#![no_std]
#![no_main]
#![feature(panic_info_message)]
#![feature(alloc_error_handler)]
#![feature(result_flattening)]

extern crate alloc;

#[macro_use]
extern crate bitflags;

#[cfg(feature = "board_qemu")]
#[path = "boards/qemu.rs"]
mod board;
#[cfg(feature = "board_fu740")]
#[path = "boards/fu740.rs"]
mod board;

#[macro_use]
mod console;
mod config;
mod drivers;
mod lang_items;
mod logger;
mod mm;
mod sbi;
mod signal;
mod syscall;
mod task;
mod timer;
mod trap;
mod utsname;
mod vfs;

use core::{
    arch::global_asm,
    sync::atomic::{AtomicBool, Ordering},
};
use riscv::register::sstatus::{self, FS};

global_asm!(include_str!("entry.asm"));
global_asm!(include_str!("link_app.S"));

fn clear_bss() {
    extern "C" {
        fn sbss();
        fn ebss();
    }
    unsafe {
        core::slice::from_raw_parts_mut(sbss as usize as *mut u8, ebss as usize - sbss as usize)
            .fill(0);
    }
}

static LOCK: AtomicBool = AtomicBool::new(false);

fn set_fs() {
    unsafe {
        sstatus::set_fs(FS::Clean);
    }
}

/// 获取当前运行的 CPU 核
#[cfg(feature = "board_qemu")]
pub fn hartid() -> usize {
    use core::arch::asm;

    let hartid;
    unsafe {
        asm! {
            "mv {}, tp",
            out(reg) hartid
        };
    }
    hartid
}
#[cfg(feature = "board_fu740")]
pub fn hartid() -> usize {
    0
}

// 避免编译器对名字进行混淆
#[no_mangle]
pub fn rust_main() -> ! {
    // rustsbi 同时开启了所有核
    // 这里先让一个核启动初始化，其它核自旋
    if hartid() == 0 {
        clear_bss();
        println!("[kernel] init core0");
        println!("[kernel] Hello, world!");
        mm::init();
        mm::activate();
        mm::remap_test();
        set_fs();
        logger::init();
        #[cfg(feature = "board_fu740")]
        board::clock_init();
        trap::init();
        trap::enable_timer_interrupt();
        timer::set_next_trigger();
        vfs::init();
        vfs::add_initproc_shell();
        task::add_initproc();
        task::run_tasks();
    } else {
        while LOCK.load(Ordering::SeqCst) == false {}
    }
    panic!("Unreachable in rust_main!");
}
