//! 该模块修改自 [buddy_system_allocator](https://github.com/rcore-os/buddy_system_allocator)
use super::{linked_list, PhysAddr, PhysPageNum};
use crate::config::*;
use crate::mm::frame_allocator::frame_allocator_test;
use core::alloc::{GlobalAlloc, Layout};
use core::cmp::{max, min};
use core::fmt;
use core::mem::size_of;
use core::ops::Deref;
use spin::Mutex;

const MAX_ORDER: usize = 40;

/// A heap that uses buddy system with configurable order.
pub struct Heap {
    // buddy system with max order of `ORDER`
    free_list: [linked_list::LinkedList; MAX_ORDER],

    // statistics
    pub allocated: usize,
    pub total: usize,
}

impl Heap {
    /// Create an empty heap
    pub const fn new() -> Self {
        Heap {
            free_list: [linked_list::LinkedList::new(); MAX_ORDER],
            allocated: 0,
            total: 0,
        }
    }

    /// Add a range of memory [start, end) to the heap
    pub unsafe fn add_to_heap(&mut self, start: PhysPageNum, end: PhysPageNum) {
        let start: usize = PhysAddr::from(start).into();
        let end: usize = PhysAddr::from(end).into();
        let mut total = 0;
        let mut current_start = start;

        while current_start < end {
            let lowbit = current_start & (!current_start + 1);
            let size = min(
                lowbit,
                min(prev_power_of_two(end - current_start), 1 << MAX_ORDER),
            );
            let order = size.trailing_zeros() as usize;
            self.free_list[order].push(current_start as *mut usize);

            // println!(
            //     "[buddy system allocator] current_start = {:#x}, size = {:#x}, order = {}",
            //     current_start, size, order
            // );

            current_start += size;
            total += size;
        }

        self.total += total;
    }

    /// Alloc a range of memory from the heap satifying `layout` requirements
    pub fn alloc(&mut self, layout: Layout) -> Option<PhysAddr> {
        let size = max(
            layout.size().next_power_of_two(),
            max(layout.align(), size_of::<usize>()),
        );
        let order = size.trailing_zeros() as usize;
        for i in order..self.free_list.len() {
            // Find the first non-empty size class
            if !self.free_list[i].is_empty() {
                // Split buffers
                for j in (order + 1..i + 1).rev() {
                    if let Some(block) = self.free_list[j].pop() {
                        unsafe {
                            self.free_list[j - 1]
                                .push((block as usize + (1 << (j - 1))) as *mut usize);
                            self.free_list[j - 1].push(block);
                        }
                    } else {
                        return None;
                    }
                }

                let result = self.free_list[order]
                    .pop()
                    .expect("current block should have free space now")
                    as usize;
                self.allocated += size;
                return Some(result.into());
            }
        }
        None
    }

    /// Dealloc a range of memory from the heap
    pub fn dealloc(&mut self, ptr: PhysAddr, layout: Layout) {
        let ptr = usize::from(ptr);
        let size = max(
            layout.size().next_power_of_two(),
            max(layout.align(), size_of::<usize>()),
        );
        let order = size.trailing_zeros() as usize;

        unsafe {
            // Put back into free list
            self.free_list[order].push(ptr as *mut usize);

            // Merge free buddy lists
            let mut current_ptr = ptr;
            let mut current_order = order;
            while current_order < self.free_list.len() {
                let buddy = current_ptr ^ (1 << current_order);
                let mut flag = false;
                for block in self.free_list[current_order].iter_mut() {
                    if block.value() as usize == buddy {
                        block.pop();
                        flag = true;
                        break;
                    }
                }

                // Free buddy found
                if flag {
                    self.free_list[current_order].pop();
                    current_ptr = min(current_ptr, buddy);
                    current_order += 1;
                    self.free_list[current_order].push(current_ptr as *mut usize);
                } else {
                    break;
                }
            }
        }

        self.allocated -= size;
    }
}

impl fmt::Debug for Heap {
    fn fmt(&self, fmt: &mut fmt::Formatter) -> fmt::Result {
        fmt.debug_struct("[buddy system allocator]")
            .field("allocated", &self.allocated)
            .field("total", &self.total)
            .finish()
    }
}

pub struct LockedHeap(Mutex<Heap>);

impl LockedHeap {
    /// Creates an empty heap
    pub const fn new() -> Self {
        LockedHeap(Mutex::new(Heap::new()))
    }
}

impl Deref for LockedHeap {
    type Target = Mutex<Heap>;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

unsafe impl GlobalAlloc for LockedHeap {
    unsafe fn alloc(&self, layout: Layout) -> *mut u8 {
        self.0
            .lock()
            .alloc(layout)
            .map_or(0 as *mut u8, |pa| usize::from(pa) as *mut u8)
    }

    unsafe fn dealloc(&self, ptr: *mut u8, layout: Layout) {
        self.0.lock().dealloc(PhysAddr::from(ptr as usize), layout)
    }
}

impl fmt::Debug for LockedHeap {
    fn fmt(&self, fmt: &mut fmt::Formatter) -> fmt::Result {
        write!(fmt, "{:#x?}", self.0.lock())
    }
}

pub(crate) fn prev_power_of_two(num: usize) -> usize {
    1 << (8 * (size_of::<usize>()) - num.leading_zeros() as usize - 1)
}

#[global_allocator]
pub static MEMORY_ALLOCATOR: LockedHeap = LockedHeap::new();

#[alloc_error_handler]
pub fn handle_alloc_error(layout: core::alloc::Layout) -> ! {
    panic!("Heap allocation error, layout = {:?}", layout);
}

pub fn init_memory_allocator() {
    extern "C" {
        fn ekernel();
    }
    unsafe {
        MEMORY_ALLOCATOR.lock().add_to_heap(
            PhysAddr::from(ekernel as usize).ceil(),
            PhysAddr::from(MEMORY_END).floor(),
        );
        #[cfg(feature = "board_fu740")]
        MEMORY_ALLOCATOR.lock().add_to_heap(
            PhysAddr::from(MEMORY_START2).ceil(),
            PhysAddr::from(MEMORY_END2).floor(),
        );
        println!("{:#x?}", MEMORY_ALLOCATOR);
        frame_allocator_test();
    }
}

pub fn alloc_pages(pages: usize) -> Option<PhysPageNum> {
    MEMORY_ALLOCATOR
        .lock()
        .alloc(Layout::from_size_align(pages * PAGE_SIZE, PAGE_SIZE).unwrap())
        .map(PhysPageNum::from)
}

pub fn alloc_page() -> Option<PhysPageNum> {
    alloc_pages(1)
}

pub fn free_pages(ppn: PhysPageNum, pages: usize) {
    MEMORY_ALLOCATOR.lock().dealloc(
        PhysAddr::from(ppn),
        Layout::from_size_align(pages * PAGE_SIZE, PAGE_SIZE).unwrap(),
    );
}

pub fn free_page(ppn: PhysPageNum) {
    free_pages(ppn, 1);
}
