use super::{
    buddy_system_allocator::{alloc_page, alloc_pages, free_page, free_pages},
    PhysPageNum,
};
use alloc::vec::Vec;
use core::fmt::{self, Debug, Formatter};

pub struct FrameTracker {
    pub ppn: PhysPageNum,
}

impl FrameTracker {
    pub fn new(ppn: PhysPageNum) -> Self {
        // page cleaning
        let bytes_array = ppn.get_bytes_array();
        for i in bytes_array {
            *i = 0;
        }
        Self { ppn }
    }
}

impl Debug for FrameTracker {
    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
        f.write_fmt(format_args!("FrameTracker:PPN={:#x}", self.ppn.0))
    }
}

impl Drop for FrameTracker {
    fn drop(&mut self) {
        frame_dealloc(self.ppn);
    }
}

pub fn frames_alloc(pages: usize) -> Option<Vec<FrameTracker>> {
    alloc_pages(pages).map(|pa| {
        (0..pages)
            .map(|i| FrameTracker::new(pa + i))
            .collect::<Vec<_>>()
    })
}

pub fn frame_alloc() -> Option<FrameTracker> {
    alloc_page().map(FrameTracker::new)
}

pub fn frames_dealloc(ppn: PhysPageNum, pages: usize) {
    free_pages(ppn, pages)
}

pub fn frame_dealloc(ppn: PhysPageNum) {
    free_page(ppn)
}

#[allow(unused)]
pub fn frame_allocator_test() {
    let mut v: Vec<FrameTracker> = Vec::new();
    for i in 0..5 {
        let frame = frame_alloc().unwrap();
        println!("{:?}", frame);
        v.push(frame);
    }
    v.clear();
    for i in 0..5 {
        let frame = frame_alloc().unwrap();
        println!("{:?}", frame);
        v.push(frame);
    }
    drop(v);
    println!("frame_allocator_test passed!");
}
