use super::{frame_alloc, FrameTracker, MapPermission};
use crate::{
    config::PAGE_SIZE,
    mm::{memory_set::MapArea, page_table::PTEFlags},
    syscall::errno::SyscallError,
    task::current_task,
};
use alloc::{collections::BTreeMap, vec::Vec};
use lazy_static::lazy_static;
use spin::Mutex;

pub struct Shm {
    pages: Vec<FrameTracker>,
}

impl Shm {
    pub fn new(num: usize) -> Self {
        Self {
            pages: { (0..num).map(|_| frame_alloc().unwrap()).collect() },
        }
    }
}

pub struct ShmManager {
    next_key: usize,
    map: BTreeMap<usize, Shm>,
}

impl ShmManager {
    pub fn new() -> Self {
        Self {
            next_key: 1,
            map: BTreeMap::new(),
        }
    }
}

lazy_static! {
    pub static ref SHM_MANAGER: Mutex<ShmManager> = Mutex::new(ShmManager::new());
}

/// 创建共享内存段，返回共享内存段标识符
///
/// `size` 为共享内存段的大小
pub fn shm_create(size: usize) -> usize {
    let num = (size + PAGE_SIZE - 1) / PAGE_SIZE;
    let mut manager = SHM_MANAGER.lock();
    let key = manager.next_key;
    manager.map.insert(key, Shm::new(num));
    manager.next_key += 1;
    key
}

/// 判断共享内存段是否存在
pub fn shm_find(key: usize) -> bool {
    let manager = SHM_MANAGER.lock();
    manager.map.get(&key).is_some()
}

/// 将共享内存段映射到用户地址空间上
///
/// ## 参数
/// - `key` 为共享内存段标识符
/// - `addr` 为插入的起始地址，为 0 表示随机插入
pub fn shm_attach(
    key: usize,
    addr: usize,
    permission: MapPermission,
) -> Result<usize, SyscallError> {
    let manager = SHM_MANAGER.lock();
    if let Some(shm) = manager.map.get(&key) {
        let task = current_task().unwrap();
        let task_inner = task.inner_exclusive_access();
        let memory_set = &task_inner.memory_set;
        if addr == 0 {
            let size = shm.pages.len() * PAGE_SIZE;
            let va = memory_set.find_insert_addr(size, None);
            let mut memory_set_inner = memory_set.inner_exclusive_access();
            for (i, page) in shm.pages.iter().enumerate() {
                memory_set_inner.page_table.map(
                    (va + i * PAGE_SIZE).into(),
                    page.ppn,
                    PTEFlags::from_bits(permission.bits()).unwrap(),
                );
            }
            memory_set_inner.areas.push(MapArea::new(
                va,
                va + size,
                crate::mm::memory_set::MapType::Share,
                permission,
            ));
            return Ok(va.0);
        }
        panic!("[shm_attach] unimplement attach addr");
    } else {
        Err(SyscallError::EINVAL)
    }
}
