mod action;
mod info;
mod signal;

pub use action::{SignalAction, SignalActions};
pub use info::SignalInfo;
pub use signal::{SignalFlags, SignalOperation, MAX_SIG};

use crate::{
    mm::{get_data, put_data, MapPermission},
    task::{
        add_task, current_task, exit_current_and_run_next, stop_current_and_run_next,
        TaskControlBlock, TaskStatus, PIDTYPE_TGID, PIDTYPE_TID, PID_HASH,
    },
    trap::TrapContext,
};
use alloc::sync::Arc;
use log::*;

/// 向指定进程发送信号
pub fn add_signal(task: Arc<TaskControlBlock>, signal: SignalFlags) {
    let mut task_inner = task.inner_exclusive_access();
    task_inner.signal_pending |= signal;
    if signal == SignalFlags::SIGCONT {
        task_inner.task_status = TaskStatus::Ready;
        drop(task_inner);
        add_task(Arc::clone(&task));
    }
}

/// 向线程组发送信号
pub fn send_thread_group_signal(pid: usize, signal: SignalFlags) {
    let hash = PID_HASH[PIDTYPE_TGID].lock();
    if let Some(vec) = hash.get(&pid) {
        for task in vec.iter() {
            add_signal(Arc::clone(task), signal);
        }
    }
}

/// 向进程组发送信号
///
/// 目前没有进程组，因此等同于线程组
pub fn send_process_group_singal(pgid: usize, signal: SignalFlags) {
    send_thread_group_signal(pgid, signal);
}

/// 向除自身以及 `init` 进程之外的所有进程发送信号
pub fn send_access_signal(signal: SignalFlags) {
    let tid = current_task().unwrap().get_tid();
    let hash = PID_HASH[PIDTYPE_TID].lock();
    for _ in hash
        .iter()
        .filter(|(key, _)| **key != tid)
        .map(|(_, vec)| vec.get(0).map(|task| add_signal(Arc::clone(task), signal)))
    {}
}

pub fn current_add_signal(signal: SignalFlags) {
    let task = current_task().unwrap();
    let mut task_inner = task.inner_exclusive_access();
    task_inner.signal_pending |= signal;
    // println!(
    //     "[K] current_add_signal:: current task sigflag {:?}",
    //     process_inner.signals
    // );
}

fn call_kernel_signal_handler(sig: usize, signal: SignalFlags) {
    match signal.default_operation() {
        SignalOperation::Terminate => {
            debug!("[call_kernel_signal_handler] Terminate: {:?}", signal);
            exit_current_and_run_next(sig as i32);
        }
        SignalOperation::Dump => {
            debug!("[call_kernel_signal_handler] Dump: {:?}", signal);
            exit_current_and_run_next((sig + 128) as i32);
        }
        SignalOperation::Ignore => {}
        SignalOperation::Stop => {
            debug!("[call_kernel_signal_handler] Stop: {:?}", signal);
            stop_current_and_run_next();
        }
        SignalOperation::Continue => {}
    };
    let task = current_task().unwrap();
    let mut task_inner = task.inner_exclusive_access();
    task_inner.signal_pending ^= signal;
}

/// 在用户态堆栈建立页面保存相关信息
///
/// ```
/// |   (checkout)      |   低地址
/// |   signal mask     |
/// |   Trap context    |   高地址
fn setup_frame(sig: usize, sigaction: SignalAction) {
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    let token = task_inner.get_user_token();
    let mut trap_ctx = task_inner.get_trap_cx();
    let mut user_sp = trap_ctx.x[2];

    // 保存 Trap 上下文
    user_sp -= core::mem::size_of::<TrapContext>();
    put_data(token, user_sp as *mut TrapContext, *trap_ctx);
    // signal mask
    user_sp -= core::mem::size_of::<SignalFlags>();
    put_data(token, user_sp as *mut SignalFlags, task_inner.signal_mask);
    // checkout
    user_sp -= core::mem::size_of::<usize>();
    put_data(token, user_sp as *mut usize, 0xdeadbeef);

    trap_ctx.sepc = sigaction.handler;

    if sigaction.restorer == 0 {
        // 如果 libc 库没有提供信号恢复代码，需要自行补充，
        // 这里将分配一个 mmap 区域页用于映射信号恢复代码，之后把该段地址填入相应的信号处理函数结构
        let (start_va, _) = task_inner.memory_set.insert_framed_area_randomly(
            4096,
            None,
            MapPermission::R | MapPermission::X | MapPermission::W | MapPermission::U,
        );
        trace!(
            "[setup_frame] mmap signal restore area, signum = {}, start_va = {:#x?}",
            sig,
            start_va
        );

        let mut rip = start_va.0;
        // li a7, 139
        put_data(token, rip as *mut u32, 0x08b00893);
        rip += core::mem::size_of::<u32>();
        // ecall
        put_data(token, rip as *mut u32, 0x00000073);

        task_inner.signal_actions.inner_exclusive_access().table[sig].restorer = start_va.0;
        // ra
        trap_ctx.x[1] = start_va.0;
    } else {
        // ra
        trap_ctx.x[1] = sigaction.restorer;
    }

    // sp
    trap_ctx.x[2] = user_sp;
    // a0
    trap_ctx.x[10] = sig;
}

pub fn restore_frame() {
    let task = current_task().unwrap();
    let mut task_inner = task.inner_exclusive_access();
    let token = task_inner.get_user_token();
    let trap_ctx = task_inner.get_trap_cx();
    let mut user_sp = trap_ctx.x[2];

    // checkout
    let checkout = get_data(token, user_sp as *const usize);
    trace!("[restore_frame] checkout = {:#x}", checkout);
    user_sp += core::mem::size_of::<usize>();
    // signal mask
    task_inner.signal_mask = get_data(token, user_sp as *const SignalFlags);
    user_sp += core::mem::size_of::<SignalFlags>();
    // Trap cx
    *trap_ctx = get_data(token, user_sp as *const TrapContext);
}

fn call_user_signal_handler(sig: usize, signal: SignalFlags, sigaction: SignalAction) {
    debug!("[call_user_signal_handler] signal = {:?}", signal);
    setup_frame(sig, sigaction);
    let task = current_task().unwrap();
    let mut task_inner = task.inner_exclusive_access();
    task_inner.signal_mask |= sigaction.mask;
    task_inner.signal_mask |= signal;
    task_inner.signal_pending ^= signal;
}

pub fn handle_signals() {
    for signum in 1..(MAX_SIG + 1) {
        let task = current_task().unwrap();
        let mut task_inner = task.inner_exclusive_access();
        let signal = SignalFlags::from_sig(signum);
        if task_inner.signal_pending.contains(signal) && (!task_inner.signal_mask.contains(signal))
        {
            let sigaction = task_inner.signal_actions.inner_exclusive_access().table[signum];
            match sigaction.handler {
                // 缺省操作
                0 => {
                    drop(task_inner);
                    drop(task);
                    call_kernel_signal_handler(signum, signal);
                }
                // 忽略信号
                1 => {
                    task_inner.signal_pending ^= signal;
                }
                // 返回到用户的信号处理程序
                _ => {
                    drop(task_inner);
                    drop(task);
                    call_user_signal_handler(signum, signal, sigaction);
                    return;
                }
            };
        }
    }
}
