use super::errno::*;
use crate::mm::{
    get_data, put_data, translated_byte_buffer, translated_ref, translated_refmut, translated_str,
    UserBuffer,
};
use crate::signal::SignalFlags;
use crate::task::{current_task, current_user_token, suspend_current_and_run_next};
use crate::timer::{get_timespec, Timespec};
use crate::vfs::{
    dev_root_inode, get_dir, insert_mounts, make_pipe, open_file_at, procfs_root_inode,
    remove_mounts, root_inode, types::*, Inode, Memfile, ModeFlags, Mount, OpenFlags, MOUNTS,
};
use alloc::string::String;
use alloc::sync::Arc;
use alloc::{vec, vec::Vec};
use log::*;

/// 获取文件绝对路径
fn get_path(mut inode: Arc<dyn Inode>) -> String {
    let mut path = inode.get_name();
    let root_inode = root_inode();
    if Arc::as_ptr(&inode) != Arc::as_ptr(&root_inode) {
        loop {
            inode = inode.lookup("..").unwrap();
            path = String::from("/") + &path;
            if Arc::as_ptr(&inode) == Arc::as_ptr(&root_inode) {
                break;
            }
            path = inode.get_name() + &path;
        }
    }
    path
}

pub fn sys_getcwd(buf: *mut u8, size: usize) -> Result<usize, SyscallError> {
    let token = current_user_token();
    let mut buffer = UserBuffer::new(translated_byte_buffer(token, buf, size));
    let inode = Arc::clone(
        &current_task()
            .unwrap()
            .inner_exclusive_access()
            .fs_info
            .inner_exclusive_access()
            .cwd,
    );
    let path = get_path(inode) + "\0";
    if path.len() > size {
        return Err(SyscallError::ERANGE);
    }
    buffer.write(path.as_bytes());
    return Ok(buf as usize);
}

pub fn sys_write(fd: usize, buf: *const u8, len: usize) -> Result<usize, SyscallError> {
    let token = current_user_token();
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if let Some(file) = task_inner.fd_table.get(fd) {
        // 文件读写会挂起当前线程，需要释放锁
        drop(task_inner);
        drop(task);
        Ok(file.write(UserBuffer::new(translated_byte_buffer(token, buf, len))))
    } else {
        Err(SyscallError::EBADF)
    }
}

pub fn sys_read(fd: usize, buf: *const u8, len: usize) -> Result<usize, SyscallError> {
    let token = current_user_token();
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if let Some(file) = task_inner.fd_table.get(fd) {
        // 文件读写会挂起当前线程，需要释放锁
        drop(task_inner);
        drop(task);
        Ok(file.read(UserBuffer::new(translated_byte_buffer(token, buf, len))))
    } else {
        Err(SyscallError::EBADF)
    }
}

pub fn sys_lseek(fd: usize, offset: isize, whence: usize) -> Result<usize, SyscallError> {
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if let Some(file) = task_inner.fd_table.get(fd) {
        file.lseek(offset, whence)
    } else {
        Err(SyscallError::EBADF)
    }
}

pub fn sys_openat(
    dirfd: isize,
    path: *const u8,
    flags: u32,
    mode: u32,
) -> Result<usize, SyscallError> {
    let token = current_user_token();
    let path = translated_str(token, path);
    //println!(
    // "in sys_openat: dirfd = {}, path = {}, flags = {:#x}, mode = {:#o}",
    //     dirfd, path, flags, mode
    //);
    let flags = OpenFlags::from_bits(flags).unwrap();
    let mode = if flags.contains(OpenFlags::O_CREAT) {
        ModeFlags::from_bits(mode).unwrap()
    } else {
        ModeFlags::empty()
    };
    debug!(
        "[sys_openat] dirfd = {}, path = {:?}, flags = {:?}, mode = {:?}",
        dirfd, path, flags, mode
    );
    open_file_at(dirfd, &path, flags, mode).map(|file| {
        if flags.contains(OpenFlags::O_APPEND) {
            file.lseek(0, SEEK_END).unwrap();
        }
        let task = current_task().unwrap();
        let task_inner = task.inner_exclusive_access();
        let fd = match task_inner.try_alloc_fd() {
            Ok(fd) => fd,
            Err(e) => return Err(e),
        };
        task_inner
            .fd_table
            .set_fd(fd, file, flags.contains(OpenFlags::O_CLOEXEC));
        Ok(fd)
    })?
}

pub fn sys_mkdirat(dirfd: isize, path: *const u8, mode: u32) -> Result<usize, SyscallError> {
    let token = current_user_token();
    let path = translated_str(token, path);
    let mode = ModeFlags::from_bits(mode).unwrap();
    let inode = get_dir(dirfd, &path)?;
    let name = path
        .trim_end_matches('/')
        .split('/')
        .collect::<Vec<_>>()
        .pop()
        .unwrap();
    if inode.lookup(name).is_some() {
        Err(SyscallError::EEXIST)
    } else {
        inode.mkdir(name, mode);
        Ok(0)
    }
}

pub fn sys_chdir(path: *const u8) -> Result<usize, SyscallError> {
    let token = current_user_token();
    let path = translated_str(token, path);
    // 在 open_file_at 中会用到当前进程变量，会获得锁，
    open_file_at(AT_FDCWD, &path, OpenFlags::O_RDWR, ModeFlags::empty()).map(|file| {
        let task = current_task().unwrap();
        let task_inner = task.inner_exclusive_access();
        let mut fs_info_inner = task_inner.fs_info.inner_exclusive_access();
        fs_info_inner.cwd = file.get_inode();
        0
    })
}

pub fn sys_close(fd: usize) -> Result<usize, SyscallError> {
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if task_inner.fd_table.get(fd).is_some() {
        task_inner.fd_table.remove(fd);
        // 释放节点
        crate::vfs::sync();
        Ok(0)
    } else {
        Err(SyscallError::EBADF)
    }
}

pub fn sys_getdents64(fd: usize, buf: *const u8, len: usize) -> Result<usize, SyscallError> {
    let token = current_user_token();
    let mut buf = UserBuffer::new(translated_byte_buffer(token, buf, len));
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if let Some(file) = task_inner.fd_table.get(fd) {
        let inode = file.get_inode();
        let mut vec = Vec::new();
        let mut off = file.lseek(0, SEEK_CUR).unwrap();
        let mut res = 0;
        while let Some(dirent) = inode.read_dentry(off) {
            if res + dirent.d_reclen as usize > len {
                break;
            }
            res += dirent.d_reclen as usize;
            vec.extend_from_slice(dirent.as_bytes());
            off = dirent.d_off as usize;
        }
        buf.write(vec.as_slice());
        file.lseek(off as isize, SEEK_SET).unwrap();
        Ok(res)
    } else {
        Err(SyscallError::EBADF)
    }
}

pub fn sys_readv(fd: usize, iov: *const IOVec, iovcnt: usize) -> Result<usize, SyscallError> {
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if let Some(file) = task_inner.fd_table.get(fd) {
        drop(task_inner);
        drop(task);
        Ok(file.readv(iov, iovcnt))
    } else {
        Err(SyscallError::EBADF)
    }
}

pub fn sys_writev(fd: usize, iov: *const IOVec, iovcnt: usize) -> Result<usize, SyscallError> {
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if let Some(file) = task_inner.fd_table.get(fd) {
        drop(task_inner);
        drop(task);
        Ok(file.writev(iov, iovcnt))
    } else {
        Err(SyscallError::EBADF)
    }
}

pub fn sys_sendfile(
    out_fd: usize,
    in_fd: usize,
    ptr: *mut i64,
    count: usize,
) -> Result<usize, SyscallError> {
    const BUF_SIZE: usize = 4096;
    let token = current_user_token();
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    let mut buf = Vec::<u8>::with_capacity(BUF_SIZE);
    let mut len;
    let mut size = 0;

    if let Some(fout) = task_inner.fd_table.get(out_fd) {
        if let Some(fin) = task_inner.fd_table.get(in_fd) {
            drop(task_inner);
            drop(task);
            let old_off = fin.lseek(0, SEEK_CUR).unwrap();
            if ptr as usize != 0 {
                fin.lseek(*translated_ref(token, ptr) as isize, SEEK_SET)
                    .unwrap();
            }
            loop {
                let buf_read = UserBuffer::new(vec![unsafe {
                    core::slice::from_raw_parts_mut(buf.as_mut_ptr(), BUF_SIZE.min(count - size))
                }]);
                len = fin.read(buf_read);
                if len == 0 {
                    break;
                }
                let buf_write = UserBuffer::new(vec![unsafe {
                    core::slice::from_raw_parts_mut(buf.as_mut_ptr(), len)
                }]);
                assert_eq!(fout.write(buf_write), len);
                size += len;
                if size >= count || len < BUF_SIZE {
                    break;
                }
            }
            if ptr as usize != 0 {
                *translated_refmut(token, ptr) = fin.lseek(0, SEEK_CUR).unwrap() as i64;
                fin.lseek(old_off as isize, SEEK_SET).unwrap();
            }
            Ok(size)
        } else {
            Err(SyscallError::EBADF)
        }
    } else {
        Err(SyscallError::EBADF)
    }
}

pub fn sys_fstat(fd: usize, buf: *mut Kstat) -> Result<usize, SyscallError> {
    let token = current_user_token();
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if let Some(file) = task_inner.fd_table.get(fd) {
        let inode = file.get_inode();
        put_data(token, buf, inode.get_attr());
        Ok(0)
    } else {
        Err(SyscallError::EBADF)
    }
}

pub fn sys_newfstatat(
    dirfd: isize,
    path: *const u8,
    buf: *mut Kstat,
    _flags: usize,
) -> Result<usize, SyscallError> {
    let token = current_user_token();
    let path = translated_str(token, path);
    debug!("[sys_newfstatat] dirfd = {}, path = {}", dirfd, path);
    open_file_at(dirfd, &path, OpenFlags::O_RDONLY, ModeFlags::empty()).map(|file| {
        let inode = file.get_inode();
        *translated_refmut(token, buf) = inode.get_attr();
        0
    })
}

pub fn sys_pipe2(pipe: *mut u32, flags: u32) -> Result<usize, SyscallError> {
    bitflags! {
        struct PipeFlags: u32 {
            const O_CLOEXEC = OpenFlags::O_CLOEXEC.bits();
        }
    }

    let flags = PipeFlags::from_bits(flags).unwrap();
    let token = current_user_token();
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    let (pipe_read, pipe_write) = make_pipe();
    let read_fd = task_inner.alloc_fd();
    task_inner
        .fd_table
        .set_fd(read_fd, pipe_read, flags.contains(PipeFlags::O_CLOEXEC));
    let write_fd = task_inner.alloc_fd();

    task_inner
        .fd_table
        .set_fd(write_fd, pipe_write, flags.contains(PipeFlags::O_CLOEXEC));
    // println!("in sys_pipe2: read_fd = {}, write_fd = {}", read_fd, write_fd);
    *translated_refmut(token, pipe) = read_fd as u32;
    *translated_refmut(token, unsafe { pipe.add(1) }) = write_fd as u32;
    Ok(0)
}

pub fn sys_dup(fd: usize) -> Result<usize, SyscallError> {
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if let Some(file) = task_inner.fd_table.get(fd) {
        let new_fd = match task_inner.try_alloc_fd() {
            Ok(new_fd) => new_fd,
            Err(err) => return Err(err),
        };
        task_inner.fd_table.set_fd(new_fd, file, false);
        Ok(new_fd)
    } else {
        Err(SyscallError::EBADF)
    }
}

pub fn sys_dup3(oldfd: usize, newfd: usize, flags: u32) -> Result<usize, SyscallError> {
    bitflags! {
        struct DupFlags: u32 {
            const O_CLOEXEC = OpenFlags::O_CLOEXEC.bits();
        }
    }
    if oldfd == newfd {
        return Err(SyscallError::EINVAL);
    }
    let flags = DupFlags::from_bits(flags).unwrap();
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if let Some(file) = task_inner.fd_table.get(oldfd) {
        match task_inner
            .fd_table
            .try_set_fd(newfd, file, flags.contains(DupFlags::O_CLOEXEC))
        {
            Ok(_) => {}
            Err(e) => {
                println!("{:?}", e);
                return Err(e);
            }
        }
        Ok(newfd)
    } else {
        Err(SyscallError::EBADF)
    }
}

pub fn sys_fcntl(fd: usize, cmd: usize, arg: i32) -> Result<usize, SyscallError> {
    const F_DUPFD: usize = 0;
    const F_GETFD: usize = 1;
    const F_SETFD: usize = 2;
    const F_GETFL: usize = 3;
    const F_SETFL: usize = 4;
    const F_DUPFD_CLOEXEC: usize = 1030;

    bitflags! {
        struct FdFlags: u32 {
            const FD_CLOEXEC = 1;
        }
    }

    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if let Some(file) = task_inner.fd_table.get(fd) {
        match cmd {
            F_DUPFD | F_DUPFD_CLOEXEC => {
                let newfd = task_inner.alloc_fd();
                if newfd >= arg as usize {
                    task_inner
                        .fd_table
                        .set_fd(newfd, file, cmd == F_DUPFD_CLOEXEC);
                    Ok(newfd)
                } else {
                    task_inner
                        .fd_table
                        .set_fd(arg as usize, file, cmd == F_DUPFD_CLOEXEC);
                    Ok(arg as usize)
                }
            }
            F_GETFD => Ok(task_inner.fd_table.get_flag(fd).unwrap() as usize),
            F_SETFD => {
                if let Some(flag) = FdFlags::from_bits(arg as u32) {
                    task_inner
                        .fd_table
                        .set_flag(fd, flag.contains(FdFlags::FD_CLOEXEC));
                    Ok(0)
                } else {
                    Err(SyscallError::EINVAL)
                }
            }
            F_GETFL => Ok(file.get_flags().bits() as usize),
            F_SETFL => {
                let flag = OpenFlags::from_bits(arg as u32).unwrap();
                file.set_flags(flag);
                if flag.contains(OpenFlags::O_NONBLOCK) {
                    warn!("[sys_fcntl] unimplement O_NONBLOCK flag = {:?}", flag);
                }
                Ok(0)
            }
            _ => {
                panic!("in sys_fcntl: unsupport operation, cmd = {}", cmd);
            }
        }
    } else {
        Err(SyscallError::EBADF)
    }
}

pub fn sys_unlinkat(dirfd: isize, path: *const u8, _flags: u32) -> Result<usize, SyscallError> {
    let token = current_user_token();
    let path = translated_str(token, path);
    get_dir(dirfd, &path).map(|dir| {
        let name = path
            .trim_end_matches('/')
            .split('/')
            .collect::<Vec<_>>()
            .pop()
            .unwrap();
        dir.unlink(name);
        0
    })
}

pub fn sys_faccessat(dirfd: isize, path: *const u8, mode: u32) -> Result<usize, SyscallError> {
    bitflags! {
        struct FaccessFlags: u32 {
            const F_OK = 0;
            const X_OK = 1;
            const W_OK = 2;
            const R_OK = 4;
        }
    }

    let token = current_user_token();
    let path = translated_str(token, path);
    if let Some(mode) = FaccessFlags::from_bits(mode) {
        let file = open_file_at(dirfd, &path, OpenFlags::O_RDWR, ModeFlags::empty())?;
        if mode.contains(FaccessFlags::W_OK) && !file.get_mode().contains(ModeFlags::WRITE) {
            return Err(SyscallError::EROFS);
        }
        Ok(0)
    } else {
        Err(SyscallError::EINVAL)
    }
}

pub fn sys_renameat2(
    old_dirfd: isize,
    old_path: *const u8,
    new_dirfd: isize,
    new_path: *const u8,
    _flags: u32,
) -> Result<usize, SyscallError> {
    let token = current_user_token();
    let old_path = translated_str(token, old_path);
    let new_path = translated_str(token, new_path);
    let old_dir = get_dir(old_dirfd, &old_path)?;
    let old_name = old_path
        .trim_end_matches('/')
        .split('/')
        .collect::<Vec<_>>()
        .pop()
        .unwrap();
    get_dir(new_dirfd, &new_path).map(|new_dir| {
        let new_name = new_path
            .trim_end_matches('/')
            .split('/')
            .collect::<Vec<_>>()
            .pop()
            .unwrap();
        old_dir.rename(old_name, new_dir, new_name);
        0
    })
}

pub fn sys_utimensat(
    dirfd: isize,
    path: *const u8,
    times: *const [Timespec; 2],
    flags: u32,
) -> Result<usize, SyscallError> {
    const UTIME_NOW: usize = 0x3fffffff; // 当前时间
    const UTIME_OMIT: usize = 0x3ffffffe; // 时间戳不变
    const AT_SYMLINK_NOFOLLOW: u32 = 0x100;
    let now = Timespec::now();
    let mut atime_sec = Some(now.tv_sec as i64);
    let mut atime_nsec = Some(now.tv_nsec as i64);
    let mut mtime_sec = Some(now.tv_sec as i64);
    let mut mtime_nsec = Some(now.tv_nsec as i64);

    let token = current_user_token();
    if flags == AT_SYMLINK_NOFOLLOW {
        warn!("[sys_utimensat] unimplent AT_SYMLINK_NOFOLLOW");
        return Err(SyscallError::EINVAL);
    } else if flags != 0 {
        warn!("[sys_utimensat] unknown flags");
        return Err(SyscallError::EINVAL);
    }

    let times: &[Timespec; 2] = if times.is_null() {
        &[Timespec {
            tv_sec: 0,
            tv_nsec: UTIME_NOW,
        }; 2]
    } else {
        translated_ref(token, times)
    };

    let inode;
    if !path.is_null() {
        let path = translated_str(token, path);
        match open_file_at(dirfd, &path, OpenFlags::O_RDWR, ModeFlags::empty()) {
            Ok(file) => {
                inode = file.get_inode();
            }
            Err(e) => return Err(e),
        }
    } else {
        let task = current_task().unwrap();
        let task_inner = task.inner_exclusive_access();
        match task_inner.fd_table.get(dirfd as usize) {
            Some(file) => {
                inode = file.get_inode();
            }
            None => return Err(SyscallError::EBADF),
        }
    }
    match times[0].tv_nsec {
        UTIME_NOW => (),
        UTIME_OMIT => {
            atime_sec = None;
            atime_nsec = None
        }
        _ => {
            atime_sec = Some(times[0].tv_sec as i64);
            atime_nsec = Some(times[0].tv_nsec as i64)
        }
    }
    match times[1].tv_nsec {
        UTIME_NOW => (),
        UTIME_OMIT => {
            mtime_sec = None;
            mtime_nsec = None
        }
        _ => {
            mtime_sec = Some(times[1].tv_sec as i64);
            mtime_nsec = Some(times[1].tv_nsec as i64)
        }
    }
    inode.set_timestamps(atime_sec, atime_nsec, mtime_sec, mtime_nsec);
    return Ok(0);
}

pub fn sys_mount(
    source: *const u8,
    target: *const u8,
    filesystemtype: *const u8,
    _mountflags: usize,
    _data: *const usize,
) -> Result<usize, SyscallError> {
    let token = current_user_token();
    let source = translated_str(token, source);
    let target = translated_str(token, target);
    let filesystemtype = translated_str(token, filesystemtype);
    info!(
        "[sys_mount] source = {}, target = {}, filesystemtype = {}",
        source, target, filesystemtype
    );
    let inode = open_file_at(
        AT_FDCWD,
        &target,
        OpenFlags::O_RDONLY | OpenFlags::O_DIRECTORY,
        ModeFlags::empty(),
    )?
    .get_inode();
    match filesystemtype.as_str() {
        "vfat" => {
            /* 由于不确定 source 指定的路径要进行怎样的解析，如本身指向一个设备，所以需要我们打开
            该设备的同时进行解析得到相应文件系统的根目录还是可以直接指向一个文件根目录节点，因此先
            当作一个 inode 节点作为根节点进行解析，之后再更改
             */
            let root = open_file_at(AT_FDCWD, &source, OpenFlags::O_RDONLY, ModeFlags::empty())?
                .get_inode();
            insert_mounts(Mount::new(filesystemtype, root, inode));
        }
        "proc" => {
            // 挂载 procfs 的 source 参数为字符串 "proc"
            insert_mounts(Mount::new(filesystemtype, procfs_root_inode(), inode));
        }
        "devtmpfs" => {
            insert_mounts(Mount::new(filesystemtype, dev_root_inode(), inode));
        }
        _ => {
            warn!("[sys_mount] unsupport filesystem: {:?}", filesystemtype);
        }
    }
    Ok(0)
}

pub fn sys_umount2(target: *const u8, _flags: i32) -> Result<usize, SyscallError> {
    let token = current_user_token();
    let target = translated_str(token, target);
    let inode =
        open_file_at(AT_FDCWD, &target, OpenFlags::O_RDONLY, ModeFlags::empty())?.get_inode();
    remove_mounts(inode)
}

pub fn sys_memfd_create(name: *const u8, flags: u32) -> Result<usize, SyscallError> {
    bitflags! {
        struct MemfdFlags: u32 {
            const MFD_CLOEXEC = 1;
        }
    }

    let token = current_user_token();
    let name = translated_str(token, name);
    let flags = MemfdFlags::from_bits(flags).unwrap();
    let file = Arc::new(Memfile::new(name));
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    let fd = task_inner.fd_table.alloc_fd();
    task_inner
        .fd_table
        .set_fd(fd, file, flags.contains(MemfdFlags::MFD_CLOEXEC));
    Ok(fd)
}

pub fn sys_statfs(path: *const u8, buf: *mut Statfs) -> Result<usize, SyscallError> {
    const PROC_SUPER_MAGIC: u64 = 0x9fa0;
    const TMPFS_MAGIC: u64 = 0x01021994;
    const MSDOS_SUPER_MAGIC: u64 = 0x4d44;

    bitflags! {
        struct MountFlags: u64 {
            const ST_RDONLY = 1;
            const ST_NOSUID = 2;
            const ST_NODEV  = 4;
            const ST_NOEXEC = 8;
            const ST_SYNCHRONOUS = 16;
            const ST_MANDLOCK    = 64;
            const ST_WRITE       = 128;
            const ST_APPEND      = 256;
            const ST_IMMUTABLE   = 512;
            const ST_NOATIME     = 1024;
            const ST_NODIRATIME  = 2048;
            const ST_RELATIME    = 4096;
        }
    }

    let token = current_user_token();
    let path = translated_str(token, path);
    if path == "/" {
        let res = Statfs {
            f_type: MSDOS_SUPER_MAGIC,
            f_bsize: 4096,
            f_blocks: 1,
            f_bfree: 0,
            f_bavail: 0,
            f_files: 1,
            f_ffree: 0,
            f_fsid: [0; 2],
            f_namelen: 255,
            f_frsize: 4096,
            f_flags: (MountFlags::ST_NOSUID | MountFlags::ST_RELATIME).bits(),
            f_spare: [0; 4],
        };
        put_data(token, buf, res);
        return Ok(0);
    }
    let inode = open_file_at(AT_FDCWD, &path, OpenFlags::O_RDONLY, ModeFlags::empty())?.get_inode();
    let mounts = MOUNTS.lock();
    if let Some((id, _)) = mounts
        .iter()
        .enumerate()
        .find(|(_, mount)| Arc::as_ptr(&mount.root) == Arc::as_ptr(&inode))
    {
        match mounts[id].filesystemtype.as_str() {
            "proc" => {
                let res = Statfs {
                    f_type: PROC_SUPER_MAGIC,
                    f_bsize: 4096,
                    f_blocks: 0,
                    f_bfree: 0,
                    f_bavail: 0,
                    f_files: 0,
                    f_ffree: 0,
                    f_fsid: [0; 2],
                    f_namelen: 255,
                    f_frsize: 4096,
                    f_flags: (MountFlags::ST_NOSUID
                        | MountFlags::ST_NODEV
                        | MountFlags::ST_NOEXEC
                        | MountFlags::ST_RELATIME)
                        .bits(),
                    f_spare: [0; 4],
                };
                put_data(token, buf, res);
            }
            "devtmpfs" => {
                let res = Statfs {
                    f_type: TMPFS_MAGIC,
                    f_bsize: 4096,
                    f_blocks: 0,
                    f_bfree: 0,
                    f_bavail: 0,
                    f_files: 0,
                    f_ffree: 0,
                    f_fsid: [0; 2],
                    f_namelen: 255,
                    f_frsize: 4096,
                    f_flags: (MountFlags::ST_NOSUID
                        | MountFlags::ST_NOEXEC
                        | MountFlags::ST_RELATIME)
                        .bits(),
                    f_spare: [0; 4],
                };
                put_data(token, buf, res);
            }
            filesystemtype => {
                warn!("[sys_statfs] unsupport filesystem: {:?}", filesystemtype);
            }
        }
    } else {
        warn!("[sys_statfs] can't find mount point");
    }
    Ok(0)
}

pub fn sys_readlinkat(
    dirfd: isize,
    path: *const u8,
    buf: *const u8,
    len: usize,
) -> Result<usize, SyscallError> {
    let token = current_user_token();
    let path = translated_str(token, path);
    debug!("[sys_readlinkat] dirfd = {}, path = {}", dirfd, path);
    if &path == "/proc/self/exe" {
        let task = current_task().unwrap();
        let pwd = get_path(Arc::clone(
            &task.inner_exclusive_access().exec.as_ref().unwrap(),
        ));
        let mut buf = UserBuffer::new(translated_byte_buffer(token, buf, len));
        let size = pwd.len().min(len);
        buf.write(&pwd.as_bytes()[..size]);
        return Ok(size);
    } else {
        todo!()
    }
}

pub fn sys_pselect6(
    nfds: i32,
    readfds: *mut FdSet,
    writefds: *mut FdSet,
    exceptfds: *mut FdSet,
    timeout: *const Timespec,
    data: *mut usize,
) -> Result<usize, SyscallError> {
    // data 所属类型结构体
    // struct {
    //     const kernel_sigset_t *ss;   /* Pointer to signal set */
    //     size_t ss_len;   /* Size (in bytes) of object pointed to by 'ss' */
    // };

    let token = current_user_token();
    let check_readfds = if readfds.is_null() {
        FdSet::new()
    } else {
        get_data(token, readfds)
    };
    let check_writefds = if writefds.is_null() {
        FdSet::new()
    } else {
        get_data(token, writefds)
    };
    let end = if timeout.is_null() {
        None
    } else {
        Some(get_data(token, timeout) + get_timespec())
    };
    let mask = if data.is_null() {
        None
    } else {
        let sigset_pointer = get_data(token, data);
        if sigset_pointer == 0 {
            None
        } else {
            Some(get_data(token, sigset_pointer as *const SignalFlags))
        }
    };
    debug!(
        "[sys_pselect6] nfds = {}, readfds = {:?}, writefds = {:?}, timeout = {:?}, mask = {:?}",
        nfds,
        check_readfds,
        check_writefds,
        get_data(token, timeout),
        mask
    );

    let mut ready_readfds = FdSet::new();
    let mut ready_writefds = FdSet::new();
    let mut res;

    loop {
        res = 0;
        let task = current_task().unwrap();
        let mut task_inner = task.inner_exclusive_access();
        let fd_table = Arc::clone(&task_inner.fd_table);
        // 检测读事件
        for fd in 0..nfds as usize {
            if check_readfds.is_set(fd) {
                if let Some(file) = fd_table.get(fd) {
                    if file.ready_read() {
                        ready_readfds.set(fd);
                        res += 1;
                    }
                }
            }
        }
        // 检测写事件
        for fd in 0..nfds as usize {
            if check_writefds.is_set(fd) {
                if let Some(file) = fd_table.get(fd) {
                    if file.ready_write() {
                        ready_writefds.set(fd);
                        res += 1;
                    }
                }
            }
        }
        if res != 0 {
            break;
        }
        if let Some(end_time) = end {
            if get_timespec() > end_time {
                break;
            } else {
                let origin_mask = task_inner.signal_mask;
                if let Some(signal_mask) = mask {
                    task_inner.signal_mask = signal_mask;
                }
                drop(fd_table);
                drop(task_inner);
                drop(task);
                suspend_current_and_run_next();
                if mask.is_some() {
                    current_task().unwrap().inner_exclusive_access().signal_mask = origin_mask;
                }
            }
        }
    }

    if !readfds.is_null() {
        put_data(token, readfds, ready_readfds);
    }
    if !writefds.is_null() {
        put_data(token, writefds, ready_writefds);
    }
    if !exceptfds.is_null() {
        put_data(token, exceptfds, FdSet::new());
    }
    Ok(res)
}

pub fn sys_ppoll(
    fds: *mut Pollfd,
    nfds: usize,
    timeout: *const Timespec,
    sigmask: *const SignalFlags,
) -> Result<usize, SyscallError> {
    bitflags! {
        struct PollFlags: i16 {
            const POLLIN = 0x001;
            const POLLOUT = 0x004;
        }
    }

    let token = current_user_token();
    let mut pollfds = (0..nfds)
        .map(|i| {
            let mut pollfd = get_data(token, unsafe { fds.offset(i as isize) });
            pollfd.revents = 0;
            pollfd
        })
        .collect::<Vec<_>>();
    let end = if timeout.is_null() {
        None
    } else {
        Some(get_data(token, timeout) + get_timespec())
    };
    let mask = if sigmask.is_null() {
        None
    } else {
        Some(get_data(token, sigmask))
    };
    debug!(
        "[sys_ppoll] fds = {:?}, nfds = {}, timeout = {:?}, sigmask = {:?}",
        pollfds, nfds, end, mask
    );

    let mut res;
    loop {
        res = 0;
        let task = current_task().unwrap();
        let mut task_inner = task.inner_exclusive_access();
        let fd_table = Arc::clone(&task_inner.fd_table);
        for pollfd in pollfds.iter_mut() {
            if pollfd.fd >= 0 {
                let flags = PollFlags::from_bits(pollfd.events).unwrap();
                if let Some(file) = fd_table.get(pollfd.fd as usize) {
                    if flags.contains(PollFlags::POLLIN) {
                        if file.ready_read() {
                            pollfd.revents |= PollFlags::POLLIN.bits();
                            res += 1;
                        }
                    }
                    if flags.contains(PollFlags::POLLOUT) {
                        if file.ready_write() {
                            pollfd.revents |= PollFlags::POLLOUT.bits();
                            res += 1;
                        }
                    }
                }
            }
        }
        if res != 0 {
            break;
        }
        if let Some(end_time) = end {
            if get_timespec() > end_time {
                break;
            }
        }
        // 无限阻塞或者未达到规定时间
        let origin_mask = task_inner.signal_mask;
        if let Some(signal_mask) = mask {
            task_inner.signal_mask = signal_mask;
        }
        drop(fd_table);
        drop(task_inner);
        drop(task);
        suspend_current_and_run_next();
        if mask.is_some() {
            current_task().unwrap().inner_exclusive_access().signal_mask = origin_mask;
        }
    }
    debug!("[sys_ppoll] ret fds = {:?}", pollfds);
    (0..nfds).for_each(|i| {
        put_data(token, unsafe { fds.offset(i as isize) }, pollfds[i]);
    });
    Ok(res)
}

pub fn sys_ftruncate(fd: i32, length: i64) -> Result<usize, SyscallError> {
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    if let Some(file) = task_inner.fd_table.get(fd as usize) {
        if length == 0 {
            file.get_inode().truncate();
        } else {
            panic!("[sys_ftruncate] unimplement truncate length > 0");
        }
        Ok(0)
    } else {
        Err(SyscallError::EBADF)
    }
}

pub fn sys_umask(_umask: usize) -> Result<usize, SyscallError> {
    return Ok(0);
}

pub fn sys_pread64(
    fd: usize,
    ptr: *mut u8,
    len: usize,
    offset: usize,
) -> Result<usize, SyscallError> {
    // debug!(
    //     "[sys_pread] fd: {}, ptr: {:?}, len: {}, offset: {}",
    //     fd, ptr, len, offset
    // );
    let token = current_user_token();
    let task = current_task().unwrap();
    let inner = task.inner_exclusive_access();
    let fd_table = &inner.fd_table;

    if let Some(file) = fd_table.get(fd) {
        drop(inner);
        drop(task);
        let old_off = file.lseek(0, SEEK_CUR)?;
        let _ = file.lseek(offset as isize, SEEK_SET)?;
        let result = file.read(UserBuffer::new(translated_byte_buffer(token, ptr, len)));
        file.lseek(old_off as isize, SEEK_SET)?;
        Ok(result)
    } else {
        return Err(SyscallError::EBADF);
    }
}
