use super::errno::*;
use crate::{
    mm::{
        shm_attach, shm_create, shm_find, translated_byte_buffer, MapPermission, UserBuffer,
        VirtAddr,
    },
    task::{current_task, current_user_token},
    vfs::types::*,
};

pub fn sys_brk(addr: usize) -> Result<usize, SyscallError> {
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    let mm = &task_inner.memory_set;
    let brk = mm.inner_exclusive_access().brk;
    if addr != 0 {
        let permission = MapPermission::U | MapPermission::R | MapPermission::W;
        let new_brk = VirtAddr::from(addr).ceil().into();
        mm.insert_framed_area(brk, new_brk, permission);

        // 初赛brk与标准有所不同，以下是初赛版本
        // mm.inner_exclusive_access().brk = VirtAddr::from(addr);
        // Ok(addr)

        // 以下是规范版本
        mm.inner_exclusive_access().brk = new_brk;
        Ok(usize::from(new_brk))
    } else {
        Ok(usize::from(brk))
    }
}

bitflags! {
    pub struct ProtFlags: u32 {
        const PROT_NONE = 0;
        const PROT_READ = 0x1;
        const PROT_WRITE = 0x2;
        const PROT_EXEC = 0x4;
    }
}

bitflags! {
    pub struct MapFlags: u32 {
        const MAP_SHARED = 0x1;
        const MAP_PRIVATE = 0x2;
        // MAP_FIXED 表示直接使用传递的参数做为 mmap 空间的起始地址
        const MAP_FIXED = 0x10;
        const MAP_ANONYMOUS = 0x20;
    }
}

/// 只能读取文件，不能写回文件
pub fn sys_mmap(
    start: usize,
    len: usize,
    prot: u32,
    flags: u32,
    fd: usize,
    off: i64,
) -> Result<usize, SyscallError> {
    let prot = ProtFlags::from_bits(prot).unwrap();
    let flags = MapFlags::from_bits(flags).unwrap();
    // println!(
    //     "in sys_mmap: start = {:#x}, len = {:#x}, prot = {:?}, flags = {:?}, fd = {}, off = {:#x}",
    //     start, len, prot, flags, fd as isize, off
    // );
    let mut permission = MapPermission::U;
    if prot.contains(ProtFlags::PROT_READ) {
        permission |= MapPermission::R;
    }
    if prot.contains(ProtFlags::PROT_WRITE) {
        permission |= MapPermission::W;
    }
    if prot.contains(ProtFlags::PROT_EXEC) {
        permission |= MapPermission::X;
    }

    let token = current_user_token();
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    // 分配 mmap 空间
    // 可能有的重用之前已经分配的空间，就造成 mmap 空间的分割
    let mut res = start;
    if flags.contains(MapFlags::MAP_FIXED) {
        task_inner
            .memory_set
            .insert_framed_area(start.into(), (start + len).into(), permission);
    } else {
        let va = task_inner
            .memory_set
            .insert_framed_area_randomly(len, None, permission)
            .0;
        task_inner.memory_set.inner_exclusive_access().mmap_bottom = va;
        res = va.into()
    }

    // 匿名映射不关联任何文件
    if flags.contains(MapFlags::MAP_ANONYMOUS) {
        Ok(res)
    } else {
        if let Some(file) = task_inner.fd_table.get(fd) {
            let vec = translated_byte_buffer(token, res as *const u8, len);
            file.lseek(off as isize, SEEK_SET).unwrap();
            file.read(UserBuffer::new(vec));
            return Ok(res);
        }
        Err(SyscallError::EBADF)
    }
}

pub fn sys_mprotect(addr: usize, len: usize, prot: u32) -> Result<usize, SyscallError> {
    let prot = ProtFlags::from_bits(prot).unwrap();
    let task = current_task().unwrap();
    let mut permission = MapPermission::U;
    if prot.contains(ProtFlags::PROT_READ) {
        permission |= MapPermission::R;
    }
    if prot.contains(ProtFlags::PROT_WRITE) {
        permission |= MapPermission::W;
    }
    if prot.contains(ProtFlags::PROT_EXEC) {
        permission |= MapPermission::X;
    }
    task.inner_exclusive_access().memory_set.insert_framed_area(
        addr.into(),
        (addr + len).into(),
        permission,
    );
    Ok(0)
}

pub fn sys_munmap(start: usize, _len: usize) -> Result<usize, SyscallError> {
    let task = current_task().unwrap();
    task.inner_exclusive_access()
        .memory_set
        .remove_area_with_start_vpn(start.into());
    Ok(0)
}

pub fn sys_shmget(key: i32, size: usize, shmflg: i32) -> Result<usize, SyscallError> {
    const IPC_PRIVATE: i32 = 0;
    bitflags! {
        struct ShmFlags: i32 {
            const IPC_CREAT = 0o1000;
            const IPC_EXCL = 0o2000;
        }
    }

    // 忽略权限位
    let flags = ShmFlags::from_bits(shmflg & !0x1ff).unwrap();
    match key {
        IPC_PRIVATE => Ok(shm_create(size)),
        key if key > 0 => {
            if shm_find(key as usize) {
                if flags.contains(ShmFlags::IPC_CREAT | ShmFlags::IPC_EXCL) {
                    Err(SyscallError::EEXIST)
                } else {
                    Ok(key as usize)
                }
            } else {
                if flags.contains(ShmFlags::IPC_CREAT) {
                    Ok(shm_create(size))
                } else {
                    Err(SyscallError::ENOENT)
                }
            }
        }
        _ => Err(SyscallError::ENOENT),
    }
}

pub fn sys_shmat(shmid: i32, shmaddr: usize, shmflg: i32) -> Result<usize, SyscallError> {
    bitflags! {
        struct ShmFlags: i32 {
            const SHM_RDONLY = 0o10000;
            // const SHM_RND    = 0o20000;
            // const SHM_REMAP  = 0o40000;
            const SHM_EXEC   = 0o100000;
        }
    }

    let mut permission = MapPermission::U | MapPermission::R;
    if shmflg == 0 {
        permission |= MapPermission::W | MapPermission::X
    } else {
        let shmflg = ShmFlags::from_bits(shmflg).unwrap();
        if shmflg.contains(ShmFlags::SHM_EXEC) {
            permission |= MapPermission::X;
        }
        if !shmflg.contains(ShmFlags::SHM_RDONLY) {
            permission |= MapPermission::W;
        }
    }

    match shmid {
        key if key < 0 => Err(SyscallError::EINVAL),
        _ => shm_attach(shmid as usize, shmaddr, permission),
    }
}

pub fn sys_shmctl(_shmid: i32, cmd: i32, _buf: usize) -> Result<usize, SyscallError> {
    const IPC_RMID: i32 = 0;
    match cmd {
        // 直接返回
        IPC_RMID => Ok(0),
        _ => {
            panic!("[sys_shmctl] unsupport cmd");
        }
    }
}
