pub mod errno;
mod fs;
mod mm;
mod process;
mod signal;

pub use fs::*;
pub use mm::*;
pub use process::*;
pub use signal::*;

use self::errno::SyscallError;
use crate::sbi::shutdown;
use crate::signal::{SignalAction, SignalFlags};
use crate::task::Rusage;
use crate::timer::{Itimerval, Timespec, Timeval, Tms};
use crate::utsname::*;
use crate::vfs::types::*;
use log::*;
use num_enum::TryFromPrimitive;

#[derive(Debug, PartialEq, Eq, Clone, Copy, TryFromPrimitive)]
#[repr(usize)]
pub enum SyscallID {
    Shutdown = 0,
    Getcwd = 17,
    Dup = 23,
    Dup3 = 24,
    Fcntl = 25,
    Ioctl = 29,
    Mkdirat = 34,
    Unlinkat = 35,
    Umount2 = 39,
    Mount = 40,
    Statfs = 43,
    Ftruncate = 46,
    Faccessat = 48,
    Chdir = 49,
    Openat = 56,
    Close = 57,
    Pipe2 = 59,
    Getdents64 = 61,
    Lseek = 62,
    Read = 63,
    Write = 64,
    Readv = 65,
    Writev = 66,
    Pread64 = 67,
    Sendfile = 71,
    Pselect6 = 72,
    Ppoll = 73,
    Readlinkat = 78,
    Newfstatat = 79,
    Fstat = 80,
    Sync = 81,
    Fsync = 82,
    Utimensat = 88,
    Exit = 93,
    ExitGroup = 94,
    SetTidAddress = 96,
    Nanosleep = 101,
    Setitimer = 103,
    ClockGettime = 113,
    Syslog = 116,
    SchedYield = 124,
    Kill = 129,
    RtSigaction = 134,
    RtSigprocmask = 135,
    RtSigtimedwait = 137,
    RtSigreturn = 139,
    Times = 153,
    Uname = 160,
    Getrusage = 165,
    Umask = 166,
    GetTimeOfDay = 169,
    Getpid = 172,
    Getppid = 173,
    Getuid = 174,
    Geteuid = 175,
    Getegid = 177,
    Gettid = 178,
    Sysinfo = 179,
    Shmget = 194,
    Shmctl = 195,
    Shmat = 196,
    Brk = 214,
    Munmap = 215,
    Clone = 220,
    Execve = 221,
    Mmap = 222,
    Mprotect = 226,
    Madvise = 233,
    Wait4 = 260,
    Prlimit = 261,
    Renameat2 = 276,
    Seccomp = 277,
    Getrandom = 278,
    MemfdCreate = 279,
    #[num_enum(default)]
    Unsupported,
}

/// `syscall_id` 只保证已经实现的系统调用的系统调用号是正确的；
/// 如果是未实现的系统调用，其系统调用号放置在 `real_id` 上，
/// 这时 `syscall_id` 只固定为最后一个系统调用号 + 1
pub fn syscall(
    syscall_id: SyscallID,
    args: [usize; 6],
    real_id: usize,
) -> Result<usize, errno::SyscallError> {
    debug!(
        "[syscall] {:?} id = {}, args = [{:#x}, {:#x}, {:#x}, {:#x}, {:#x}, {:#x}]",
        syscall_id as SyscallID, real_id, args[0], args[1], args[2], args[3], args[4], args[5],
    );

    match syscall_id {
        SyscallID::Getcwd => sys_getcwd(args[0] as *mut u8, args[1] as usize),
        SyscallID::Dup => sys_dup(args[0]),
        SyscallID::Dup3 => sys_dup3(args[0] as usize, args[1] as usize, args[2] as u32),
        SyscallID::Fcntl => sys_fcntl(args[0] as usize, args[1] as usize, args[2] as i32),
        SyscallID::Ioctl => unimplement(syscall_id),
        SyscallID::Mkdirat => sys_mkdirat(args[0] as isize, args[1] as *const u8, args[2] as u32),
        SyscallID::Unlinkat => sys_unlinkat(args[0] as isize, args[1] as *const u8, args[2] as u32),
        SyscallID::Umount2 => sys_umount2(args[0] as *const u8, args[1] as i32),
        SyscallID::Mount => sys_mount(
            args[0] as *const u8,
            args[1] as *const u8,
            args[2] as *const u8,
            args[3],
            args[4] as *const usize,
        ),
        SyscallID::Statfs => sys_statfs(args[0] as *const u8, args[1] as *mut Statfs),
        SyscallID::Ftruncate => sys_ftruncate(args[0] as i32, args[1] as i64),
        SyscallID::Faccessat => {
            sys_faccessat(args[0] as isize, args[1] as *const u8, args[2] as u32)
        }
        SyscallID::Chdir => sys_chdir(args[0] as *const u8),
        SyscallID::Openat => sys_openat(
            args[0] as isize,
            args[1] as *const u8,
            args[2] as u32,
            args[3] as u32,
        ),
        SyscallID::Close => sys_close(args[0]),
        SyscallID::Pipe2 => sys_pipe2(args[0] as *mut u32, args[1] as u32),
        SyscallID::Getdents64 => {
            sys_getdents64(args[0] as usize, args[1] as *const u8, args[2] as usize)
        }
        SyscallID::Lseek => sys_lseek(args[0] as usize, args[1] as isize, args[2] as usize),
        SyscallID::Read => sys_read(args[0], args[1] as *const u8, args[2]),
        SyscallID::Write => sys_write(args[0], args[1] as *const u8, args[2]),
        SyscallID::Readv => sys_readv(args[0], args[1] as *const IOVec, args[2]),
        SyscallID::Writev => sys_writev(args[0], args[1] as *const IOVec, args[2]),
        SyscallID::Pread64 => sys_pread64(args[0], args[1] as *mut u8, args[2], args[3]),
        SyscallID::Sendfile => sys_sendfile(
            args[0] as usize,
            args[1] as usize,
            args[2] as *mut i64,
            args[3] as usize,
        ),
        SyscallID::Pselect6 => sys_pselect6(
            args[0] as i32,
            args[1] as *mut FdSet,
            args[2] as *mut FdSet,
            args[3] as *mut FdSet,
            args[4] as *const Timespec,
            args[5] as *mut usize,
        ),
        SyscallID::Ppoll => sys_ppoll(
            args[0] as *mut Pollfd,
            args[1] as usize,
            args[2] as *const Timespec,
            args[3] as *const SignalFlags,
        ),
        SyscallID::Readlinkat => sys_readlinkat(
            args[0] as isize,
            args[1] as *const u8,
            args[2] as *const u8,
            args[3] as usize,
        ),
        SyscallID::Newfstatat => sys_newfstatat(
            args[0] as isize,
            args[1] as *const u8,
            args[2] as *mut Kstat,
            args[3] as usize,
        ),
        SyscallID::Fstat => sys_fstat(args[0] as usize, args[1] as *mut Kstat),
        SyscallID::Sync => unimplement(syscall_id),
        SyscallID::Fsync => unimplement(syscall_id),
        SyscallID::Utimensat => sys_utimensat(
            args[0] as isize,
            args[1] as *const u8,
            args[2] as *const [Timespec; 2],
            args[3] as u32,
        ),
        SyscallID::Exit => sys_exit(args[0] as i32),
        SyscallID::ExitGroup => sys_exit_group(args[0] as i32),
        SyscallID::SetTidAddress => sys_set_tid_address(args[0] as usize),
        SyscallID::Nanosleep => sys_nanosleep(args[0] as *mut Timespec, args[1] as *mut Timespec),
        SyscallID::Setitimer => sys_setitimer(
            args[0] as i32,
            args[1] as *const Itimerval,
            args[2] as *mut Itimerval,
        ),
        SyscallID::ClockGettime => sys_clock_gettime(args[0] as usize, args[1] as *mut Timespec),
        SyscallID::Syslog => unimplement(syscall_id),
        SyscallID::SchedYield => sys_sched_yield(),
        SyscallID::Kill => sys_kill(args[0] as isize, args[1]),
        SyscallID::RtSigaction => sys_rt_sigaction(
            args[0] as usize,
            args[1] as *const SignalAction,
            args[2] as *mut SignalAction,
            args[3] as usize,
        ),
        SyscallID::RtSigprocmask => sys_rt_sigprocmask(
            args[0] as u32,
            args[1] as *const SignalFlags,
            args[2] as *mut SignalFlags,
            args[3] as usize,
        ),
        SyscallID::RtSigtimedwait => sys_rt_sigtimedwait(
            args[0] as *const SignalFlags,
            args[1] as usize,
            args[2] as *const Timespec,
            args[3] as usize,
        ),
        SyscallID::RtSigreturn => sys_rt_sigretrun(),
        SyscallID::Times => sys_times(args[0] as *mut Tms),
        SyscallID::Uname => sys_uname(args[0] as *mut Utsname),
        SyscallID::Getrusage => sys_getrusage(args[0] as i32, args[1] as *mut Rusage),
        SyscallID::Umask => sys_umask(args[0] as usize),
        SyscallID::GetTimeOfDay => sys_get_time_of_day(args[0] as *mut Timeval),
        SyscallID::Brk => sys_brk(args[0]),
        SyscallID::Getpid => sys_getpid(),
        SyscallID::Getppid => sys_getppid(),
        SyscallID::Getuid => sys_getuid(),
        SyscallID::Geteuid => sys_geteuid(),
        SyscallID::Getegid => sys_getegid(),
        SyscallID::Gettid => sys_gettid(),
        SyscallID::Sysinfo => unimplement(syscall_id),
        SyscallID::Shmget => sys_shmget(args[0] as i32, args[1] as usize, args[2] as i32),
        SyscallID::Shmctl => sys_shmctl(args[0] as i32, args[1] as i32, args[2]),
        SyscallID::Shmat => sys_shmat(args[0] as i32, args[1] as usize, args[2] as i32),
        SyscallID::Clone => sys_clone(
            args[0] as u32,
            args[1] as usize,
            args[2] as *mut u32,
            args[3] as usize,
            args[4] as *mut u32,
        ),
        SyscallID::Execve => sys_execve(
            args[0] as *const u8,
            args[1] as *const usize,
            args[2] as *const usize,
        ),
        SyscallID::Wait4 => sys_wait4(
            args[0] as isize,
            args[1] as *mut i32,
            args[2] as u32,
            args[3] as *mut Rusage,
        ),
        SyscallID::Mmap => sys_mmap(
            args[0] as usize,
            args[1] as usize,
            args[2] as u32,
            args[3] as u32,
            args[4] as usize,
            args[5] as i64,
        ),
        SyscallID::Mprotect => sys_mprotect(args[0] as usize, args[1] as usize, args[2] as u32),
        SyscallID::Madvise => unimplement(syscall_id),
        SyscallID::Munmap => sys_munmap(args[0] as usize, args[1] as usize),
        SyscallID::Prlimit => sys_prlimit(
            args[0],
            args[1] as u32,
            args[2] as *const RLimit,
            args[3] as *mut RLimit,
        ),
        SyscallID::Renameat2 => sys_renameat2(
            args[0] as isize,
            args[1] as *const u8,
            args[2] as isize,
            args[3] as *const u8,
            args[4] as u32,
        ),
        SyscallID::Shutdown => shutdown(),
        SyscallID::Seccomp => unimplement(syscall_id),
        SyscallID::Getrandom => sys_getrandom(args[0] as *mut u8, args[1], args[2] as u32),
        SyscallID::MemfdCreate => sys_memfd_create(args[0] as *const u8, args[1] as u32),
        SyscallID::Unsupported => {
            warn!("[Syscall] Unsupported syscall_id: {}", real_id as usize);
            sys_exit_group(0)
        }
    }
}

fn unimplement(syscall_id: SyscallID) -> Result<usize, SyscallError> {
    warn!(
        "[Syscall] Unimplement: {:?}, id = {}, default return 0",
        syscall_id, syscall_id as usize
    );
    Ok(0)
}
