use super::errno::*;
use crate::mm::{
    put_data, translated_byte_buffer, translated_ref, translated_refmut, translated_str,
    UserBuffer, VirtAddr,
};
use crate::task::{
    current_task, current_user_token, exit_current_and_run_next, exit_current_group,
    move_child_process_to_init, remove_from_thread_group, suspend_current_and_run_next, Rusage,
    TaskControlBlock, PIDTYPE_PPID, PID_HASH,
};
use crate::timer::*;
use crate::utsname::*;
use crate::vfs::{open_file_at, types::AT_FDCWD, ModeFlags, OpenFlags};
use alloc::string::String;
use alloc::sync::Arc;
use alloc::vec::Vec;
use core::mem::size_of;
use log::*;

pub fn sys_times(buf: *mut Tms) -> Result<usize, SyscallError> {
    let token = current_user_token();
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    let tms = Tms {
        tms_utime: task_inner.rusage.ru_utime.to_tick(),
        tms_stime: task_inner.rusage.ru_stime.to_tick(),
        tms_cutime: 0,
        tms_cstime: 0,
    };
    put_data(token, buf, tms);
    Ok(get_time())
}

pub fn sys_get_time_of_day(time: *mut Timeval) -> Result<usize, SyscallError> {
    let token = current_user_token();
    (*translated_refmut(token, time)) = get_timeval();
    Ok(0)
}

pub fn sys_clock_gettime(_clk_id: usize, tp: *mut Timespec) -> Result<usize, SyscallError> {
    let token = current_user_token();
    (*translated_refmut(token, tp)) = get_timespec();
    Ok(0)
}

pub fn sys_nanosleep(req: *const Timespec, _rem: *mut Timespec) -> Result<usize, SyscallError> {
    let token = current_user_token();
    let end_timespec = get_timespec() + (*translated_ref(token, req));
    loop {
        let cur_timespec = get_timespec();
        if cur_timespec >= end_timespec {
            return Ok(0);
        } else {
            suspend_current_and_run_next();
        }
    }
}

pub fn sys_setitimer(
    which: i32,
    new_value: *const Itimerval,
    old_value: *mut Itimerval,
) -> Result<usize, SyscallError> {
    let token = current_user_token();
    let new_value = translated_ref(token, new_value);
    debug!(
        "[sys_setitimer] which = {}, new_value = {:?}",
        which, new_value
    );
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    match which {
        id if id == ITIMER_REAL as i32 => {
            task_inner.timer.inner_exclusive_access().itimer[ITIMER_REAL] = *new_value;
            if !old_value.is_null() {
                put_data(
                    token,
                    old_value,
                    task_inner.timer.inner_exclusive_access().itimer[ITIMER_REAL],
                );
            }
        }
        _ => {
            panic!("[sys_setitimer] unsupport arg, which = {}", which)
        }
    }
    Ok(0)
}

pub fn sys_uname(buf: *mut Utsname) -> Result<usize, SyscallError> {
    let token = current_user_token();
    let mut buffer = UserBuffer::new(translated_byte_buffer(
        token,
        buf as usize as *const u8,
        size_of::<Utsname>(),
    ));
    let uname = Utsname::new();
    buffer.write(uname.as_bytes());
    Ok(0)
}

pub fn sys_gettid() -> Result<usize, SyscallError> {
    Ok(current_task().unwrap().get_tid())
}

pub fn sys_getpid() -> Result<usize, SyscallError> {
    Ok(current_task().unwrap().get_pid())
}

pub fn sys_getppid() -> Result<usize, SyscallError> {
    Ok(current_task().unwrap().get_ppid())
}

pub fn sys_getuid() -> Result<usize, SyscallError> {
    Ok(0)
}

pub fn sys_geteuid() -> Result<usize,SyscallError> {
    Ok(0)
}

pub fn sys_getegid() -> Result<usize,SyscallError> {
    Ok(0)
}
pub fn sys_exit(exit_code: i32) -> ! {
    exit_current_and_run_next(exit_code);
    panic!("Unreachable in sys_exit!");
}

pub fn sys_exit_group(exit_code: i32) -> ! {
    exit_current_group(exit_code);
    panic!("Unreachable in sys_exit!");
}

pub fn sys_sched_yield() -> Result<usize, SyscallError> {
    suspend_current_and_run_next();
    Ok(0)
}

pub fn sys_prlimit(
    pid: usize,
    resource: u32,
    new_limit: *const RLimit,
    old_limit: *mut RLimit,
) -> Result<usize, SyscallError> {
    const RLIMIT_NOFILE: u32 = 7;
    if resource != RLIMIT_NOFILE {
        return Ok(0);
    }

    if pid == 0 {
        let token = current_user_token();
        let task = current_task().unwrap();
        let mut inner = task.inner_exclusive_access();
        let fd_table = &mut inner.fd_table;
        if !old_limit.is_null() {
            // 说明是get
            let limit = translated_refmut(token, old_limit);
            limit.rlim_cur = fd_table.get_soft_limit();
            limit.rlim_max = fd_table.get_hard_limit();
        }
        if !new_limit.is_null() {
            // 说明是set
            let limit: &RLimit = translated_ref(token, new_limit);
            fd_table.set_limit(limit.rlim_cur, limit.rlim_max);
        }

        drop(inner);
    } else {
        unimplemented!("未实现的非0 pid");
    }

    Ok(0)
}

#[allow(unused)]
#[derive(Clone, Copy, Debug)]
pub struct RLimit {
    pub rlim_cur: usize, /* Soft limit */
    pub rlim_max: usize, /* Hard limit (ceiling for rlim_cur) */
}

bitflags! {
    pub struct CloneFlags: u32 {
        const CSIGNAL = 0x000000ff;
        // const CLONE_NEWTIME= 0x00000080;
        const CLONE_VM = 0x00000100;
        const CLONE_FS = 0x00000200;
        const CLONE_FILES = 0x00000400;
        const CLONE_SIGHAND = 0x00000800;
        // const CLONE_PIDFD = 0x00001000;
        // const CLONE_PTRACE = 0x00002000;
        // const CLONE_VFORK =0x00004000;
        // const CLONE_PARENT = 0x00008000;
        const CLONE_THREAD = 0x00010000;
        // const CLONE_NEWNS = 0x00020000;
        const CLONE_SYSVSEM = 0x00040000;
        const CLONE_SETTLS = 0x00080000;
        const CLONE_PARENT_SETTID = 0x00100000;
        const CLONE_CHILD_CLEARTID = 0x00200000;
        const CLONE_DETACHED = 0x00400000;
        // const CLONE_UNTRACED = 0x00800000;
        const CLONE_CHILD_SETTID = 0x01000000;
        // const CLONE_NEWCGROUP = 0x02000000;
        // const CLONE_NEWUTS = 0x04000000;
        // const CLONE_NEWIPC = 0x08000000;
        // const CLONE_NEWUSER = 0x10000000;
        // const CLONE_NEWPID = 0x20000000;
        // const CLONE_NEWNET = 0x40000000;
        // const CLONE_IO = 0x80000000;
    }
}

pub fn sys_clone(
    flags: u32,
    stack: usize,
    parent_tid: *mut u32,
    tls: usize,
    child_tid: *mut u32,
) -> Result<usize, SyscallError> {
    let flags = CloneFlags::from_bits(flags).unwrap();
    debug!("[sys_clone] flags = {:#?}", flags);
    let current_process = current_task().unwrap();
    let stack = if stack == 0 {
        None
    } else {
        Some(VirtAddr::from(stack))
    };
    let new_process =
        TaskControlBlock::clone(&current_process, flags, stack, parent_tid, tls, child_tid);
    let new_tid = new_process.get_tid();
    // modify trap context of new_process, because it returns immediately after switching
    let new_process_inner = new_process.inner_exclusive_access();
    let trap_cx = new_process_inner.get_trap_cx();
    // we do not have to move to next instruction since we have done it before
    // for child process, fork returns 0
    trap_cx.x[10] = 0;
    Ok(new_tid)
}

pub fn sys_execve(
    path: *const u8,
    mut args: *const usize,
    mut envp: *const usize,
) -> Result<usize, SyscallError> {
    let token = current_user_token();
    let mut path = translated_str(token, path);
    let mut args_vec: Vec<String> = Vec::new();
    if !args.is_null() {
        loop {
            let arg_str_ptr = *translated_ref(token, args);
            if arg_str_ptr == 0 {
                break;
            }
            args_vec.push(translated_str(token, arg_str_ptr as *const u8));
            unsafe {
                args = args.add(1);
            }
        }
    }
    let mut envp_vec: Vec<String> = Vec::new();
    if !envp.is_null() {
        loop {
            let env_str_ptr = *translated_ref(token, envp);
            if env_str_ptr == 0 {
                break;
            }
            envp_vec.push(translated_str(token, env_str_ptr as *const u8));
            unsafe {
                envp = envp.add(1);
            }
        }
    }
    if path.ends_with(".sh") {
        path = String::from("busybox");
        args_vec.insert(0, String::from("sh"));
        args_vec.insert(0, String::from("busybox"));
    }
    debug!(
        "[sys_execve] app name = {}, argv = {:?}, envp = {:?}",
        path.as_str(),
        args_vec,
        envp_vec
    );
    open_file_at(AT_FDCWD, &path, OpenFlags::O_RDONLY, ModeFlags::empty()).map(|file| {
        let all_data = file.read_all();
        let process = current_task().unwrap();
        process.execve(all_data.as_slice(), args_vec, envp_vec);
        let mut process_inner = process.inner_exclusive_access();
        process_inner.exec = Some(file.get_inode());
        process_inner.last_utime = get_timeval();
        process_inner.last_stime = get_timeval();
        0
    })
}

bitflags! {
    pub struct WaitFlags: u32 {
        const WNOHANG = 0x00000001;
        // const WUNTRACED = 0x00000002;
        // const WCONTINUED = 0x00000008;
    }
}

pub fn sys_wait4(
    pid: isize,
    wstatus: *mut i32,
    flags: u32,
    rusage: *mut Rusage,
) -> Result<usize, SyscallError> {
    let flags = WaitFlags::from_bits(flags).unwrap();
    loop {
        let mut hash = PID_HASH[PIDTYPE_PPID].lock();
        let task = current_task().unwrap();
        let task_inner = task.inner_exclusive_access();

        let children = hash.get_mut(&task.get_pid()).unwrap();
        let mut wait_child_id = None;
        let mut exit_child = None;
        for (id, child) in children.iter().enumerate() {
            // 检查子线程组是否退出
            let child_inner = child.inner_exclusive_access();
            if child_inner
                .signal_info
                .inner_exclusive_access()
                .group_exit_code
                .is_some()
            {
                exit_child = Some(Arc::clone(child));
            }
            drop(child_inner);

            // 检查子线程组是否满足要求
            match pid {
                _ if pid > 0 => {
                    if child.get_pid() == pid as usize {
                        wait_child_id = Some(id);
                        break;
                    }
                }
                _ => {
                    wait_child_id = Some(id);
                    if exit_child.is_some() {
                        break;
                    }
                }
            }
        }

        if let Some(id) = wait_child_id {
            if let Some(child) = exit_child {
                let pid = child.get_pid();
                let child_inner = child.inner_exclusive_access();
                let exic_code = child_inner
                    .signal_info
                    .inner_exclusive_access()
                    .group_exit_code
                    .unwrap();
                // C 库中 #define WEXITSTATUS(s) ((s & 0xFF00) >> 8)
                // exit_code 需要左移 8 位
                if wstatus as usize != 0 {
                    *translated_refmut(task_inner.get_user_token(), wstatus) = exic_code << 8;
                }
                if !rusage.is_null() {
                    put_data(
                        task_inner.get_user_token(),
                        rusage,
                        Rusage::clone(&child_inner.rusage),
                    );
                }
                drop(child_inner);

                children.remove(id);
                remove_from_thread_group(pid);
                // 调用 wait4() 的可能是 init 进程，因此需要释放进程锁
                drop(task_inner);
                drop(task);
                drop(hash);
                move_child_process_to_init(pid);

                // 此时 TCB 引用应该只剩一个，可以验证一下
                assert_eq!(Arc::strong_count(&child), 1);

                return Ok(pid);
            } else if flags.contains(WaitFlags::WNOHANG) {
                return Ok(0);
            } else {
                drop(task_inner);
                drop(task);
                drop(hash);
                suspend_current_and_run_next();
            }
        } else {
            return Err(SyscallError::ECHILD);
        }
    }
}

pub fn sys_set_tid_address(ptr: usize) -> Result<usize, SyscallError> {
    let task = current_task().unwrap();
    let mut task_inner = task.inner_exclusive_access();
    task_inner.clear_child_tid = Some(ptr.into());
    Ok(task.get_tid())
}

pub fn sys_getrandom(buf: *mut u8, buflen: usize, _flags: u32) -> Result<usize, SyscallError> {
    let token = current_user_token();
    let tick = get_time();
    let len = buflen.min(core::mem::size_of::<usize>());
    let mut vec = translated_byte_buffer(token, buf, len);
    assert_eq!(len, vec[0].len());
    vec[0].copy_from_slice(&tick.to_be_bytes()[..len]);
    Ok(len)
}

pub fn sys_getrusage(who: i32, usage: *mut Rusage) -> Result<usize, SyscallError> {
    const RUSAGE_SELF: i32 = 0;

    let token = current_user_token();
    let task = current_task().unwrap();
    match who {
        RUSAGE_SELF => {
            let inner = task.inner_exclusive_access();
            put_data(token, usage, Rusage::clone(&inner.rusage))
        }
        _ => {
            panic!("[sys_getrusage] unsupport arg, who = {}", who);
        }
    }
    Ok(0)
}
