use super::errno::*;
use crate::{
    mm::{translated_ref, translated_refmut},
    signal::{
        restore_frame, send_access_signal, send_process_group_singal, send_thread_group_signal,
        SignalAction, SignalFlags, MAX_SIG,
    },
    task::{current_task, current_user_token, suspend_current_and_run_next},
    timer::{get_timespec, Timespec},
};
use log::*;

pub fn sys_kill(pid: isize, sig: usize) -> Result<usize, SyscallError> {
    if sig != 0 {
        let signal = SignalFlags::from_sig(sig);
        match pid {
            _ if pid > 0 => {
                send_thread_group_signal(pid as usize, signal);
            }
            0 => {
                let pgid = current_task().unwrap().get_pid();
                send_process_group_singal(pgid, signal);
            }
            -1 => {
                send_access_signal(signal);
            }
            _ => {
                send_process_group_singal(-pid as usize, signal);
            }
        };
    }
    Ok(0)
}

pub fn sys_rt_sigprocmask(
    how: u32,
    set: *const SignalFlags,
    old_set: *mut SignalFlags,
    _sig_set_size: usize,
) -> Result<usize, SyscallError> {
    const SIG_BLOCK: u32 = 0;
    const SIG_UNBLOCK: u32 = 1;
    const SIG_SETMASK: u32 = 2;
    let task = current_task().unwrap();
    let mut task_inner = task.inner_exclusive_access();
    let token = task_inner.get_user_token();
    if old_set as usize != 0 {
        *translated_refmut(token, old_set) = task_inner.signal_mask;
    }
    if set as usize != 0 {
        let mut mask = *translated_ref(token, set);
        mask &= !SignalFlags::SIGKILL;
        mask &= !SignalFlags::SIGSTOP;
        trace!("[sys_rt_sigprocmask] set = {:?}, how = {:?}", mask, how);
        match how {
            SIG_BLOCK => task_inner.signal_mask |= mask,
            SIG_UNBLOCK => task_inner.signal_mask &= !mask,
            SIG_SETMASK => task_inner.signal_mask = mask,
            _ => return Err(SyscallError::EINVAL),
        }
    }
    Ok(0)
}

pub fn sys_rt_sigretrun() -> Result<usize, SyscallError> {
    restore_frame();
    Ok(0)
}

pub fn sys_rt_sigaction(
    signum: usize,
    action: *const SignalAction,
    old_action: *mut SignalAction,
    _sig_set_size: usize,
) -> Result<usize, SyscallError> {
    let signal = SignalFlags::from_sig(signum);
    debug!("[sys_rt_sigaction] signal = {:?}", signal);
    if signal == SignalFlags::SIGKILL || signal == SignalFlags::SIGSTOP {
        return Err(SyscallError::EINVAL);
    }
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    let token = task_inner.get_user_token();
    let mut signal_actions_inner = task_inner.signal_actions.inner_exclusive_access();
    if old_action as usize != 0 {
        *translated_refmut(token, old_action) = signal_actions_inner.table[signum];
    }
    if action as usize != 0 {
        debug!(
            "[sys_rt_sigaction] signal action = {:#x?}",
            *translated_ref(token, action)
        );
        signal_actions_inner.table[signum] = *translated_ref(token, action);
    }
    Ok(0)
}

pub fn sys_rt_sigtimedwait(
    set: *const SignalFlags,
    _info: usize,
    timeout: *const Timespec,
    _sig_set_size: usize,
) -> Result<usize, SyscallError> {
    let token = current_user_token();
    let set = *translated_ref(token, set);
    let timeout = *translated_ref(token, timeout);
    trace!(
        "[sys_rt_sigtimedwait] set = {:?}, timeout = {:?}, info = {:?}",
        set,
        timeout,
        _info
    );
    let end_time = timeout + get_timespec();
    loop {
        let task = current_task().unwrap();
        let mut task_inner = task.inner_exclusive_access();
        for signum in 1..(MAX_SIG + 1) {
            let signal = SignalFlags::from_sig(signum);
            if set.contains(signal) && task_inner.signal_pending.contains(signal) {
                task_inner.signal_pending &= !signal;
                return Ok(signum);
            }
        }
        if get_timespec() > end_time {
            return Err(SyscallError::EAGAIN);
        } else {
            drop(task_inner);
            drop(task);
            suspend_current_and_run_next();
        }
    }
}
