pub mod aux;
mod context;
mod id;
mod manager;
mod processor;
mod switch;
#[allow(clippy::module_inception)]
mod task;

use crate::mm::translated_refmut;
use crate::signal::{send_thread_group_signal, SignalFlags};
use crate::timer::get_timeval;
use crate::vfs::{open_file, root_inode, ModeFlags, OpenFlags};
use alloc::sync::Arc;
pub use context::TaskContext;
use lazy_static::*;
use manager::fetch_task;
use switch::__switch;
pub use task::{Rusage, TaskControlBlock, TaskStatus};

pub use id::*;
pub use manager::add_task;
pub use processor::{
    current_task, current_trap_cx, current_trap_cx_user_va, current_user_token, run_tasks,
    schedule, take_current_task,
};

use self::manager::remove_task;

lazy_static! {
    pub static ref INITPROC: Arc<TaskControlBlock> = {
        let file = open_file(
            root_inode(),
            "initproc",
            OpenFlags::O_RDONLY,
            ModeFlags::empty(),
        )
        .unwrap();
        let v = file.read_all();
        TaskControlBlock::new(v.as_slice())
    };
}

pub fn add_initproc() {
    // 由于使用了懒分配，因此显示使用 INITPROC 生成 init 进程
    let _initproc = Arc::clone(&INITPROC);
}

pub fn suspend_current_and_run_next() {
    // There must be an application running.
    let task = take_current_task().unwrap();

    // ---- access current TCB exclusively
    let mut task_inner = task.inner_exclusive_access();
    let task_cx_ptr = &mut task_inner.task_cx as *mut TaskContext;
    // Change status to Ready
    task_inner.task_status = TaskStatus::Ready;
    task_inner.rusage.ru_stime = task_inner.rusage.ru_stime + get_timeval() - task_inner.last_stime;
    drop(task_inner);
    // ---- release current PCB

    // push back to ready queue.
    add_task(task);
    // jump to scheduling cycle
    schedule(task_cx_ptr);
}

pub fn stop_current_and_run_next() {
    let task = take_current_task().unwrap();
    let mut task_inner = task.inner_exclusive_access();
    let task_cx_ptr = &mut task_inner.task_cx as *mut TaskContext;
    task_inner.task_status = TaskStatus::Stopped;
    task_inner.rusage.ru_stime = task_inner.rusage.ru_stime + get_timeval() - task_inner.last_stime;
    drop(task_inner);
    schedule(task_cx_ptr);
}

/// 退出当前进程
///
/// 此时只是将进程从调度队列中移除，`PID_HASH` 还保留进程的 TCB，
/// 该 TCB 将等待 `wait4()` 系统调用释放
pub fn exit_current_and_run_next(exit_code: i32) {
    let task = take_current_task().unwrap();
    let mut task_inner = task.inner_exclusive_access();
    // CLONE_CHILD_CLEARTID
    if let Some(ptr) = task_inner.clear_child_tid {
        let token = task_inner.get_user_token();
        *translated_refmut(token, usize::from(ptr) as *mut u32) = 0;
    }
    // 释放用户资源
    task_inner.dealloc_user_res();
    task_inner.exec = None;
    task_inner.exit_code = Some(exit_code);
    task_inner.task_status = TaskStatus::Zombie;
    drop(task_inner);

    // 通知父进程
    {
        let hash = PID_HASH[PIDTYPE_TGID].lock();
        if let Some(vec) = hash.get(&task.get_pid()) {
            if vec
                .iter()
                .all(|t| t.inner_exclusive_access().task_status == TaskStatus::Zombie)
            {
                drop(hash);
                send_thread_group_signal(task.get_ppid(), SignalFlags::SIGCHLD);
                // 释放资源
                let task_inner = task.inner_exclusive_access();
                let mut sig_info_inner = task_inner.signal_info.inner_exclusive_access();
                if sig_info_inner.group_exit_code.is_none() {
                    sig_info_inner.group_exit_code = Some(exit_code);
                }
                task_inner.memory_set.recycle_data_pages();
                task_inner.fd_table.clear();
            }
        }
    }

    remove_task(task);
    // we do not have to save task context
    let mut _unused = TaskContext::zero_init();
    schedule(&mut _unused as *mut _);
}

pub fn exit_current_group(exit_code: i32) {
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    let mut sig_info_inner = task_inner.signal_info.inner_exclusive_access();
    if sig_info_inner.group_exit_code.is_none() {
        sig_info_inner.group_exit_code = Some(exit_code);
        let pid = task.get_pid();
        drop(sig_info_inner);
        drop(task_inner);
        drop(task);
        send_thread_group_signal(pid, SignalFlags::SIGKILL);
    } else {
        drop(sig_info_inner);
        drop(task_inner);
        drop(task);
    }
    exit_current_and_run_next(exit_code);
}
