#![allow(unused)]

use crate::config::CLOCK_FREQ;
use crate::sbi::set_timer;
use core::ops::{Add, Sub};
use riscv::register::time;
use spin::{Mutex, MutexGuard};

const TICKS_PER_SEC: usize = 100;
const MSEC_PER_SEC: usize = 1000;
const USEC_PER_SEC: usize = 1000000;
const NSEC_PER_SEC: usize = 1000000000;

#[repr(C)]
#[derive(Debug, Clone, Copy, PartialEq, PartialOrd)]
pub struct Timeval {
    pub tv_sec: usize,
    pub tv_usec: usize,
}
impl Timeval {
    pub fn new() -> Self {
        Self {
            tv_sec: 0,
            tv_usec: 0,
        }
    }
    pub fn from_tick(tick: usize) -> Self {
        Self {
            tv_sec: tick / CLOCK_FREQ,
            tv_usec: (tick % CLOCK_FREQ) * USEC_PER_SEC / CLOCK_FREQ,
        }
    }
    pub fn to_tick(&self) -> usize {
        self.tv_sec * CLOCK_FREQ + (self.tv_usec * CLOCK_FREQ / USEC_PER_SEC)
    }
}
impl Add for Timeval {
    type Output = Timeval;

    fn add(self, rhs: Self) -> Self::Output {
        Self::Output::from_tick(self.to_tick() + rhs.to_tick())
    }
}
impl Sub for Timeval {
    type Output = Timeval;

    fn sub(self, rhs: Self) -> Self::Output {
        Self::Output::from_tick(self.to_tick() - rhs.to_tick())
    }
}

#[repr(C)]
#[derive(Clone, Copy, Debug, PartialEq, PartialOrd)]
pub struct Timespec {
    pub tv_sec: usize,
    pub tv_nsec: usize,
}
impl Timespec {
    pub fn from_tick(tick: usize) -> Self {
        Self {
            tv_sec: tick / CLOCK_FREQ,
            tv_nsec: (tick % CLOCK_FREQ) * NSEC_PER_SEC / CLOCK_FREQ,
        }
    }
    pub fn to_tick(&self) -> usize {
        self.tv_sec * CLOCK_FREQ + (self.tv_nsec * CLOCK_FREQ / NSEC_PER_SEC)
    }

    pub fn now() -> Self {
        get_timespec()
    }

    pub fn now_sec() -> usize {
        get_timespec().tv_sec
    }
}
impl Add for Timespec {
    type Output = Timespec;

    fn add(self, other: Timespec) -> Timespec {
        Timespec::from_tick(self.to_tick() + other.to_tick())
    }
}

#[repr(C)]
#[derive(Debug, Clone)]
pub struct Tms {
    // 时间单位为滴答数，即 tick
    pub tms_utime: usize,  /* 用户使用的 CPU 时间 */
    pub tms_stime: usize,  /* 系统（内核）CPU 时间 */
    pub tms_cutime: usize, /* 已终止的子进程使用的用户 CPU 时间 */
    pub tms_cstime: usize, /* 已终止的子进程使用的系统 CPU 时间 */
}

#[repr(C)]
#[derive(Debug, Clone, Copy)]
pub struct Itimerval {
    pub it_interval: Timeval,
    pub it_value: Timeval,
}

impl Itimerval {
    pub fn new() -> Self {
        Self {
            it_interval: Timeval::new(),
            it_value: Timeval::new(),
        }
    }
}

pub const ITIMER_REAL: usize = 0;
pub const ITIMER_VIRTUAL: usize = 1;
pub const ITIMER_PROF: usize = 2;

/// 三种 itimer
pub struct Timer {
    inner: Mutex<TimerInner>,
}
pub struct TimerInner {
    pub itimer: [Itimerval; 3],
}

impl Timer {
    pub fn inner_exclusive_access(&self) -> MutexGuard<'_, TimerInner> {
        self.inner.lock()
    }

    pub fn new() -> Self {
        Self {
            inner: Mutex::new(TimerInner {
                itimer: [Itimerval::new(); 3],
            }),
        }
    }

    /// 检查 `id` 类型的定时器是否到达，如果到达则重置
    pub fn check_timer(&self, id: usize) -> bool {
        let mut inner = self.inner_exclusive_access();
        let itimer = &mut inner.itimer[id];
        let empty = Timeval::new();
        if itimer.it_value != empty && itimer.it_value <= get_interval_timeval() {
            itimer.it_value = if itimer.it_interval != empty {
                itimer.it_interval
            } else {
                empty
            };
            true
        } else {
            itimer.it_value = itimer.it_value - get_interval_timeval();
            false
        }
    }
}

pub fn get_time() -> usize {
    time::read()
}

pub fn get_timeval() -> Timeval {
    let tick = get_time();
    Timeval::from_tick(tick)
}

pub fn get_timespec() -> Timespec {
    let tick = get_time();
    Timespec::from_tick(tick)
}

/// 获取设置时钟中断的时间间隔，并将其转换为 `Timeval` 类型
pub fn get_interval_timeval() -> Timeval {
    let tick = CLOCK_FREQ / TICKS_PER_SEC;
    Timeval::from_tick(tick)
}

pub fn set_next_trigger() {
    set_timer(get_time() + CLOCK_FREQ / TICKS_PER_SEC);
}
