mod context;

use crate::config::*;
use crate::mm::MapPermission;
use crate::signal::{current_add_signal, handle_signals, SignalFlags};
use crate::syscall::{syscall, SyscallID};
use crate::task::{
    current_task, current_trap_cx, current_trap_cx_user_va, current_user_token,
    suspend_current_and_run_next,
};
use crate::timer::*;
use alloc::format;
use core::arch::{asm, global_asm};
use core::convert::TryFrom;
use log::*;
use riscv::register::{
    mtvec::TrapMode,
    scause::{self, Exception, Interrupt, Trap},
    sie, stval, stvec,
};

pub use context::TrapContext;

global_asm!(include_str!("trap.S"));

pub fn init() {
    set_kernel_trap_entry();
}

fn set_kernel_trap_entry() {
    unsafe {
        stvec::write(trap_from_kernel as usize, TrapMode::Direct);
    }
}

fn set_user_trap_entry() {
    unsafe {
        stvec::write(TRAMPOLINE as usize, TrapMode::Direct);
    }
}

pub fn enable_timer_interrupt() {
    unsafe {
        sie::set_stimer();
    }
}

#[no_mangle]
pub fn trap_handler() -> ! {
    // println!("trap handler");
    set_kernel_trap_entry();
    {
        let task = current_task().unwrap();
        let mut task_inner = task.inner_exclusive_access();
        task_inner.rusage.ru_utime =
            task_inner.rusage.ru_utime + get_timeval() - task_inner.last_utime;
        task_inner.last_stime = get_timeval();
    }
    let scause = scause::read();
    let stval = stval::read();
    match scause.cause() {
        Trap::Exception(Exception::UserEnvCall) => {
            // jump to next instruction anyway
            let mut cx = current_trap_cx();
            trace!("[trap_handler] UserEnvCall sepc = {:#x}", cx.sepc);
            let syscall_id = SyscallID::try_from(cx.x[17]).unwrap();
            cx.sepc += 4;
            // get system call return value
            let result = syscall(
                syscall_id,
                [cx.x[10], cx.x[11], cx.x[12], cx.x[13], cx.x[14], cx.x[15]],
                cx.x[17],
            );

            match result {
                Ok(ret) => debug!(
                    "[syscall ret] {:?} ret = {}",
                    syscall_id,
                    format!("Ok({:#x})", ret)
                ),
                Err(errno) => warn!(
                    "[syscall ret] {:?} ret = {}",
                    syscall_id,
                    format!("Err({:?})", errno)
                ),
            }

            // rt_sigreturn 直接返回之前被信号处理中断的系统调用返回值
            if syscall_id != SyscallID::RtSigreturn {
                // cx is changed during sys_exec, so we have to call it again
                cx = current_trap_cx();
                cx.x[10] = match result {
                    Ok(res) => res,
                    Err(errno) => (-(errno as isize)) as usize,
                }
            }
        }
        Trap::Exception(Exception::StorePageFault) | Trap::Exception(Exception::LoadPageFault) => {
            // 增加用户栈大小
            if stval > MMAP_TOP && stval <= USTACK_TOP {
                debug!(
                    "[trap_handler] {:?} in application, bad addr = {:#x}, bad instruction = {:#x}, kernel killed it.",
                    scause.cause(),
                    stval,
                    current_trap_cx().sepc,
                );
                debug!("[trap_handler] increase application stack");
                let task = current_task().unwrap();
                let mut task_inner = task.inner_exclusive_access();
                if let Some(ustack_top) = task_inner.ustack_top {
                    debug!(
                        "[trap_handler] ustack_top = {:#x?}, ustack_size = {:#x}",
                        ustack_top, task_inner.ustack_size
                    );
                    task_inner.memory_set.insert_framed_area(
                        ustack_top - (task_inner.ustack_size + USER_STACK_SIZE),
                        ustack_top - task_inner.ustack_size,
                        MapPermission::R | MapPermission::W | MapPermission::U,
                    );
                    task_inner.ustack_size += USER_STACK_SIZE;
                }
            } else {
                error!(
                    "[kernel] {:?} in application, bad addr = {:#x}, bad instruction = {:#x}, kernel killed it.",
                    scause.cause(),
                    stval,
                    current_trap_cx().sepc,
                );
                current_add_signal(SignalFlags::SIGSEGV);
            }
        }
        Trap::Exception(Exception::StoreFault)
        | Trap::Exception(Exception::InstructionFault)
        | Trap::Exception(Exception::InstructionPageFault)
        | Trap::Exception(Exception::LoadFault) => {
            error!(
                "[kernel] {:?} in application, bad addr = {:#x}, bad instruction = {:#x}, kernel killed it.",
                scause.cause(),
                stval,
                current_trap_cx().sepc,
            );

            current_add_signal(SignalFlags::SIGSEGV);
        }
        Trap::Exception(Exception::IllegalInstruction) => {
            println!(
                "[kernel] {:?} in application, bad addr = {:#x}, bad instruction = {:#x}, kernel killed it.",
                scause.cause(),
                stval,
                current_trap_cx().sepc,
            );
            current_add_signal(SignalFlags::SIGILL);
        }
        Trap::Interrupt(Interrupt::SupervisorTimer) => {
            set_next_trigger();
            suspend_current_and_run_next();

            // 检查定时器
            let task = current_task().unwrap();
            let task_inner = task.inner_exclusive_access();
            if task_inner.timer.check_timer(ITIMER_REAL) {
                drop(task_inner);
                drop(task);
                current_add_signal(SignalFlags::SIGALRM);
            }
        }
        _ => {
            panic!(
                "Unsupported trap {:?}, stval = {:#x}!",
                scause.cause(),
                stval
            );
        }
    }
    // handle signals (handle the sent signal)
    // println!("[K] trap_handler:: handle_signals");
    handle_signals();
    // println!("[K] trap_handler:: handle_signals complete");
    trap_return();
}

#[no_mangle]
pub fn trap_return() -> ! {
    {
        let task = current_task().unwrap();
        let mut task_inner = task.inner_exclusive_access();
        task_inner.rusage.ru_stime =
            task_inner.rusage.ru_stime + get_timeval() - task_inner.last_stime;
        task_inner.last_utime = get_timeval();
    }
    set_user_trap_entry();
    let trap_cx_ptr: usize = current_trap_cx_user_va().into();
    let user_satp = current_user_token();
    extern "C" {
        fn __alltraps();
        fn __restore();
    }
    let restore_va = __restore as usize - __alltraps as usize + TRAMPOLINE;
    // println!("trap return");
    unsafe {
        asm!(
            "fence.i",
            "jr {restore_va}",
            restore_va = in(reg) restore_va,
            in("a0") trap_cx_ptr,
            in("a1") user_satp,
            options(noreturn)
        );
    }
}

#[no_mangle]
pub fn trap_from_kernel() -> ! {
    use riscv::register::sepc;
    println!("stval = {:#x}, sepc = {:#x}", stval::read(), sepc::read());
    panic!("a trap {:?} from kernel!", scause::read().cause());
}
