use super::{Null, Zero};
use crate::vfs::{types::*, File, Inode, Memfile, ModeFlags};
use alloc::{string::String, sync::Arc};
use lazy_static::lazy_static;
use spin::{Mutex, MutexGuard};

/// devtmpfs 文件系统根目录
///
/// 根目录下虽然存储了目录项，但读取的时候并没有获取对应的 inode 节点
/// （甚至没有 inode 节点），都是直接获取得到对应的文件。因此没有实现 lookup 接口
pub struct RootInode {
    buf: Memfile,
    inner: Mutex<RootInodeInner>,
}
pub struct RootInodeInner {
    // 下一个目录项的 inode 节点号
    inode: usize,
    // 存储下一个目录项的偏移
    offset: usize,
}
impl RootInode {
    pub fn new() -> RootInode {
        let res = RootInode {
            buf: Memfile::new(String::from("dev")),
            inner: Mutex::new(RootInodeInner {
                inode: 1,
                offset: 0,
            }),
        };
        // dev 文件不会像 proc 一样动态更新，所以直接写入就行，不用每次刷新
        res.push(
            String::from("zero"),
            ModeFlags::CHAR | ModeFlags::READ | ModeFlags::WRITE,
        );
        res.push(
            String::from("null"),
            ModeFlags::CHAR | ModeFlags::READ | ModeFlags::WRITE,
        );
        res
    }

    pub fn inner_exclusive_access(&self) -> MutexGuard<'_, RootInodeInner> {
        self.inner.lock()
    }

    /// 存储目录项
    fn push(&self, name: String, mode: ModeFlags) {
        let mut inner = self.inner_exclusive_access();
        let mut dirent = Dirent::new(name, inner.inode as u64, mode, 0);
        inner.inode += 1;
        inner.offset += dirent.d_reclen as usize;
        dirent.d_off = inner.offset as i64;
        self.buf.write_bytes(dirent.as_bytes());
    }
}
impl Inode for RootInode {
    fn get_attr(&self) -> Kstat {
        Kstat {
            st_dev: 0,
            st_ino: 0,
            st_mode: (ModeFlags::DIRECTORY | ModeFlags::RWX).bits(),
            st_nlink: 1,
            st_uid: 0,
            st_gid: 0,
            st_rdev: 0,
            __pad: 0,
            st_size: 4096,
            st_blksize: 512,
            __pad2: 0,
            st_blocks: 0,
            st_atime_sec: 0,
            st_atime_nsec: 0,
            st_mtime_sec: 0,
            st_mtime_nsec: 0,
            st_ctime_sec: 0,
            st_ctime_nsec: 0,
            __unused: [0; 2],
        }
    }

    fn lookup(&self, name: &str) -> Option<Arc<dyn Inode>> {
        match name {
            "zero" => Some(Zero.get_inode()),
            "null" => Some(Null.get_inode()),
            _ => None,
        }
    }

    fn read_dentry(&self, off: usize) -> Option<Dirent> {
        // 读取目录项长度(d_reclen)
        self.buf.lseek(
            (off + core::mem::size_of::<u64>() + core::mem::size_of::<i64>()) as isize,
            SEEK_SET,
        );
        let mut bytes = [0u8; 2];
        self.buf.read_bytes(&mut bytes);
        let len = u16::from_le_bytes(bytes);
        if len == 0 {
            return None;
        }

        // 读取目录项
        self.buf.lseek(off as isize, SEEK_SET);
        let mut res = Dirent::empty();
        res.d_reclen = len;
        let read_size = self.buf.read_bytes(res.as_mut_bytes());
        assert_eq!(len, res.d_reclen);
        assert_eq!(len, read_size as u16);
        Some(res)
    }
}

lazy_static! {
    pub static ref DEV_ROOT_INODE: Arc<dyn Inode> = Arc::new(RootInode::new());
}
