use crate::{
    mm::UserBuffer,
    vfs::{types::*, File, Inode, ModeFlags},
};
use alloc::sync::Arc;
use lazy_static::lazy_static;

/// `/dev/null` 是一个虚拟设备文件，它可以被视为"黑洞"。
///
/// 从 `/dev/null` 读取将立即结束。
/// 用户读取将返回一个"文件结束"指示符，因为它不包含任何数据。
/// 而对于内核来说应该是直接返回 0
///
/// 任何写入 `/dev/null` 的数据都会被丢弃，不保存任何内容。
/// 但对于 write() 来说返回写入的字节数
pub struct Null;
impl File for Null {
    fn get_inode(&self) -> Arc<dyn Inode> {
        Arc::clone(&NULL_FAKE_INODE)
    }

    fn read(&self, buf: UserBuffer) -> usize {
        0
    }

    fn write(&self, buf: UserBuffer) -> usize {
        buf.len()
    }
}

/// 设备文件没有 inode 节点，该结构是虚构的 inode 节点
struct NullFakeInode;
impl Inode for NullFakeInode {
    fn get_attr(&self) -> Kstat {
        Kstat {
            st_dev: 0,
            st_ino: 0,
            st_mode: (ModeFlags::CHAR | ModeFlags::READ | ModeFlags::WRITE).bits(),
            st_nlink: 1,
            st_uid: 0,
            st_gid: 0,
            st_rdev: 0,
            __pad: 0,
            st_size: 0,
            st_blksize: 512,
            __pad2: 0,
            st_blocks: 0,
            st_atime_sec: 0,
            st_atime_nsec: 0,
            st_mtime_sec: 0,
            st_mtime_nsec: 0,
            st_ctime_sec: 0,
            st_ctime_nsec: 0,
            __unused: [0; 2],
        }
    }
}

lazy_static! {
    static ref NULL_FAKE_INODE: Arc<dyn Inode> = Arc::new(NullFakeInode);
}
