use crate::{
    mm::UserBuffer,
    vfs::{types::*, File, Inode, ModeFlags},
};
use alloc::sync::Arc;
use lazy_static::lazy_static;

/// `/dev/zero` 是一个虚拟设备文件，它提供了一个连续的流零字节。
///
/// 读取 `/dev/zero` 将返回无限的零字节流。
/// 这对于需要填充数据的操作很有用，例如初始化内存或创建空白文件。
///
/// 写入 `/dev/zero` 没有实际效果，因为该设备会直接忽略写入的内容。
/// 但对于 write() 来说返回写入的字节数
pub struct Zero;
impl File for Zero {
    fn get_inode(&self) -> Arc<dyn Inode> {
        Arc::clone(&ZERO_FAKE_INODE)
    }

    fn read(&self, buf: UserBuffer) -> usize {
        let read_size = buf.len();
        for slice in buf.buffers {
            slice.fill(0);
        }
        read_size
    }

    fn write(&self, buf: UserBuffer) -> usize {
        buf.len()
    }
}

/// 设备文件没有 inode 节点，该结构是虚构的 inode 节点
struct ZeroFakeInode;
impl Inode for ZeroFakeInode {
    fn get_attr(&self) -> Kstat {
        Kstat {
            st_dev: 0,
            st_ino: 0,
            st_mode: (ModeFlags::CHAR | ModeFlags::READ | ModeFlags::WRITE).bits(),
            st_nlink: 1,
            st_uid: 0,
            st_gid: 0,
            st_rdev: 0,
            __pad: 0,
            st_size: 0,
            st_blksize: 512,
            __pad2: 0,
            st_blocks: 0,
            st_atime_sec: 0,
            st_atime_nsec: 0,
            st_mtime_sec: 0,
            st_mtime_nsec: 0,
            st_ctime_sec: 0,
            st_ctime_nsec: 0,
            __unused: [0; 2],
        }
    }
}

lazy_static! {
    static ref ZERO_FAKE_INODE: Arc<dyn Inode> = Arc::new(ZeroFakeInode);
}
