use crate::{
    timer::{get_time, Timespec},
    vfs::ModeFlags,
};
use alloc::{string::String, vec::Vec};
use core::{convert::TryInto, fmt::Display};

pub const DENTRY_SZ: usize = 32;

#[derive(Clone, Copy, Debug)]
pub struct Fat32Date(u16);
impl Fat32Date {
    /// 将日期戳转为“年-月-日”格式
    pub fn transform(&self) -> (usize, usize, usize) {
        let day = (self.0 & 0x001F) as usize;
        let month = ((self.0 >> 5) & 0x000F) as usize;
        let year = 1980 + (((self.0 >> 11) & 0x007F) as usize);
        (year, month, day)
    }
    /// 将时间戳转换成usize
    pub fn fat32date_tv_sec(&self) -> usize {
        const DAY_SEC: usize = 24 * 60 * 60;
        let (year, month, day) = self.transform();
        println!("{} {} {}", year, month, day);
        let total_days = {
            let mut year_days = (year - 1980) * 365;
            for y in 1980..year {
                if (y % 4 == 0 && y % 100 != 0) || y % 400 == 0 {
                    year_days += 1;
                }
            }
            // 判断当前年份是否为闰年
            let february = if (year % 4 == 0 && year % 100 != 0) || year % 400 == 0 {
                1
            } else {
                0
            };
            let month_days = match month {
                1 => 0,
                2 => 31,
                3 => 59 + february,
                4 => 90 + february,
                5 => 120 + february,
                6 => 151 + february,
                7 => 181 + february,
                8 => 212 + february,
                9 => 243 + february,
                10 => 273 + february,
                11 => 304 + february,
                12 => 334 + february,
                _ => return 0,
            };
            year_days + month_days + day - 1
        };

        total_days * DAY_SEC
    }

    /// 将自1980年后的时间进行转换  tv_sec 为 0 即1980.1.1 0:0:0
    /// tv_sec 为系统上电后的秒数 下同
    pub fn tv_sec_fat32date(tv_sec: u64) -> Self {
        const DAY_SEC: u64 = 24 * 60 * 60;
        let day_after_1980 = (tv_sec / DAY_SEC) as u16;
        let year = day_after_1980 / 365;
        let remaining_days = day_after_1980 % 365;

        let get_days_in_month = |year: u16, month: u8| -> u8 {
            match month {
                1 | 3 | 5 | 7 | 8 | 10 | 12 => 31,
                4 | 6 | 9 | 11 => 30,
                2 => {
                    if (year % 4 == 0 && year % 100 != 0) || (year % 400 == 0) {
                        29
                    } else {
                        28
                    }
                }
                _ => panic!("Invalid month"),
            }
        };
        let mut month = 1;
        let mut day = (remaining_days + 1) as u32;
        while day > get_days_in_month(year, month) as u32 {
            day -= get_days_in_month(year, month) as u32;
            month += 1;
        }
        Self((year as u16) << 11 | (month as u16) << 5 | (day as u16))
    }
    pub fn get_u16(&self) -> u16 {
        self.0
    }
}
impl Display for Fat32Date {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        write!(f, "year-month-day: {:?}", self.transform())
    }
}

#[derive(Clone, Copy, Debug)]
pub struct Fat32Time(u16);
impl Fat32Time {
    /// 将时间戳转为“时-分-秒”格式
    pub fn transform(&self) -> (usize, usize, usize) {
        let second = (((self.0 & 0x001F) % 60) as usize) << 1;
        let minute = (((self.0 >> 5) & 0x003F) % 60) as usize;
        let hour = (((self.0 >> 11) & 0x001F) % 24) as usize;
        (hour, minute, second)
    }

    pub fn fat32time_tv_sec(&self) -> usize {
        let (hour, minute, second) = self.transform();
        hour * 60 * 60 + minute * 60 + second
    }

    /// 将Timespec 中tv_sec 转换成 Fat32Time
    pub fn tv_sec_fat32time(tv_sec: u64) -> Self {
        const DAY_SEC: u64 = 86400;
        let tv_sec = tv_sec % DAY_SEC;
        Self(
            ((tv_sec / 3600) as u16) << 11
                | ((tv_sec / 60 % 60) as u16) << 5
                | ((tv_sec % 60) as u16) / 2,
        )
    }
    pub fn get_u16(&self) -> u16 {
        self.0
    }
}
impl Display for Fat32Time {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        write!(f, "hour-minute-second: {:?}", self.transform())
    }
}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub enum Fat32Type {
    File,
    Directory,
    VolumeLabel,
    Invalid,
}

#[repr(packed)]
#[derive(Clone, Copy, Debug)]
pub struct ShortDirEntry {
    name: [u8; 11],
    pub attribute: u8,
    _winnt_reserved: u8,
    pub creation_tenth: u8,
    pub creation_time: Fat32Time,
    pub creation_date: Fat32Date,
    pub last_acc_date: Fat32Date,
    cluster_high: u16,
    pub last_write_time: Fat32Time,
    pub last_write_date: Fat32Date,
    cluster_low: u16,
    pub size: u32,
}

impl ShortDirEntry {
    /// 空短目录项
    pub fn empty() -> Self {
        unsafe { *(&[0; 0x20] as *const _ as usize as *const Self) }
    }

    /// 新建短目录项
    ///
    /// 对于新创建的目录文件分配一个簇的空间，其它文件不分配
    pub fn new(short_name: &str, attribute: u8, cluster: u32) -> Self {
        let mut name = [0x20; 11];
        let mut i = 0;
        let mut flag = false;
        for c in short_name.as_bytes() {
            if flag && *c == '.' as u8 {
                i = 8;
                continue;
            }
            name[i] = *c;
            i += 1;
            if *c != '.' as u8 {
                flag = true;
            }
        }
        let now_time = Fat32Time::tv_sec_fat32time(Timespec::now_sec() as u64);
        let now_date = Fat32Date::tv_sec_fat32date(Timespec::now_sec() as u64);
        let mut res = Self {
            name,
            attribute,
            _winnt_reserved: 0,
            creation_tenth: 0,
            creation_time: now_time,
            creation_date: now_date,
            last_acc_date: now_date,
            cluster_high: 0,
            last_write_time: now_time,
            last_write_date: now_date,
            cluster_low: 0,
            size: 0,
        };
        res.set_first_cluster(cluster);
        res
    }

    /// 0 表示该目录项是空闲的，同时表明该目录项之后的所有目录项也是空闲的，
    /// 可以直接使用
    pub fn is_free(&self) -> bool {
        self.name[0] == 0x00
    }

    pub fn is_valid(&self) -> bool {
        ((self.attribute & Attribute::LONG_NAME_MASK.bits()) != Attribute::LONG_NAME.bits())
            && self.name[0] != 0xE5
            && self.name[0] != 0x00
    }

    pub fn get_type(&self) -> Fat32Type {
        let attr = self.attribute & (Attribute::DIRECTORY | Attribute::VOLUME_ID).bits();
        if attr == 0x00 {
            Fat32Type::File
        } else if attr == Attribute::DIRECTORY.bits() {
            Fat32Type::Directory
        } else if attr == Attribute::VOLUME_ID.bits() {
            Fat32Type::VolumeLabel
        } else {
            Fat32Type::Invalid
        }
    }

    pub fn get_mode(&self) -> ModeFlags {
        let mut mode = ModeFlags::empty();
        let attribute = Attribute::from_bits(self.attribute).unwrap();
        match self.get_type() {
            Fat32Type::File => mode |= ModeFlags::REGULAR,
            Fat32Type::Directory => mode |= ModeFlags::DIRECTORY,
            _ => {}
        };
        if attribute.contains(Attribute::RDONLY) {
            mode |= ModeFlags::READ | ModeFlags::EXEC;
        } else {
            mode |= ModeFlags::RWX;
        }
        mode
    }

    /// 获取文件首簇号
    ///
    /// 注：根据文档，若 `..` 目录项代表根目录或者未分配空间的文件，其首簇号设置为 0
    pub fn get_first_cluster(&self) -> u32 {
        // 簇号只有低 28 位有效
        let cluster =
            (((self.cluster_high as u32) << 16) + (self.cluster_low as u32)) & 0x0FFF_FFFF;
        match cluster {
            0 => {
                if self.name() == ".." {
                    2
                } else {
                    0
                }
            }
            first_cluster => first_cluster,
        }
    }

    /// 获取短文件名
    ///
    /// 短文件名都以大写形式存储，因此返回的字符串为大写
    pub fn name(&self) -> String {
        let mut short_name: String = String::new();
        for i in 0..8 {
            // 记录文件名
            if self.name[i] == 0x20 {
                break;
            } else {
                short_name.push(self.name[i] as char);
            }
        }
        for i in 0..3 {
            // 记录扩展名
            if self.name[i + 8] == 0x20 {
                break;
            } else {
                if i == 0 {
                    short_name.push('.');
                }
                short_name.push(self.name[i + 8] as char);
            }
        }
        short_name
    }

    /// 计算校验和
    ///
    /// 由于 rust 对算术运算溢出 panic，因此这里使用更大的类型，然后每次对高位清零
    pub fn checksum(&self) -> u8 {
        let mut sum: usize = 0;
        for i in 0..11 {
            if (sum & 1) != 0 {
                sum = 0x80 + (sum >> 1) + self.name[i] as usize;
            } else {
                sum = (sum >> 1) + self.name[i] as usize;
            }
            sum &= u8::MAX as usize;
        }
        sum as u8
    }

    /// 设置文件首簇号
    pub fn set_first_cluster(&mut self, cluster: u32) {
        self.cluster_high = ((cluster & 0xFFFF0000) >> 16) as u16;
        self.cluster_low = (cluster & 0x0000FFFF) as u16;
    }

    /// 清空文件
    pub fn clear(&mut self) {
        self.size = 0;
        self.set_first_cluster(0);
    }

    pub fn delete(&mut self) {
        self.size = 0;
        self.name[0] = 0xE5;
        self.set_first_cluster(0);
    }

    pub fn as_bytes(&self) -> &[u8] {
        unsafe { core::slice::from_raw_parts(self as *const _ as usize as *const u8, DENTRY_SZ) }
    }

    pub fn as_bytes_mut(&mut self) -> &mut [u8] {
        unsafe { core::slice::from_raw_parts_mut(self as *mut _ as usize as *mut u8, DENTRY_SZ) }
    }
}

#[repr(packed)]
#[derive(Clone, Copy, Debug)]
pub struct LongDirEntry {
    order: u8,
    name1: [u16; 5],
    attribute: u8,
    _type: u8,
    pub checksum: u8,
    name2: [u16; 6],
    _first_cluster_low: u16,
    name3: [u16; 2],
}

impl LongDirEntry {
    pub fn empty() -> Self {
        unsafe { core::mem::zeroed() }
    }

    fn portion(portion_name: &[u8], order: u8, checksum: u8) -> Self {
        let mut name = [0xFFFF; 13];
        for (i, c) in portion_name.iter().enumerate() {
            name[i] = *c as u16;
        }
        if portion_name.len() < 13 {
            name[portion_name.len()] = 0x0000;
        }
        let (name1, right) = name.split_at(5);
        let (name2, name3) = right.split_at(6);
        Self {
            order,
            name1: name1.try_into().unwrap(),
            attribute: Attribute::LONG_NAME.bits(),
            _type: 0,
            checksum,
            name2: name2.try_into().unwrap(),
            _first_cluster_low: 0,
            name3: name3.try_into().unwrap(),
        }
    }

    /// 只支持 ASCII 字符
    pub fn new(name: &str, checksum: u8) -> Vec<Self> {
        let mut res = Vec::new();
        let num = (name.len() + 12) / 13;
        for (i, portion_name) in name.as_bytes().chunks(13).enumerate() {
            let mut order = i + 1;
            if order == num {
                order |= 0x40;
            }
            res.push(LongDirEntry::portion(portion_name, order as u8, checksum));
        }
        res
    }

    pub fn is_valid(&self) -> bool {
        ((self.attribute & Attribute::LONG_NAME_MASK.bits()) == Attribute::LONG_NAME.bits())
            && self.order != 0xE5
    }

    pub fn delete(&mut self) {
        self.order = 0xE5;
    }

    pub fn order(&self) -> usize {
        (self.order & (!0x40)) as usize
    }

    pub fn name(&self) -> String {
        let mut name = String::new();
        for i in 0..5 {
            match self.name1[i] as u8 {
                0 => return name,
                c => name.push(c as char),
            }
        }
        for i in 0..6 {
            match self.name2[i] as u8 {
                0 => return name,
                c => name.push(c as char),
            }
        }
        for i in 0..2 {
            match self.name3[i] as u8 {
                0 => return name,
                c => name.push(c as char),
            }
        }
        return name;
    }

    pub fn as_bytes(&self) -> &[u8] {
        unsafe { core::slice::from_raw_parts(self as *const _ as usize as *const u8, DENTRY_SZ) }
    }

    pub fn as_bytes_mut(&mut self) -> &mut [u8] {
        unsafe { core::slice::from_raw_parts_mut(self as *mut _ as usize as *mut u8, DENTRY_SZ) }
    }

    pub fn to_short_dentry(self) -> ShortDirEntry {
        unsafe { *(&self as *const _ as usize as *const ShortDirEntry) }
    }
}

bitflags! {
    pub struct Attribute: u8 {
        const RDONLY = 0x01;
        const HIDDEN = 0x02;
        const SYSTEM = 0x04;
        const VOLUME_ID = 0x08;
        const DIRECTORY = 0x10;
        const ARCHIVE = 0x20;
        const LONG_NAME = Self::RDONLY.bits
                        | Self::HIDDEN.bits
                        | Self::SYSTEM.bits
                        | Self::VOLUME_ID.bits;
        const LONG_NAME_MASK = Self::LONG_NAME.bits
                                | Self::DIRECTORY.bits
                                | Self::ARCHIVE.bits;
    }
}

impl Attribute {
    pub fn new(mode: ModeFlags) -> Self {
        let mut res = Attribute::empty();
        if mode.contains(ModeFlags::DIRECTORY) {
            res |= Attribute::DIRECTORY;
        }
        if !mode.contains(ModeFlags::WRITE) {
            res |= Attribute::RDONLY;
        }
        res
    }
}
