use core::{fmt::Display, usize};

use super::{
    dir_entry::{
        Attribute, Fat32Date, Fat32Time, Fat32Type, LongDirEntry, ShortDirEntry, DENTRY_SZ,
    },
    Fat32SuperBlock,
};
use crate::{
    console::print,
    vfs::{types::*, Inode, ModeFlags},
};
use alloc::{
    string::{String, ToString},
    sync::Arc,
    vec::Vec,
};
use lazy_static::*;
use spin::{Mutex, MutexGuard};

/// 文件 inode 节点
///
/// 对于 FAT32 文件系统，一个文件可以由首簇号识别，对应文件的数据区。
/// 而 FAT32 的目录项其实可以认为就是文件的 inode 节点，存储着文件的相关信息。
/// `Fat32Inode` 同时保存父目录的首簇号以及文件目录项在父目录的偏移，
/// 从而确定文件目录项的位置，这样设计也是为了便于获得文件名以及文件路径。
///
/// 我们获取得到的 inode 节点中的 `short_dentry` 对应文件本身的目录项，而不含 `.`
/// 以及 `..` 目录项。另外除了目录文件，普通文件不会预先分配空间，此时得到的首簇号都是 0。
/// 这时节点中的 `parent_first_cluster` 以及 `parent_offset` 字段还可用来进一步区分文件，
/// 因为其父目录必定分配了相应的空间。
///
/// 对于根目录，会创建一个虚拟的 inode 节点。
pub struct Fat32Inode {
    /// 文件超级块对象
    sb: Arc<Fat32SuperBlock>,
    /// 文件父目录的首簇号
    parent_first_cluster: u32,
    name: String,
    inner: Mutex<Fat32InodeInner>,
}

pub struct Fat32InodeInner {
    dirt: bool,
    /// 文件短目录项
    short_dentry: ShortDirEntry,
    /// `short_dentry` 在父目录的偏移
    ///
    /// 在更改文件名的时候会更改偏移
    parent_offset: usize,
    /// 添加 时间信息通过utime测试 其实nsec没什么用...
    st_atime_sec: i64,
    st_atime_nsec: i64,
    st_mtime_sec: i64,
    st_mtime_nsec: i64,
}

impl Fat32Inode {
    fn inner_exclusive_access(&self) -> MutexGuard<'_, Fat32InodeInner> {
        self.inner.lock()
    }

    fn new(
        sb: Arc<Fat32SuperBlock>,
        parent_first_cluster: u32,
        parent_offset: usize,
        name: String,
        short_dentry: ShortDirEntry,
    ) -> Self {
        Self {
            sb,
            parent_first_cluster,
            name,
            inner: Mutex::new(Fat32InodeInner {
                dirt: false,
                short_dentry,
                parent_offset,
                st_atime_sec: 0,
                st_atime_nsec: 0,
                st_mtime_sec: 0,
                st_mtime_nsec: 0,
            }),
        }
    }

    pub fn root_inode(sb: Arc<Fat32SuperBlock>) -> Self {
        Fat32Inode::new(
            sb,
            2,
            0,
            String::from("/"),
            ShortDirEntry::new("/", Attribute::DIRECTORY.bits(), 2),
        )
    }

    /// 比较两个 inode 节点是否相同
    fn compare(&self, other: &Fat32Inode) -> bool {
        // 两个 inode 节点可能分属不同的 FAT32 文件系统
        if Arc::as_ptr(&self.sb) != Arc::as_ptr(&other.sb) {
            return false;
        }
        let first_cluster = self.get_first_cluster();
        let other_first_cluster = other.get_first_cluster();
        if first_cluster != 0 && other_first_cluster != 0 {
            return first_cluster == other_first_cluster;
        } else {
            return self.parent_first_cluster == other.parent_first_cluster
                && self.inner_exclusive_access().parent_offset
                    == other.inner_exclusive_access().parent_offset;
        }
    }

    /// 获取文件首簇号
    fn get_first_cluster(&self) -> u32 {
        self.inner_exclusive_access()
            .short_dentry
            .get_first_cluster()
    }

    fn get_type(&self) -> Fat32Type {
        self.inner_exclusive_access().short_dentry.get_type()
    }

    fn get_mode(&self) -> ModeFlags {
        self.inner_exclusive_access().short_dentry.get_mode()
    }

    fn get_size(&self) -> usize {
        match self.get_type() {
            Fat32Type::File => self.inner_exclusive_access().short_dentry.size as usize,
            Fat32Type::Directory => {
                self.sb.chain_of_cluster(self.get_first_cluster()).len()
                    * self.sb.bytes_per_cluster()
            }
            _ => 0,
        }
    }

    pub fn sync(&self) {
        let inner = self.inner_exclusive_access();
        if inner.dirt {
            self.sb.write(
                self.parent_first_cluster,
                inner.parent_offset,
                inner.short_dentry.as_bytes(),
            );
        }
    }

    pub fn list(&self) {
        println!("----------------- ls -----------------");
        let mut off = 0;
        while let Some((_, parent_offset, dentry_name, _)) =
            self.sb.read_dentry(self.get_first_cluster(), off)
        {
            println!("{}", dentry_name);
            off = parent_offset + DENTRY_SZ;
        }
        println!("--------------------------------------");
    }

    pub fn get_atime_sec(&self) -> i64 {
        let inner = self.inner_exclusive_access();
        (inner.st_atime_sec) as i64
    }

    pub fn get_atime_nsec(&self) -> i64 {
        0
    }

    pub fn get_mtime_sec(&self) -> i64 {
        let inner = self.inner_exclusive_access();
        // let last_write_date = inner.short_dentry.last_write_date;
        // let last_write_time = inner.short_dentry.last_write_time;
        //(last_write_date.fat32date_tv_sec() + last_write_time.fat32time_tv_sec()) as i64
        inner.st_mtime_sec as i64
    }
    pub fn get_mtime_nsec(&self) -> i64 {
        0
    }
}

impl Inode for Fat32Inode {
    fn get_name(&self) -> String {
        String::clone(&self.name)
    }

    fn get_attr(&self) -> Kstat {
        let first_cluster = self.get_first_cluster();
        let st_blocks = (self.sb.chain_of_cluster(first_cluster).len()
            * self.sb.bytes_per_cluster()
            / 512) as i64;
        Kstat {
            st_dev: 0,
            st_ino: self.get_first_cluster() as u64,
            st_mode: self.get_mode().bits(),
            st_nlink: 1,
            st_uid: 0,
            st_gid: 0,
            st_rdev: 0,
            __pad: 0,
            st_size: self.get_size() as i64,
            st_blksize: 512,
            __pad2: 0,
            st_blocks,
            st_atime_sec: self.get_atime_sec(),
            st_atime_nsec: self.get_atime_nsec(),
            st_mtime_sec: self.get_mtime_sec(),
            st_mtime_nsec: self.get_mtime_nsec(),
            st_ctime_sec: 0,
            st_ctime_nsec: 0,
            __unused: [0; 2],
        }
    }

    fn create(&self, name: &str, mode: ModeFlags) -> Option<Arc<dyn Inode>> {
        let short_name = self.sb.generate_short_name(name);
        let attribute = Attribute::new(mode);
        let short_dentry = ShortDirEntry::new(&short_name, attribute.bits(), 0);
        let long_dentry_vec = LongDirEntry::new(name, short_dentry.checksum());
        let first_cluster = self.get_first_cluster();
        if let Some(mut off) = self
            .sb
            .alloc_dentry(first_cluster, long_dentry_vec.len() + 1)
        {
            for long_dentry in long_dentry_vec.iter().rev() {
                self.sb.write(first_cluster, off, long_dentry.as_bytes());
                off += DENTRY_SZ;
            }
            self.sb.write(first_cluster, off, short_dentry.as_bytes());
            Some(insert_inode_table(Fat32Inode::new(
                Arc::clone(&self.sb),
                first_cluster,
                off,
                name.to_string(),
                short_dentry,
            )))
        } else {
            None
        }
    }

    fn lookup(&self, name: &str) -> Option<Arc<dyn Inode>> {
        // FAT32 文件名匹配不区分大小写
        let name = name.to_uppercase();
        match name.as_str() {
            "." => get_inode(self),
            ".." => {
                // 文件本身或者其父目录是根节点
                if self.get_first_cluster() == 2 || self.parent_first_cluster == 2 {
                    return get_root_inode(&self.sb);
                }

                // 读取父目录的父目录
                let mut tmp = ShortDirEntry::empty();
                self.sb
                    .read(self.parent_first_cluster, 0x20, tmp.as_bytes_mut());
                // 查找父目录实际的短目录项
                // 根据首簇号进行匹配，这时父目录肯定分配了空间
                let mut off = 0;
                while let Some((_, parent_offset, name, short_dentry)) =
                    self.sb.read_dentry(tmp.get_first_cluster(), off)
                {
                    if short_dentry.get_first_cluster() == self.parent_first_cluster {
                        return Some(insert_inode_table(Fat32Inode::new(
                            Arc::clone(&self.sb),
                            tmp.get_first_cluster(),
                            parent_offset,
                            name,
                            short_dentry,
                        )));
                    }
                    off = parent_offset + DENTRY_SZ;
                }
                None
            }
            _ => {
                let mut off = 0;
                while let Some((_, parent_offset, dentry_name, short_dentry)) =
                    self.sb.read_dentry(self.get_first_cluster(), off)
                {
                    if dentry_name.to_uppercase() == name {
                        return Some(insert_inode_table(Fat32Inode::new(
                            Arc::clone(&self.sb),
                            self.get_first_cluster(),
                            parent_offset,
                            dentry_name,
                            short_dentry,
                        )));
                    }
                    off = parent_offset + DENTRY_SZ;
                }
                None
            }
        }
    }

    fn read(&self, off: usize, buf: &mut [u8]) -> usize {
        let size = self.get_size();
        let read_size = (size - off).min(buf.len());
        if read_size != 0 {
            let buf = &mut buf[..read_size];
            self.sb.read(self.get_first_cluster(), off, buf);
        }
        read_size
    }

    fn write(&self, off: usize, buf: &[u8]) -> usize {
        let new_size = off + buf.len();
        if let Some(first_cluster) = self.sb.increase_size(self.get_first_cluster(), new_size) {
            self.sb.write(first_cluster, off, buf);
            let mut inner = self.inner_exclusive_access();
            inner.dirt = true;
            inner.short_dentry.size = new_size as u32;
            inner.short_dentry.set_first_cluster(first_cluster);
            buf.len()
        } else {
            0
        }
    }

    fn truncate(&self) {
        let mut inner = self.inner_exclusive_access();
        self.sb
            .dealloc_chain_of_cluster(inner.short_dentry.get_first_cluster());
        inner.dirt = true;
        inner.short_dentry.clear();
    }

    fn mkdir(&self, name: &str, mode: ModeFlags) {
        let short_name = self.sb.generate_short_name(name);
        let attribute = Attribute::new(mode) | Attribute::DIRECTORY;
        let alloc_cluster = self.sb.alloc_cluster().unwrap();
        let short_dentry = ShortDirEntry::new(&short_name, attribute.bits(), alloc_cluster);
        let long_dentry_vec = LongDirEntry::new(name, short_dentry.checksum());
        let first_cluster = self.get_first_cluster();
        if let Some(mut off) = self
            .sb
            .alloc_dentry(first_cluster, long_dentry_vec.len() + 1)
        {
            for long_dentry in long_dentry_vec.iter().rev() {
                self.sb.write(first_cluster, off, long_dentry.as_bytes());
                off += DENTRY_SZ;
            }
            self.sb.write(first_cluster, off, short_dentry.as_bytes());
        }
        let dot = ShortDirEntry::new(".", Attribute::DIRECTORY.bits(), alloc_cluster);
        let dotdot = ShortDirEntry::new(
            "..",
            Attribute::DIRECTORY.bits(),
            if first_cluster == 2 { 0 } else { first_cluster },
        );
        self.sb.write(alloc_cluster, 0, dot.as_bytes());
        self.sb.write(alloc_cluster, 0x20, dotdot.as_bytes());
    }

    fn read_dentry(&self, off: usize) -> Option<Dirent> {
        self.sb.read_dentry(self.get_first_cluster(), off).map(
            |(_, parent_offset, name, short_dentry)| {
                Dirent::new(
                    name,
                    short_dentry.get_first_cluster() as u64,
                    short_dentry.get_mode(),
                    (parent_offset + 0x20) as i64,
                )
            },
        )
    }

    /// 不管可能存在已经被打开的文件，直接删除
    fn unlink(&self, name: &str) {
        let name = name.to_uppercase();
        let first_cluster = self.get_first_cluster();
        let mut off = 0;
        while let Some((option, short_dentry_off, dentry_name, mut short_dentry)) =
            self.sb.read_dentry(first_cluster, off)
        {
            if dentry_name.to_uppercase() == name {
                // 删除簇链
                self.sb
                    .dealloc_chain_of_cluster(short_dentry.get_first_cluster());
                // 删除长目录项
                if let Some(mut long_dentry_off) = option {
                    let mut long_dentry = LongDirEntry::empty();
                    while long_dentry_off < short_dentry_off {
                        self.sb
                            .read(first_cluster, long_dentry_off, long_dentry.as_bytes_mut());
                        long_dentry.delete();
                        self.sb
                            .write(first_cluster, long_dentry_off, long_dentry.as_bytes());
                        long_dentry_off += DENTRY_SZ;
                    }
                }
                // 删除短目录项
                short_dentry.delete();
                self.sb
                    .write(first_cluster, short_dentry_off, short_dentry.as_bytes());
                break;
            }
            off = short_dentry_off + DENTRY_SZ;
        }
    }

    /// 默认两个文件处于同一个文件系统
    fn rename(&self, old_name: &str, new_dir: Arc<dyn Inode>, new_name: &str) {
        let old_name = old_name.to_uppercase();
        let first_cluster = self.get_first_cluster();
        let mut off = 0;
        while let Some((option, short_dentry_off, dentry_name, mut short_dentry)) =
            self.sb.read_dentry(first_cluster, off)
        {
            if dentry_name.to_uppercase() == old_name {
                let new_first_cluster = new_dir.get_attr().st_ino as u32;
                let new_short_name = self.sb.generate_short_name(new_name);
                let mut new_short_dentry = ShortDirEntry::new(
                    &new_short_name,
                    short_dentry.attribute,
                    short_dentry.get_first_cluster(),
                );
                new_short_dentry.size = short_dentry.size;
                let new_long_dentry_vec = LongDirEntry::new(new_name, new_short_dentry.checksum());
                let mut new_off = self
                    .sb
                    .alloc_dentry(new_first_cluster, new_long_dentry_vec.len() + 1)
                    .unwrap();
                // 写入新长目录项
                for new_long_dentry in new_long_dentry_vec.iter().rev() {
                    self.sb
                        .write(new_first_cluster, new_off, new_long_dentry.as_bytes());
                    new_off += DENTRY_SZ;
                }
                // 写入新短目录项
                self.sb
                    .write(new_first_cluster, new_off, new_short_dentry.as_bytes());

                // 删除旧长目录项
                if let Some(mut long_dentry_off) = option {
                    let mut long_dentry = LongDirEntry::empty();
                    while long_dentry_off < short_dentry_off {
                        self.sb
                            .read(first_cluster, long_dentry_off, long_dentry.as_bytes_mut());
                        long_dentry.delete();
                        self.sb
                            .write(first_cluster, long_dentry_off, long_dentry.as_bytes());
                        long_dentry_off += DENTRY_SZ;
                    }
                }
                // 删除旧短目录项
                short_dentry.delete();
                self.sb
                    .write(first_cluster, short_dentry_off, short_dentry.as_bytes());

                break;
            }
            off = short_dentry_off + DENTRY_SZ;
        }
    }

    fn set_timestamps(
        &self,
        st_atime_sec: Option<i64>,
        st_atime_nsec: Option<i64>,
        st_mtime_sec: Option<i64>,
        st_mtime_nsec: Option<i64>,
    ) {
        let mut inner = self.inner_exclusive_access();
        if let Some(atime_sec) = st_atime_sec {
            inner.st_atime_sec = atime_sec;
            inner.short_dentry.last_acc_date = Fat32Date::tv_sec_fat32date(atime_sec as u64);
            inner.dirt = true;
        }

        if let Some(mtime_sec) = st_mtime_sec {
            inner.st_mtime_sec = mtime_sec;
            inner.short_dentry.last_write_date = Fat32Date::tv_sec_fat32date(mtime_sec as u64);
            inner.short_dentry.last_write_time = Fat32Time::tv_sec_fat32time(mtime_sec as u64);
            inner.dirt = true;
        }
    }
}

impl Drop for Fat32Inode {
    fn drop(&mut self) {
        let inner = self.inner_exclusive_access();
        if inner.dirt {
            self.sb.write(
                self.parent_first_cluster,
                inner.parent_offset,
                inner.short_dentry.as_bytes(),
            );
        }
    }
}

lazy_static! {
    /// 所有 FAT32 文件系统共享的 inode 节点表，存储文件的 `Fat32Inode` 节点，
    /// 向 vfs 层输出可用 inode 节点，保证节点只有一份。
    static ref INODE_TABLE: Mutex<Vec<Option<Arc<Fat32Inode>>>> = Mutex::new(Vec::new());
}

/// 向 inode 表插入结点
pub fn insert_inode_table(inode: Fat32Inode) -> Arc<dyn Inode> {
    let mut inode_table = INODE_TABLE.lock();
    // 边查找边将不被使用的节点回收
    for i in 0..inode_table.len() {
        if let Some(tmp) = inode_table[i].as_ref() {
            if inode.compare(tmp) {
                let res = Arc::clone(tmp);
                return res;
            }
            if Arc::strong_count(tmp) == 1 {
                inode_table[i].take();
            }
        }
    }
    let res = Arc::new(inode);
    if let Some(num) = (0..inode_table.len()).find(|num| inode_table[*num].is_none()) {
        inode_table[num] = Some(Arc::clone(&res));
    } else {
        inode_table.push(Some(Arc::clone(&res)));
    }
    res
}

fn get_inode(inode: &Fat32Inode) -> Option<Arc<dyn Inode>> {
    let inode_table = INODE_TABLE.lock();
    for i in 0..inode_table.len() {
        if inode_table[i].is_some() {
            let tmp = Arc::clone(inode_table[i].as_ref().unwrap());
            if inode as *const Fat32Inode == Arc::as_ptr(&tmp) {
                return Some(tmp);
            }
        }
    }
    None
}

pub fn get_root_inode(sb: &Fat32SuperBlock) -> Option<Arc<dyn Inode>> {
    let inode_table = INODE_TABLE.lock();
    for i in 0..inode_table.len() {
        if inode_table[i].is_some() {
            let tmp = Arc::clone(inode_table[i].as_ref().unwrap());
            if sb as *const Fat32SuperBlock == Arc::as_ptr(&tmp.sb) && tmp.get_first_cluster() == 2
            {
                return Some(tmp);
            }
        }
    }
    None
}

pub fn inode_sync_all() {
    let mut inode_table = INODE_TABLE.lock();
    for i in 0..inode_table.len() {
        if let Some(inode) = inode_table[i].as_ref() {
            if Arc::strong_count(inode) == 1 {
                inode_table[i].take();
            } else {
                inode.sync();
            }
        }
    }
}
