use core::convert::TryInto;

#[derive(Clone, Copy, Debug)]
pub struct BPB {
    pub bytes_per_sector: u16,   // 每扇区字节数
    pub sectors_per_cluster: u8, // 每簇扇区数
    pub reserved_sectors: u16,   // 保留区扇区数
    pub fats: u8,                // FAT 表数目
    pub total_sectors: u32,      // 总扇区数
    pub sectors_per_fat: u32,    // 一个 FAT 表所占扇区数
    pub root_cluster: u32,       // 根目录首簇号
    pub fsinfo_sector: u16,      // FSInfo扇区的扇区号
}

impl BPB {
    pub fn new(dbr: &[u8; 512]) -> Self {
        Self {
            bytes_per_sector: u16::from_le_bytes(dbr[0xb..0xb + 2].try_into().unwrap()),
            sectors_per_cluster: dbr[0xd],
            reserved_sectors: u16::from_le_bytes(dbr[0xe..0xe + 2].try_into().unwrap()),
            fats: dbr[0x10],
            total_sectors: u32::from_le_bytes(dbr[0x20..0x20 + 4].try_into().unwrap()),
            sectors_per_fat: u32::from_le_bytes(dbr[0x24..0x24 + 4].try_into().unwrap()),
            root_cluster: u32::from_le_bytes(dbr[0x2c..0x2c + 4].try_into().unwrap()),
            fsinfo_sector: u16::from_le_bytes(dbr[0x30..0x30 + 2].try_into().unwrap()),
        }
    }

    pub fn bytes_per_cluster(&self) -> usize {
        self.bytes_per_sector as usize * self.sectors_per_cluster as usize
    }
}

#[derive(Clone, Copy, Debug)]
pub struct FSInfo {
    pub free_count: u32,         // 空簇数
    pub last_alloc_cluster: u32, // 最后一个被分配的簇号
}

impl FSInfo {
    pub fn new(fsinfo: &[u8; 512]) -> Self {
        unsafe {
            assert_eq!(0x4161_5252, *(fsinfo.as_ptr() as *const u32));
            assert_eq!(0x6141_7272, *(fsinfo.as_ptr().add(0x1E4) as *const u32));
            assert_eq!(0xAA55_0000, *(fsinfo.as_ptr().add(0x1FC) as *const u32));
            Self {
                free_count: *(fsinfo.as_ptr().add(0x1E8) as *const u32),
                last_alloc_cluster: *(fsinfo.as_ptr().add(0x1EC) as *const u32),
            }
        }
    }
}
