use super::{
    dir_entry::{LongDirEntry, ShortDirEntry, DENTRY_SZ},
    inode::{get_root_inode, inode_sync_all, insert_inode_table, Fat32Inode},
    layout::{FSInfo, BPB},
};
use crate::{
    drivers::get_block_cache,
    vfs::{Inode, PartitionTable, SuperBlock},
};
use alloc::{
    string::{String, ToString},
    sync::Arc,
    vec,
    vec::Vec,
};
use spin::{Mutex, MutexGuard};

pub const FREE_CLUSTER: u32 = 0x0000_0000;
pub const BOUND_CLUSTER: u32 = 0x0FFF_FFF8;
pub const END_CLUSTER: u32 = 0x0FFF_FFFF;
pub const BAD_CLUSTER: u32 = 0x0FFF_FFF7;

pub struct Fat32SuperBlock {
    start_sect: usize,
    bpb: BPB,
    inner: Mutex<Fat32SuperBlockInner>,
}

pub struct Fat32SuperBlockInner {
    fsinfo: FSInfo,
    numeric_allocator: NumericAllocator,
}

impl Fat32SuperBlock {
    pub fn inner_exclusive_access(&self) -> MutexGuard<'_, Fat32SuperBlockInner> {
        self.inner.lock()
    }

    pub fn new(partition_table: PartitionTable) -> Arc<Self> {
        println!("[FAT32]: load FAT32");
        let bpb = get_block_cache(partition_table.start_sect as usize)
            .lock()
            .read(0, BPB::new);
        println!("[FAT32]: BPB = {:#x?}", bpb);
        let fsinfo =
            get_block_cache(partition_table.start_sect as usize + bpb.fsinfo_sector as usize)
                .lock()
                .read(0, FSInfo::new);
        println!("[FAT32]: FSInfo = {:#x?}", fsinfo);
        let res = Arc::new(Fat32SuperBlock {
            start_sect: partition_table.start_sect as usize,
            bpb,
            inner: Mutex::new(Fat32SuperBlockInner {
                fsinfo,
                numeric_allocator: NumericAllocator::new(),
            }),
        });
        println!("[FAT32]: chain of root = {:#x?}", res.chain_of_cluster(2));
        let root_inode = Fat32Inode::root_inode(Arc::clone(&res));
        root_inode.list();
        insert_inode_table(root_inode);
        println!("[FAT32]: load FAT32 successfully!");
        res
    }

    /// 数据区起始扇区号
    fn first_data_sector(&self) -> usize {
        self.bpb.reserved_sectors as usize
            + (self.bpb.fats as usize * self.bpb.sectors_per_fat as usize)
    }

    /// 簇号 N 对应的数据起始扇区号
    fn first_sector_of_cluster(&self, cluster: u32) -> usize {
        self.first_data_sector() + ((cluster as usize - 2) * self.bpb.sectors_per_cluster as usize)
    }

    pub fn bytes_per_cluster(&self) -> usize {
        self.bpb.bytes_per_cluster()
    }

    /// 从簇号 `cluster` 偏移量 `off` 处读出内容到 `buf`
    ///
    /// 保证 `buf` 长度小于簇的大小
    fn read_slice_from_cluster(&self, cluster: u32, off: usize, buf: &mut [u8]) {
        const SECTOR_SIZE: usize = 512;
        let mut sect = self.first_sector_of_cluster(cluster) + off / SECTOR_SIZE;
        let sect_off = off % SECTOR_SIZE;
        if sect_off + buf.len() <= SECTOR_SIZE {
            read_slice_from_block_cache(self.start_sect + sect, sect_off, buf);
        } else {
            let (left, right) = buf.split_at_mut(SECTOR_SIZE - sect_off);
            read_slice_from_block_cache(self.start_sect + sect, sect_off, left);
            for chunk in right.chunks_mut(SECTOR_SIZE) {
                sect += 1;
                read_slice_from_block_cache(self.start_sect + sect, 0, chunk);
            }
        }
    }

    /// 从首簇号 `first_cluster` 开始的簇链中偏移量 `off` 处读出内容到 `buf`
    pub fn read(&self, first_cluster: u32, mut off: usize, buf: &mut [u8]) {
        let bytes_per_cluster = self.bpb.bytes_per_cluster();
        let mut cluster = first_cluster;
        let cluster_off = off % bytes_per_cluster;
        while off >= bytes_per_cluster {
            cluster = self.get_next_cluster(cluster).unwrap();
            off -= bytes_per_cluster;
        }
        if cluster_off + buf.len() <= bytes_per_cluster {
            self.read_slice_from_cluster(cluster, cluster_off, buf);
        } else {
            let (left, right) = buf.split_at_mut(bytes_per_cluster - cluster_off);
            self.read_slice_from_cluster(cluster, cluster_off, left);
            for chunk in right.chunks_mut(bytes_per_cluster) {
                cluster = self.get_next_cluster(cluster).unwrap();
                self.read_slice_from_cluster(cluster, 0, chunk);
            }
        }
    }

    /// 从首簇为 `first_cluster`，偏移量为 `off` 的开始簇链中，查找第一个合法的短目录项
    ///
    /// 返回值：(长目录项的偏移，短目录项的偏移，文件名，短目录项)
    pub fn read_dentry(
        &self,
        first_cluster: u32,
        mut off: usize,
    ) -> Option<(Option<usize>, usize, String, ShortDirEntry)> {
        let size = self.bpb.bytes_per_cluster() * self.chain_of_cluster(first_cluster).len();
        let mut tmp = LongDirEntry::empty();
        while off < size {
            self.read(first_cluster, off, tmp.as_bytes_mut());
            if tmp.is_valid() {
                // 以长文件名形式存储，解析长目录项
                let mut name = tmp.name();
                let num = tmp.order();
                let checksum = tmp.checksum;
                let mut flag = true;
                for _ in 1..num {
                    off += DENTRY_SZ;
                    self.read(first_cluster, off, tmp.as_bytes_mut());
                    if !tmp.is_valid() || tmp.checksum != checksum {
                        off += DENTRY_SZ;
                        flag = false;
                        break;
                    }
                    name = tmp.name() + &name;
                }
                if !flag {
                    continue;
                }

                // 获取对应的短文件名，并做最终校验
                off += DENTRY_SZ;
                self.read(first_cluster, off, tmp.as_bytes_mut());
                let short_dentry = tmp.to_short_dentry();
                if short_dentry.is_valid() && short_dentry.checksum() == checksum {
                    return Some((Some(off - num * DENTRY_SZ), off, name, short_dentry));
                }

                off += DENTRY_SZ;
            } else {
                // 以短文件名存储，直接返回
                let short_dentry = tmp.to_short_dentry();
                if short_dentry.is_valid() {
                    return Some((None, off, short_dentry.name(), short_dentry));
                }
                // name[0] = 0x00，代表之后的目录项都为空
                if short_dentry.is_free() {
                    return None;
                }
                off += DENTRY_SZ;
            }
        }
        None
    }

    /// 将 `buf` 内容写到簇号 `cluster` 偏移量 `off` 处
    ///
    /// 保证 `buf` 长度小于簇的大小
    fn write_slice_to_cluster(&self, cluster: u32, off: usize, buf: &[u8]) {
        const SECTOR_SIZE: usize = 512;
        let mut sect = self.first_sector_of_cluster(cluster) + off / SECTOR_SIZE;
        let sect_off = off % SECTOR_SIZE;
        if sect_off + buf.len() <= SECTOR_SIZE {
            write_slice_to_block_cache(self.start_sect + sect, sect_off, buf);
        } else {
            let (left, right) = buf.split_at(SECTOR_SIZE - sect_off);
            write_slice_to_block_cache(self.start_sect + sect, sect_off, left);
            for chunk in right.chunks(SECTOR_SIZE) {
                sect += 1;
                write_slice_to_block_cache(self.start_sect + sect, 0, chunk);
            }
        }
    }

    /// 将 `buf` 内容写到首簇号 `first_cluster` 开始的簇链中偏移量 `off` 处
    pub fn write(&self, first_cluster: u32, mut off: usize, buf: &[u8]) {
        let bytes_per_cluster = self.bpb.bytes_per_cluster();
        let mut cluster = first_cluster;
        let cluster_off = off % bytes_per_cluster;
        while off >= bytes_per_cluster {
            // println!("ready to get next cluster: {}", cluster);
            cluster = self.get_next_cluster(cluster).unwrap();
            off -= bytes_per_cluster;
        }
        if cluster_off + buf.len() <= bytes_per_cluster {
            self.write_slice_to_cluster(cluster, cluster_off, buf);
        } else {
            let (left, right) = buf.split_at(bytes_per_cluster - cluster_off);
            self.write_slice_to_cluster(cluster, cluster_off, left);
            for chunk in right.chunks(bytes_per_cluster) {
                cluster = self.get_next_cluster(cluster).unwrap();
                self.write_slice_to_cluster(cluster, 0, chunk);
            }
        }
    }

    /// 由长文件名生成短文件名
    ///
    /// 只支持 ASCII
    pub fn generate_short_name(&self, long_name: &str) -> String {
        let mut res = String::new();
        // 转换为大写
        let upper_name = long_name.to_uppercase();
        // 去除空格
        let mut long_name = String::new();
        for c in upper_name.chars() {
            if c != ' ' {
                long_name.push(c);
            }
        }
        // 去除前导点
        let long_name = long_name.trim_start_matches(|c| c == '.');
        // 分配数字结尾串
        let numeric_tail = self.inner_exclusive_access().numeric_allocator.alloc();
        let split = long_name.split('.').collect::<Vec<_>>();
        // 构造文件名主要部分
        for c in split[0].as_bytes() {
            if res.len() >= 8 - numeric_tail.len() {
                break;
            }
            res.push(*c as char);
        }
        res.push_str(&numeric_tail);
        // 构造文件名扩展部分
        if split.len() > 1 {
            res.push('.');
            let mut i = 0;
            for c in split[split.len() - 1].as_bytes() {
                if i >= 3 {
                    break;
                }
                res.push(*c as char);
                i += 1;
            }
        }
        res
    }
}

/// 该 impl 块实现簇分配相关操作
impl Fat32SuperBlock {
    fn get_next_cluster(&self, cluster: u32) -> Option<u32> {
        let fat_off = cluster as usize * 4;
        let sect =
            self.bpb.reserved_sectors as usize + (fat_off / self.bpb.bytes_per_sector as usize);
        let sect_off = fat_off % self.bpb.bytes_per_sector as usize;
        get_block_cache(self.start_sect + sect)
            .lock()
            .read(sect_off, |val: &u32| {
                let val = val & 0x0FFF_FFFF;
                if val >= BOUND_CLUSTER || val == BAD_CLUSTER {
                    None
                } else {
                    Some(val)
                }
            })
    }

    fn set_next_cluster(&self, cluster: u32, next_cluster: u32) {
        let fat_off = cluster as usize * 4;
        let sect =
            self.bpb.reserved_sectors as usize + (fat_off / self.bpb.bytes_per_sector as usize);
        let sect_off = fat_off % self.bpb.bytes_per_sector as usize;
        get_block_cache(self.start_sect + sect)
            .lock()
            .modify(sect_off, |val: &mut u32| {
                *val &= 0xF000_0000;
                *val |= next_cluster;
            });
    }

    pub fn chain_of_cluster(&self, mut cluster: u32) -> Vec<u32> {
        if cluster == 0 {
            return Vec::new();
        }
        let mut res = vec![cluster];
        while let Some(next_cluster) = self.get_next_cluster(cluster) {
            res.push(next_cluster);
            cluster = next_cluster;
        }
        res
    }

    /// 分配一个簇的空间
    pub fn alloc_cluster(&self) -> Option<u32> {
        let mut inner = self.inner_exclusive_access();
        if inner.fsinfo.free_count > 0 {
            let mut cluster = if inner.fsinfo.last_alloc_cluster == 0xFFFF_FFFF {
                3
            } else {
                inner.fsinfo.last_alloc_cluster + 1
            };
            loop {
                let fat_off = cluster as usize * 4;
                let sect = self.bpb.reserved_sectors as usize
                    + (fat_off / self.bpb.bytes_per_sector as usize);
                let sect_off = fat_off % self.bpb.bytes_per_sector as usize;
                let flag = get_block_cache(self.start_sect + sect).lock().modify(
                    sect_off,
                    |val: &mut u32| {
                        let v = *val & 0x0FFF_FFFF;
                        if v == FREE_CLUSTER {
                            *val &= 0xF000_0000;
                            *val |= END_CLUSTER;
                            true
                        } else {
                            false
                        }
                    },
                );
                if flag {
                    inner.fsinfo.free_count -= 1;
                    inner.fsinfo.last_alloc_cluster = cluster;
                    return Some(cluster);
                } else {
                    cluster += 1;
                }
            }
        } else {
            None
        }
    }

    /// 释放一个簇
    pub fn dealloc_cluster(&self, cluster: u32) {
        let mut inner = self.inner_exclusive_access();
        let fat_off = cluster as usize * 4;
        let sect =
            self.bpb.reserved_sectors as usize + (fat_off / self.bpb.bytes_per_sector as usize);
        let sect_off = fat_off % self.bpb.bytes_per_sector as usize;
        get_block_cache(self.start_sect + sect)
            .lock()
            .modify(sect_off, |val: &mut u32| {
                *val &= 0xF000_0000;
            });
        inner.fsinfo.free_count += 1;
        if inner.fsinfo.last_alloc_cluster == cluster {
            inner.fsinfo.last_alloc_cluster -= 1;
        }
    }

    /// 释放簇链
    pub fn dealloc_chain_of_cluster(&self, first_cluster: u32) {
        let chain = self.chain_of_cluster(first_cluster);
        for cluster in chain.iter().rev() {
            self.dealloc_cluster(*cluster);
        }
    }

    /// 从首簇号 `first_cluster` 开始的簇链中分配 `num` 个目录项的空间
    pub fn alloc_dentry(&self, first_cluster: u32, num: usize) -> Option<usize> {
        let mut cluster_off = 0;
        let vec = self.chain_of_cluster(first_cluster);
        let last_cluster = *vec.last().unwrap();
        let bytes_per_cluster = self.bpb.bytes_per_cluster();
        let mut tmp = ShortDirEntry::empty();
        while cluster_off < bytes_per_cluster {
            self.read_slice_from_cluster(last_cluster, cluster_off, tmp.as_bytes_mut());
            if tmp.is_free() {
                let res = (vec.len() - 1) * bytes_per_cluster + cluster_off;
                if cluster_off + num * DENTRY_SZ > bytes_per_cluster {
                    return self.alloc_cluster().map(|next_cluster| {
                        self.set_next_cluster(last_cluster, next_cluster);
                        res
                    });
                } else {
                    return Some(res);
                }
            }
            cluster_off += DENTRY_SZ;
        }
        self.alloc_cluster().map(|next_cluster| {
            self.set_next_cluster(last_cluster, next_cluster);
            vec.len() * bytes_per_cluster
        })
    }

    /// 扩充簇链空间，返回簇链的首簇号
    pub fn increase_size(&self, mut first_cluster: u32, new_size: usize) -> Option<u32> {
        if new_size == 0 {
            return Some(first_cluster);
        }
        // 分配首簇
        if first_cluster == 0 {
            if let Some(cluster) = self.alloc_cluster() {
                first_cluster = cluster;
            } else {
                return None;
            }
        }
        // 分配剩余空间
        let mut cluster_chain = self.chain_of_cluster(first_cluster);
        let num = (new_size + self.bytes_per_cluster() - 1) / self.bytes_per_cluster()
            - cluster_chain.len();
        if num as isize <= 0 {
            return Some(first_cluster);
        }
        for _ in 0..num {
            if let Some(cluster) = self.alloc_cluster() {
                self.set_next_cluster(*cluster_chain.last().unwrap(), cluster);
                cluster_chain.push(cluster);
            } else {
                return None;
            }
        }
        Some(first_cluster)
    }
}

impl SuperBlock for Fat32SuperBlock {
    fn get_root_inode(&self) -> Arc<dyn Inode> {
        get_root_inode(self).unwrap()
    }

    fn sync(&self) {
        inode_sync_all();
        let inner = self.inner_exclusive_access();
        get_block_cache(self.start_sect + self.bpb.fsinfo_sector as usize)
            .lock()
            .modify(0, |block_cache: &mut [u8; 512]| unsafe {
                *(block_cache.as_mut_ptr().add(0x1E8) as *mut u32) = inner.fsinfo.free_count;
                *(block_cache.as_mut_ptr().add(0x1EC) as *mut u32) =
                    inner.fsinfo.last_alloc_cluster;
            });
    }
}

impl Drop for Fat32SuperBlock {
    fn drop(&mut self) {
        let inner = self.inner_exclusive_access();
        get_block_cache(self.start_sect + self.bpb.fsinfo_sector as usize)
            .lock()
            .modify(0, |block_cache: &mut [u8; 512]| unsafe {
                *(block_cache.as_mut_ptr().add(0x1E8) as *mut u32) = inner.fsinfo.free_count;
                *(block_cache.as_mut_ptr().add(0x1EC) as *mut u32) =
                    inner.fsinfo.last_alloc_cluster;
            });
    }
}

/// 从扇区号 `sect` 偏移量 `off` 处读出内容到 `buf`
fn read_slice_from_block_cache(sect: usize, off: usize, buf: &mut [u8]) {
    const BLOCK_SIZE: usize = 512;
    get_block_cache(sect)
        .lock()
        .read(0, |block_cache: &[u8; BLOCK_SIZE]| {
            buf.copy_from_slice(&block_cache[off..off + buf.len()]);
        })
}

/// 将 `buf` 内容写到扇区号 `sect` 偏移量 `off` 处
fn write_slice_to_block_cache(sect: usize, off: usize, buf: &[u8]) {
    const BLOCK_SIZE: usize = 512;
    get_block_cache(sect)
        .lock()
        .modify(0, |block_cache: &mut [u8; BLOCK_SIZE]| {
            block_cache[off..off + buf.len()].copy_from_slice(&buf);
        })
}

struct NumericAllocator {
    num: usize,
}

impl NumericAllocator {
    pub fn new() -> Self {
        NumericAllocator { num: 0 }
    }
    pub fn alloc(&mut self) -> String {
        self.num += 1;
        String::from(String::from("~") + &self.num.to_string())
    }
}
