use super::{types::*, File, Inode, ModeFlags, OpenFlags};
use crate::{
    config::{DEFAULT_FD_LIMIT, SYSTEM_FD_LIMIT},
    mm::UserBuffer,
    syscall::errno::*,
};
use alloc::{sync::Arc, vec::Vec};
use spin::{Mutex, MutexGuard};

/// 普通文件对象
pub struct OSFile {
    pub inode: Arc<dyn Inode>,
    pub mode: ModeFlags,
    inner: Mutex<OSFileInner>,
}

pub struct OSFileInner {
    flags: OpenFlags,
    offset: usize,
}

impl OSFile {
    pub fn inner_exclusive_access(&self) -> MutexGuard<'_, OSFileInner> {
        self.inner.lock()
    }

    pub fn new(inode: Arc<dyn Inode>, flags: OpenFlags) -> Arc<OSFile> {
        let mut mode = ModeFlags::EXEC;
        match inode.get_attr().get_mode().get_type() {
            ModeFlags::DIRECTORY => mode |= ModeFlags::DIRECTORY,
            ModeFlags::REGULAR => mode |= ModeFlags::REGULAR,
            _ => panic!("unknown file type"),
        }
        // flag 为 0 任何情况都成立，且不当作一个标志位
        if flags.contains(OpenFlags::O_RDWR) {
            mode |= ModeFlags::READ | ModeFlags::WRITE;
        } else if flags.contains(OpenFlags::O_WRONLY) {
            mode |= ModeFlags::WRITE;
        } else {
            mode |= ModeFlags::READ;
        }
        Arc::new(OSFile {
            inode,
            mode,
            inner: Mutex::new(OSFileInner { flags, offset: 0 }),
        })
    }
}

impl File for OSFile {
    fn get_mode(&self) -> ModeFlags {
        self.mode
    }

    fn get_flags(&self) -> OpenFlags {
        self.inner_exclusive_access().flags
    }

    fn set_flags(&self, flags: OpenFlags) {
        let mut inner = self.inner_exclusive_access();
        inner
            .flags
            .remove(OpenFlags::O_APPEND | OpenFlags::O_NONBLOCK);
        inner.flags.insert(flags);
        drop(inner);
        if flags.contains(OpenFlags::O_APPEND) {
            self.lseek(0, SEEK_END).unwrap();
        }
    }

    fn get_inode(&self) -> Arc<dyn Inode> {
        Arc::clone(&self.inode)
    }

    fn read(&self, buf: UserBuffer) -> usize {
        assert!(self.mode.contains(ModeFlags::READ));
        let mut inner = self.inner_exclusive_access();
        let mut total_read_size = 0;
        for slice in buf.buffers {
            let read_size = self.inode.read(inner.offset, slice);
            if read_size == 0 {
                break;
            }
            inner.offset += read_size;
            total_read_size += read_size;
        }
        total_read_size
    }

    fn write(&self, buf: UserBuffer) -> usize {
        assert!(self.mode.contains(ModeFlags::WRITE));
        let mut inner = self.inner_exclusive_access();
        let mut total_write_size = 0;
        for slice in buf.buffers {
            let write_size = self.inode.write(inner.offset, slice);
            assert_eq!(write_size, slice.len());
            inner.offset += write_size;
            total_write_size += write_size;
        }
        total_write_size
    }

    fn read_all(&self) -> Vec<u8> {
        assert!(self.mode.contains(ModeFlags::READ));
        let mut inner = self.inner_exclusive_access();
        let mut buffer = [0u8; 512];
        let mut res: Vec<u8> = Vec::new();
        loop {
            let len = self.inode.read(inner.offset, &mut buffer);
            if len == 0 {
                break;
            }
            inner.offset += len;
            res.extend_from_slice(&buffer[..len]);
        }
        res
    }

    fn lseek(&self, offset: isize, whence: usize) -> Result<usize, SyscallError> {
        let mut inner = self.inner_exclusive_access();
        let new_offset = match whence {
            SEEK_SET => offset,
            SEEK_CUR => inner.offset as isize + offset,
            SEEK_END => self.inode.get_attr().st_size as isize + offset,
            _ => {
                panic!("in lseek: unsupport operation, whence = {}", whence);
            }
        };
        if new_offset < 0 {
            return Err(SyscallError::EINVAL);
        }
        inner.offset = new_offset as usize;
        Ok(new_offset as usize)
    }

    fn ready_read(&self) -> bool {
        true
    }

    fn ready_write(&self) -> bool {
        true
    }
}

/// 打开文件表
pub struct FileTable {
    inner: Mutex<FileTableInner>,
}

pub struct FileTableInner {
    close_on_exec: Vec<bool>,
    table: Vec<Option<Arc<dyn File>>>,
    soft_limit: usize,
    hard_limit: usize,
}

impl FileTable {
    pub fn inner_exclusive_access(&self) -> MutexGuard<'_, FileTableInner> {
        self.inner.lock()
    }

    pub fn new() -> Self {
        Self {
            inner: Mutex::new(FileTableInner {
                close_on_exec: Vec::new(),
                table: Vec::new(),
                soft_limit: DEFAULT_FD_LIMIT,
                hard_limit: SYSTEM_FD_LIMIT,
            }),
        }
    }

    /// 获取文件描述符对应的文件
    pub fn get(&self, fd: usize) -> Option<Arc<dyn File>> {
        let inner = self.inner_exclusive_access();
        inner
            .table
            .get(fd)
            .map(|f| f.as_ref())
            .flatten()
            .map(|f| f.clone())
    }

    /// 获取文件描述符标志
    pub fn get_flag(&self, fd: usize) -> Option<u32> {
        let inner = self.inner_exclusive_access();
        inner
            .close_on_exec
            .get(fd)
            .map(|flag| if *flag { 1 } else { 0 })
    }

    /// 分配文件描述符
    ///
    /// 使用最小未被使用的编号作为新的文件描述符
    pub fn alloc_fd(&self) -> usize {
        let mut inner = self.inner_exclusive_access();

        if let Some(fd) = (0..inner.table.len()).find(|fd| inner.table[*fd].is_none()) {
            fd
        } else {
            inner.close_on_exec.push(false);
            inner.table.push(None);
            inner.table.len() - 1
        }
    }

    /// 将文件描述符与文件绑定
    pub fn set_fd(&self, fd: usize, file: Arc<dyn File>, close_on_exec: bool) {
        let mut inner = self.inner_exclusive_access();
        while fd + 1 > inner.table.len() {
            inner.close_on_exec.push(false);
            inner.table.push(None);
        }
        inner.close_on_exec[fd] = close_on_exec;
        inner.table[fd] = Some(file);
    }

    pub fn try_set_fd(
        &self,
        fd: usize,
        file: Arc<dyn File>,
        close_on_exec: bool,
    ) -> Result<(), SyscallError> {
        let mut inner = self.inner_exclusive_access();

        if fd + 1 >= inner.soft_limit {
            return Err(SyscallError::EMFILE);
        }
        while fd + 1 > inner.table.len() {
            inner.close_on_exec.push(false);
            inner.table.push(None);
        }
        inner.close_on_exec[fd] = close_on_exec;
        inner.table[fd] = Some(file);

        Ok(())
    }

    pub fn try_alloc_fd(&self) -> Result<usize, crate::vfs::SyscallError> {
        let mut inner = self.inner_exclusive_access();
        if inner.table.len() >= inner.soft_limit {
            return Err(crate::vfs::SyscallError::EMFILE);
        }
        if let Some(fd) = (0..inner.table.len()).find(|fd| inner.table[*fd].is_none()) {
            Ok(fd)
        } else {
            inner.close_on_exec.push(false);
            inner.table.push(None);
            Ok(inner.table.len() - 1)
        }
    }

    /// 设置文件描述符标志
    pub fn set_flag(&self, fd: usize, close_on_exec: bool) {
        let mut inner = self.inner_exclusive_access();
        inner
            .close_on_exec
            .get_mut(fd)
            .map(|flag| *flag = close_on_exec);
    }

    /// 删除文件描述符
    ///
    /// # Panics
    /// `fd` 超过打开文件表大小
    pub fn remove(&self, fd: usize) -> Option<Arc<dyn File>> {
        let mut inner = self.inner_exclusive_access();
        inner.close_on_exec[fd] = false;
        inner.table[fd].take()
    }

    pub fn clone(&self) -> Self {
        let inner = self.inner_exclusive_access();
        Self {
            inner: Mutex::new(FileTableInner {
                close_on_exec: Vec::clone(&inner.close_on_exec),
                table: Vec::clone(&inner.table),
                soft_limit: inner.soft_limit,
                hard_limit: inner.hard_limit,
            }),
        }
    }

    pub fn get_hard_limit(&self) -> usize {
        let inner = self.inner_exclusive_access();
        let hard_limit = inner.hard_limit;
        drop(inner);
        hard_limit
    }

    pub fn get_soft_limit(&self) -> usize {
        let inner = self.inner_exclusive_access();
        let soft_limit = inner.soft_limit;
        drop(inner);
        soft_limit
    }

    pub fn set_limit(&self, soft_limit: usize, hard_limit: usize) {
        let mut inner = self.inner_exclusive_access();
        inner.soft_limit = soft_limit;
        inner.hard_limit = hard_limit;
        drop(inner);
    }

    pub fn clear(&self) {
        let mut inner = self.inner_exclusive_access();
        inner.close_on_exec.clear();
        inner.table.clear();
    }

    /// 关闭设置了 `FD_CLOEXEC` 标志的文件
    pub fn close_on_exec(&self) {
        let mut inner = self.inner_exclusive_access();
        for fd in 0..inner.table.len() {
            if inner.close_on_exec[fd] {
                inner.close_on_exec[fd] = false;
                inner.table[fd].take();
            }
        }
    }
}
