use super::{types::*, File};
use crate::{
    config::PAGE_SIZE,
    mm::{frame_alloc, FrameTracker, UserBuffer},
    syscall::errno::SyscallError,
};
use alloc::{string::String, vec::Vec};
use spin::{Mutex, MutexGuard};

pub struct Memfile {
    name: String,
    inner: Mutex<MemfileInner>,
}

pub struct MemfileInner {
    size: usize,
    offset: usize,
    pages: Vec<FrameTracker>,
}

impl MemfileInner {
    fn increase_size(&mut self, new_size: usize) {
        if self.size >= new_size {
            return;
        }
        let max_size = self.pages.len() * PAGE_SIZE;
        if max_size < new_size {
            for _ in 0..((new_size - max_size + PAGE_SIZE - 1) / PAGE_SIZE) {
                self.pages.push(frame_alloc().unwrap());
            }
        }
        self.size = new_size;
    }
}

impl Memfile {
    pub fn new(name: String) -> Self {
        Self {
            name,
            inner: Mutex::new(MemfileInner {
                size: 0,
                offset: 0,
                pages: Vec::new(),
            }),
        }
    }

    pub fn inner_exclusive_access(&self) -> MutexGuard<'_, MemfileInner> {
        self.inner.lock()
    }

    pub fn read_bytes(&self, buf: &mut [u8]) -> usize {
        // UserBuffer 的每个 buffer 肯定不超过一页
        assert!(buf.len() <= PAGE_SIZE);
        let mut inner = self.inner_exclusive_access();
        let read_size = (inner.size - inner.offset).min(buf.len());
        let (id, off) = (inner.offset / PAGE_SIZE, inner.offset % PAGE_SIZE);
        // 读取数据在内核中以超过一页方式进行存储（最多两页）
        if off + read_size > PAGE_SIZE {
            let size1 = PAGE_SIZE - off;
            let size2 = read_size - size1;
            buf[..size1].copy_from_slice(&inner.pages[id].ppn.get_bytes_array()[off..]);
            buf[size1..read_size]
                .copy_from_slice(&inner.pages[id + 1].ppn.get_bytes_array()[..size2]);
        } else {
            buf[..read_size]
                .copy_from_slice(&inner.pages[id].ppn.get_bytes_array()[off..(off + read_size)]);
        }
        inner.offset += read_size;
        read_size
    }

    pub fn write_bytes(&self, buf: &[u8]) -> usize {
        let mut inner = self.inner_exclusive_access();
        let new_size = inner.offset + buf.len();
        inner.increase_size(new_size);
        let (id, off) = (inner.offset / PAGE_SIZE, inner.offset % PAGE_SIZE);
        // 写入数据在内核中以超过一页方式进行存储（最多两页）
        if off + buf.len() > PAGE_SIZE {
            let size1 = PAGE_SIZE - off;
            let size2 = buf.len() - size1;
            inner.pages[id].ppn.get_bytes_array()[off..].copy_from_slice(&buf[..size1]);
            inner.pages[id + 1].ppn.get_bytes_array()[..size2].copy_from_slice(&buf[size1..]);
        } else {
            inner.pages[id].ppn.get_bytes_array()[off..(off + buf.len())].copy_from_slice(buf);
        }
        inner.offset += buf.len();
        buf.len()
    }

    pub fn truncate(&self) {
        let mut inner = self.inner_exclusive_access();
        inner.size = 0;
        inner.offset = 0;
    }
}

impl File for Memfile {
    fn read(&self, buf: UserBuffer) -> usize {
        let mut total_read_size = 0;
        for slice in buf.buffers {
            let read_size = self.read_bytes(slice);
            if read_size == 0 {
                break;
            }
            total_read_size += read_size;
        }
        total_read_size
    }

    fn write(&self, buf: UserBuffer) -> usize {
        let mut total_write_size = 0;
        for slice in buf.buffers {
            let write_size = self.write_bytes(slice);
            assert_eq!(write_size, slice.len());
            total_write_size += write_size;
        }
        total_write_size
    }

    fn lseek(&self, offset: isize, whence: usize) -> Result<usize, SyscallError> {
        let mut inner = self.inner_exclusive_access();
        let new_offset = match whence {
            SEEK_SET => offset,
            SEEK_CUR => inner.offset as isize + offset,
            SEEK_END => inner.size as isize + offset,
            _ => {
                panic!(
                    "[in memfile.lseek()] unsupport operation, whence = {}",
                    whence
                );
            }
        };
        if new_offset < 0 {
            return Err(SyscallError::EINVAL);
        }
        if new_offset as usize > inner.size {
            inner.increase_size(new_offset as usize);
        }
        inner.offset = new_offset as usize;
        Ok(new_offset as usize)
    }
}
