#![allow(unused)]

mod devtmpfs;
mod fat32;
mod file;
mod info;
mod memfile;
mod mount;
mod open;
mod pipe;
mod procfs;
mod stdio;
pub mod types;

use self::procfs::PROCFS_ROOT_INODE;
use self::types::*;
use crate::drivers::{block_cache_sync_all, get_block_cache};
use crate::mm::{translated_byte_buffer, translated_ref, UserBuffer};
use crate::syscall::errno::SyscallError;
use crate::task::current_user_token;
use alloc::string::String;
use alloc::vec;
use alloc::{sync::Arc, vec::Vec};
use core::convert::TryInto;
use devtmpfs::DEV_ROOT_INODE;
use lazy_static::lazy_static;

pub use fat32::Fat32SuperBlock;
pub use file::{FileTable, OSFile};
pub use info::FSInfo;
pub use memfile::Memfile;
pub use mount::*;
pub use open::*;
pub use pipe::{make_pipe, Pipe};
pub use stdio::{Stdin, Stdout};

pub trait SuperBlock: Send + Sync {
    fn get_root_inode(&self) -> Arc<dyn Inode>;
    fn sync(&self);
}
pub trait Inode: Send + Sync {
    fn get_name(&self) -> String {
        unimplemented!("get_name");
    }
    /// 获取文件属性
    fn get_attr(&self) -> Kstat;

    /// 在当前目录下创建普通文件
    fn create(&self, name: &str, mode: ModeFlags) -> Option<Arc<dyn Inode>> {
        unimplemented!("create");
    }
    /// 在当前目录下创建目录文件
    fn mkdir(&self, name: &str, mode: ModeFlags) {
        unimplemented!("mkdir");
    }
    /// 在当前目录下查找文件
    fn lookup(&self, name: &str) -> Option<Arc<dyn Inode>> {
        unimplemented!("lookup");
    }
    fn read(&self, off: usize, buf: &mut [u8]) -> usize {
        unimplemented!("read");
    }
    fn write(&self, off: usize, buf: &[u8]) -> usize {
        unimplemented!("write");
    }
    fn truncate(&self) {
        unimplemented!("truncate");
    }
    /// 读取目录项
    fn read_dentry(&self, off: usize) -> Option<Dirent> {
        unimplemented!("read_dentry");
    }
    fn unlink(&self, name: &str) {
        unimplemented!("unlink");
    }
    /// 将文件目录下的 `old_name` 文件移动到 `new_dir` 目录下，新文件名为 `new_name`
    fn rename(&self, old_name: &str, new_dir: Arc<dyn Inode>, new_name: &str) {
        unimplemented!("rename");
    }
    /// 设置时间戳
    fn set_timestamps(
        &self,
        st_atime_sec: Option<i64>,
        st_atime_nsec: Option<i64>,
        st_mtime_sec: Option<i64>,
        st_mtime_nsec: Option<i64>,
    ) {
        unimplemented!("set_timestamps");
    }
}
pub trait File: Send + Sync {
    fn get_mode(&self) -> ModeFlags {
        unimplemented!("get_mode");
    }
    fn get_flags(&self) -> OpenFlags {
        unimplemented!("get_flags");
    }
    /// 设置文件状态标志位
    /// 
    /// 只有 `O_APPEND`、`O_ASYNC`、`O_DIRECT`、`O_NOATIME` 和 `O_NONBLOCK` 可以设置
    fn set_flags(&self, flags: OpenFlags) {
        unimplemented!("get_flags");
    }
    /// 获取文件对应 inode 节点
    fn get_inode(&self) -> Arc<dyn Inode> {
        unimplemented!("get_inode");
    }

    fn read(&self, buf: UserBuffer) -> usize;
    fn write(&self, buf: UserBuffer) -> usize;
    fn readv(&self, iov: *const IOVec, iovcnt: usize) -> usize {
        let token = current_user_token();
        let mut size = 0;
        for i in 0..iovcnt {
            let iovec = translated_ref(token, unsafe { iov.add(i) });
            let buf =
                translated_byte_buffer(token, iovec.iov_base as *const u8, iovec.iov_len as usize);
            size += self.read(UserBuffer::new(buf));
        }
        size
    }
    fn writev(&self, iov: *const IOVec, iovcnt: usize) -> usize {
        let token = current_user_token();
        let mut size = 0;
        for i in 0..iovcnt {
            let iovec = translated_ref(token, unsafe { iov.add(i) });
            let buf =
                translated_byte_buffer(token, iovec.iov_base as *const u8, iovec.iov_len as usize);
            size += self.write(UserBuffer::new(buf));
        }
        size
    }
    fn read_all(&self) -> Vec<u8> {
        unimplemented!("read_all");
    }
    #[allow(unused)]
    fn lseek(&self, offset: isize, whence: usize) -> Result<usize, SyscallError> {
        unimplemented!("lseek");
        //Ok(0)
    }
    fn ready_read(&self) -> bool {
        unimplemented!("ready_read")
    }
    fn ready_write(&self) -> bool {
        unimplemented!("ready_write")
    }
}

bitflags! {
    /// 文件标志位，包含文件的权限以及文件类型
    ///
    /// 在判断文件类型的时候，请使用 `get_type()` 函数并通过 `match` 匹配判断
    pub struct ModeFlags: u32 {
        // 权限
        const READ = 0o444;
        const WRITE = 0o222;
        const EXEC = 0o111;
        const RW = 0o600;
        const RWX = Self::READ.bits | Self::WRITE.bits | Self::EXEC.bits;

        // 文件类型
        // musl libc: include/sys/stat.h
        const SOCKET = 0o140000;
        const LINK = 0o120000;
        const REGULAR = 0o100000;
        const BLK = 0o60000;
        const DIRECTORY = 0o40000;
        const CHAR = 0o20000;
        const FIFO = 0o10000;
    }
}

impl ModeFlags {
    /// 获取文件类型
    ///
    /// 由于相关的标志位是根据实际库定义的，因此并不符合各个位互斥，
    /// 也即可能造成一种文件类型它的二进制其实是包含另一种文件类型，
    /// 导致判断失误。所以使用该函数之后通过 `match` 进行全匹配判断，
    /// 而不是使用 `contains()`
    pub fn get_type(&self) -> Self {
        const S_IFMT: u32 = 0o170000;
        let file_type = self.bits() & S_IFMT;
        assert!(file_type != 0);
        Self::from_bits(self.bits() & S_IFMT).unwrap()
    }
}

#[derive(Debug)]
pub struct PartitionTable {
    pub sys_ind: u8,
    pub start_sect: u32,
    pub nr_sects: u32,
}
impl PartitionTable {
    fn new(partition_table: &[u8; 16]) -> Self {
        let sys_ind = partition_table[0x04];
        let start_sect = u32::from_le_bytes(partition_table[0x08..0x0c].try_into().unwrap());
        let nr_sects = u32::from_le_bytes(partition_table[0x0c..0x10].try_into().unwrap());
        PartitionTable {
            sys_ind,
            start_sect,
            nr_sects,
        }
    }
}

lazy_static! {
    pub static ref SUPER_BLOCKS: Vec<Arc<dyn SuperBlock>> = {
        // sdcard 的第一个扇区可能是 MBR 也可能是 DBR，
        // 如果是 DBR，此时读到的分区起始扇区号是 0，不会影响后续操作
        let mbr = get_block_cache(0);
        let partition_table1 = mbr.lock().read(0x1BE, PartitionTable::new);
        println!("[vfs]: partition table1 = {:#x?}", partition_table1);
        vec![Fat32SuperBlock::new(partition_table1)]
    };
    pub static ref ROOT_INODE: Arc<dyn Inode> = SUPER_BLOCKS[0].get_root_inode();
}

pub fn root_inode() -> Arc<dyn Inode> {
    Arc::clone(&ROOT_INODE)
}

pub fn procfs_root_inode() -> Arc<dyn Inode> {
    Arc::clone(&PROCFS_ROOT_INODE)
}

pub fn dev_root_inode() -> Arc<dyn Inode> {
    Arc::clone(&DEV_ROOT_INODE)
}

pub fn init() {
    let _ = root_inode();
}

pub fn sync() {
    for sb in SUPER_BLOCKS.iter() {
        sb.sync();
    }
    block_cache_sync_all();
}

pub fn add_initproc_shell() {
    extern "C" {
        fn _num_app();
    }
    let num_app_ptr = _num_app as usize as *mut usize;
    let app_start = unsafe { core::slice::from_raw_parts_mut(num_app_ptr.add(1), 3) };

    if let Ok(file) = open_file(
        root_inode(),
        "initproc",
        OpenFlags::O_RDWR | OpenFlags::O_CREAT,
        ModeFlags::RWX | ModeFlags::REGULAR,
    ) {
        println!("Create initproc ");
        let mut data: Vec<&'static mut [u8]> = Vec::new();
        data.push(unsafe {
            core::slice::from_raw_parts_mut(app_start[0] as *mut u8, app_start[1] - app_start[0])
        });
        println!("Start write initproc ");
        file.write(UserBuffer::new(data));
    } else {
        panic!("initproc create fail!");
    }

    if let Ok(file) = open_file(
        root_inode(),
        "user_shell",
        OpenFlags::O_RDWR | OpenFlags::O_CREAT,
        ModeFlags::RWX | ModeFlags::REGULAR,
    ) {
        println!("Create user_shell ");
        let mut data: Vec<&'static mut [u8]> = Vec::new();
        data.push(unsafe {
            core::slice::from_raw_parts_mut(app_start[1] as *mut u8, app_start[2] - app_start[1])
        });
        println!("Start write user_shell ");
        file.write(UserBuffer::new(data));
    } else {
        panic!("user_shell create fail!");
    }
    println!("Write apps(initproc & user_shell) to disk from mem");
}
