use super::Inode;
use crate::syscall::errno::SyscallError;
use alloc::{string::String, sync::Arc, vec::Vec};
use lazy_static::*;
use spin::Mutex;

pub struct Mount {
    pub filesystemtype: String,
    /// 挂载文件系统的根目录
    pub root: Arc<dyn Inode>,
    /// 挂载到的节点
    pub inode: Arc<dyn Inode>,
}

impl Mount {
    pub fn new(filesystemtype: String, root: Arc<dyn Inode>, inode: Arc<dyn Inode>) -> Self {
        Self {
            filesystemtype,
            root,
            inode,
        }
    }
}

lazy_static! {
    pub static ref MOUNTS: Mutex<Vec<Mount>> = { Mutex::new(Vec::new()) };
}

pub fn insert_mounts(mount: Mount) {
    MOUNTS.lock().push(mount);
}

pub fn remove_mounts(inode: Arc<dyn Inode>) -> Result<usize, SyscallError> {
    let mut mounts: spin::MutexGuard<'_, Vec<Mount>> = MOUNTS.lock();
    let len = mounts.len();
    mounts.retain(|mount| Arc::as_ptr(&mount.inode) != Arc::as_ptr(&inode));
    if len == mounts.len() {
        Err(SyscallError::EINVAL)
    } else {
        Ok(0)
    }
}
