use super::{
    dev_root_inode,
    devtmpfs::{Null, Rtc, Zero},
    root_inode,
    types::*,
    File, Inode, ModeFlags, OSFile, MOUNTS,
};
use crate::{syscall::errno::*, task::current_task};
use alloc::{sync::Arc, vec::Vec};

bitflags! {
    pub struct OpenFlags: u32 {
        // 必须包含其中之一
        const O_RDONLY = 0x00;
        const O_WRONLY = 0x01;
        const O_RDWR = 0x02;

        // 以下标志位中 Linux6.1 与 musl libc 定义完全不同
        // 具体来说，它们字面值是一样的，只是一个是十六进制，一个八进制
        // ???

        // file creation flags
        const O_CREAT = 0x40;
        const O_EXCL = 0x80;
        const O_TRUNC = 0x200;
        const O_DIRECTORY = 0x10000;
        const O_CLOEXEC = 0x80000;

        // file status flags
        const O_APPEND = 0x400;
        const O_NONBLOCK = 0x800;
        const O_LARGEFILE = 0x8000;
        const O_PATH = 0x200000;
    }
}

/// 查找 inode 是否为挂载点，若是返回挂载文件系统的根节点，
/// 不是则返回原 inode 节点
fn get_mount_inode(inode: Arc<dyn Inode>) -> Arc<dyn Inode> {
    let mounts = MOUNTS.lock();
    if let Some((id, _)) = mounts
        .iter()
        .enumerate()
        .find(|(_, mount)| Arc::as_ptr(&mount.inode) == Arc::as_ptr(&inode))
    {
        Arc::clone(&mounts[id].root)
    } else {
        inode
    }
}

/// 打开指定目录下的文件
///
/// 只处理文件创建标志
///
/// 如果创建文件，只创建普通文件，不创建目录文件
pub fn open_file(
    dir: Arc<dyn Inode>,
    name: &str,
    flags: OpenFlags,
    mode: ModeFlags,
) -> Result<Arc<dyn File>, SyscallError> {
    // /dev下的文件并不像普通的文件那样可以使用 OSFile::new() 构建
    // 这里直接手动查找并返回对应文件
    if Arc::as_ptr(&dir) == Arc::as_ptr(&dev_root_inode()) {
        match name {
            "zero" => return Ok(Arc::new(Zero)),
            "null" => return Ok(Arc::new(Null)),
            "rtc" => return Ok(Arc::new(Rtc)),
            _ => return Err(SyscallError::ENOENT),
        }
    }

    if dir.get_attr().get_mode().get_type() != ModeFlags::DIRECTORY {
        return Err(SyscallError::ENOTDIR);
    }
    let inode;
    if name == "" {
        inode = dir
    } else if let Some(_inode) = dir.lookup(name) {
        if flags.contains(OpenFlags::O_CREAT | OpenFlags::O_EXCL) {
            return Err(SyscallError::EEXIST);
        }
        inode = get_mount_inode(_inode);
    } else if flags.contains(OpenFlags::O_CREAT) {
        if let Some(_inode) = dir.create(name, mode) {
            inode = _inode;
        } else {
            // 这里错误的原因应该是硬盘空间不够，但找不到相应的错误码
            return Err(SyscallError::ENOTDIR);
        }
    } else {
        return Err(SyscallError::ENOENT);
    }

    let inode_mode = inode.get_attr().get_mode();
    if flags.contains(OpenFlags::O_DIRECTORY) && (inode_mode.get_type() != ModeFlags::DIRECTORY) {
        return Err(SyscallError::ENOTDIR);
    }
    if flags.contains(OpenFlags::O_TRUNC)
        && (inode_mode.get_type() == ModeFlags::REGULAR)
        && (flags.contains(OpenFlags::O_WRONLY) || flags.contains(OpenFlags::O_RDWR))
    {
        inode.truncate();
    }
    Ok(OSFile::new(inode, flags))
}

/// 根据路径进行搜索，直到达到最顶层目录
pub fn get_dir(dirfd: isize, path: &str) -> Result<Arc<dyn Inode>, SyscallError> {
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    let mut inode;
    if path.starts_with('/') {
        inode = root_inode();
    } else if dirfd == AT_FDCWD {
        let fs_info_inner = task_inner.fs_info.inner_exclusive_access();
        inode = Arc::clone(&fs_info_inner.cwd);
    } else {
        inode = if let Some(file) = task_inner.fd_table.get(dirfd as usize) {
            file.get_inode()
        } else {
            return Err(SyscallError::EBADF);
        }
    }
    drop(task_inner);
    inode = get_mount_inode(inode);

    let mut lookup_vec = path.trim_end_matches('/').split('/').collect::<Vec<_>>();
    lookup_vec.pop();
    for name in lookup_vec {
        if name.is_empty() {
            continue;
        }
        if inode.get_attr().get_mode().get_type() != ModeFlags::DIRECTORY {
            return Err(SyscallError::ENOTDIR);
        }
        inode = if let Some(i) = inode.lookup(name) {
            i
        } else {
            return Err(SyscallError::ENOENT);
        };
        inode = get_mount_inode(inode);
    }
    Ok(inode)
}

/// 根据路径打开文件
///
/// 只处理文件创建标志
///
/// 如果创建文件，只创建普通文件，不创建目录文件
pub fn open_file_at(
    dirfd: isize,
    path: &str,
    flags: OpenFlags,
    mode: ModeFlags,
) -> Result<Arc<dyn File>, SyscallError> {
    get_dir(dirfd, path)
        .map(|parent| {
            let name = path
                .trim_end_matches('/')
                .split('/')
                .collect::<Vec<_>>()
                .pop()
                .unwrap();
            open_file(parent, name, flags, mode)
        })
        .flatten()
}
