use crate::{
    mm::{get_meminfo, UserBuffer},
    task::current_task,
    vfs::{types::*, File, Inode, Memfile, ModeFlags},
};
use alloc::{format, string::ToString};
use alloc::{string::String, sync::Arc, vec::Vec};
use lazy_static::lazy_static;
use spin::{Mutex, MutexGuard};

pub struct RootInode {
    buf: Memfile,
    inner: Mutex<RootInodeInner>,
}
pub struct RootInodeInner {
    // 下一个目录项的 inode 节点号
    inode: usize,
    // 存储下一个目录项的偏移
    offset: usize,
}
impl RootInode {
    pub fn new() -> RootInode {
        RootInode {
            buf: Memfile::new(String::from("proc")),
            inner: Mutex::new(RootInodeInner {
                inode: 1,
                offset: 0,
            }),
        }
    }

    pub fn inner_exclusive_access(&self) -> MutexGuard<'_, RootInodeInner> {
        self.inner.lock()
    }

    /// 存储目录项
    fn push(&self, name: String, mode: ModeFlags) {
        let mut inner = self.inner_exclusive_access();
        let mut dirent = Dirent::new(name, inner.inode as u64, mode, 0);
        inner.inode += 1;
        inner.offset += dirent.d_reclen as usize;
        dirent.d_off = inner.offset as i64;
        self.buf.write_bytes(dirent.as_bytes());
    }

    /// 刷新文件信息
    fn refresh(&self) {
        self.buf.truncate();
        {
            let mut inner = self.inner_exclusive_access();
            inner.inode = 1;
            inner.offset = 0;
        }
        // 刷新内核进程信息
        // let current = current_task().unwrap();
        // self.push(
        //     current.get_pid().to_string(),
        //     ModeFlags::DIRECTORY | ModeFlags::READ | ModeFlags::EXEC,
        // );

        // 刷新 mounts 文件
        self.push(String::from("mounts"), ModeFlags::REGULAR | ModeFlags::READ);

        // 刷新 meminfo 文件
        self.push(
            String::from("meminfo"),
            ModeFlags::REGULAR | ModeFlags::READ,
        );
    }
}
impl Inode for RootInode {
    fn get_attr(&self) -> Kstat {
        Kstat {
            st_dev: 0,
            st_ino: 0,
            st_mode: (ModeFlags::DIRECTORY | ModeFlags::RWX).bits(),
            st_nlink: 1,
            st_uid: 0,
            st_gid: 0,
            st_rdev: 0,
            __pad: 0,
            st_size: 4096,
            st_blksize: 512,
            __pad2: 0,
            st_blocks: 0,
            st_atime_sec: 0,
            st_atime_nsec: 0,
            st_mtime_sec: 0,
            st_mtime_nsec: 0,
            st_ctime_sec: 0,
            st_ctime_nsec: 0,
            __unused: [0; 2],
        }
    }

    fn lookup(&self, name: &str) -> Option<Arc<dyn Inode>> {
        match name {
            "mounts" => Some(Arc::clone(&PROCFS_MOUNT_INODE)),
            "meminfo" => Some(Arc::clone(&PROCFS_MEMINFO_INODE)),
            _ => None,
        }
    }

    fn read_dentry(&self, off: usize) -> Option<Dirent> {
        if off == 0 {
            self.refresh();
        }

        // 读取目录项长度
        self.buf.lseek(
            (off + core::mem::size_of::<u64>() + core::mem::size_of::<i64>()) as isize,
            SEEK_SET,
        );
        let mut bytes = [0u8; 2];
        self.buf.read_bytes(&mut bytes);
        let len = u16::from_le_bytes(bytes);
        if len == 0 {
            return None;
        }

        // 读取目录项
        self.buf.lseek(off as isize, SEEK_SET);
        let mut res = Dirent::empty();
        res.d_reclen = len;
        let read_size = self.buf.read_bytes(res.as_mut_bytes());
        assert_eq!(len, res.d_reclen);
        assert_eq!(len, read_size as u16);
        Some(res)
    }
}

pub struct MountInode;
impl Inode for MountInode {
    fn get_attr(&self) -> Kstat {
        Kstat {
            st_dev: 0,
            st_ino: 0,
            st_mode: (ModeFlags::REGULAR | ModeFlags::RWX).bits(),
            st_nlink: 1,
            st_uid: 0,
            st_gid: 0,
            st_rdev: 0,
            __pad: 0,
            st_size: 4096,
            st_blksize: 512,
            __pad2: 0,
            st_blocks: 0,
            st_atime_sec: 0,
            st_atime_nsec: 0,
            st_mtime_sec: 0,
            st_mtime_nsec: 0,
            st_ctime_sec: 0,
            st_ctime_nsec: 0,
            __unused: [0; 2],
        }
    }

    fn read(&self, off: usize, buf: &mut [u8]) -> usize {
        if off == 0 {
            let test = "proc /proc proc rw,nosuid,nodev,noexec,relatime 0 0\n\
            udev /dev devtmpfs rw,nosuid,noexec,relatime,size=4006308k,nr_inodes=1001577,mode=755,inode64 0 0\n".as_bytes();
            buf[..test.len()].copy_from_slice(test);
            test.len()
        } else {
            0
        }
    }
}

pub struct MeminfoInode {
    buf: Memfile,
}
impl MeminfoInode {
    pub fn new() -> Self {
        Self {
            buf: Memfile::new(String::from("meminfo_buf")),
        }
    }

    /// 刷新文件信息
    fn refresh(&self) {
        self.buf.truncate();
        let meminfo = get_meminfo();
        let mut formatted = String::new();
        for (key, value) in meminfo.iter() {
            let entry = format!("{}:\t{} kB\n", key, value);
            formatted.push_str(&entry);
        }
        self.buf.write_bytes(formatted.as_bytes());
    }
}
impl Inode for MeminfoInode {
    fn get_attr(&self) -> Kstat {
        Kstat {
            st_dev: 0,
            st_ino: 0,
            st_mode: (ModeFlags::REGULAR | ModeFlags::RWX).bits(),
            st_nlink: 1,
            st_uid: 0,
            st_gid: 0,
            st_rdev: 0,
            __pad: 0,
            st_size: 4096,
            st_blksize: 512,
            __pad2: 0,
            st_blocks: 0,
            st_atime_sec: 0,
            st_atime_nsec: 0,
            st_mtime_sec: 0,
            st_mtime_nsec: 0,
            st_ctime_sec: 0,
            st_ctime_nsec: 0,
            __unused: [0; 2],
        }
    }

    fn read(&self, off: usize, buf: &mut [u8]) -> usize {
        if off == 0 {
            self.refresh();
        }
        self.buf.lseek(off as isize, SEEK_SET);
        self.buf.read_bytes(buf)
    }
}

lazy_static! {
    pub static ref PROCFS_ROOT_INODE: Arc<dyn Inode> = Arc::new(RootInode::new());
    pub static ref PROCFS_MOUNT_INODE: Arc<dyn Inode> = Arc::new(MountInode);
    pub static ref PROCFS_MEMINFO_INODE: Arc<dyn Inode> = Arc::new(MeminfoInode::new());
}
