#![allow(unused)]

use super::ModeFlags;
use alloc::string::String;
// use core::ops::{BitAnd, BitOr, Not};

pub const DT_UNKNOWN: u8 = 0;
pub const DT_FIFO: u8 = 1;
pub const DT_CHR: u8 = 2;
pub const DT_DIR: u8 = 4;
pub const DT_BLK: u8 = 6;
pub const DT_REG: u8 = 8;
pub const DT_LNK: u8 = 10;
pub const DT_SOCK: u8 = 12;
pub const DT_WHT: u8 = 14;

pub const AT_FDCWD: isize = -100;

pub const SEEK_SET: usize = 0;
pub const SEEK_CUR: usize = 1;
pub const SEEK_END: usize = 2;

#[repr(C)]
#[derive(Debug)]
pub struct Dirent {
    pub d_ino: u64,
    pub d_off: i64,
    pub d_reclen: u16,
    pub d_type: u8,
    pub d_name: [u8; 256],
}
impl Dirent {
    pub fn empty() -> Self {
        Self {
            d_ino: 0,
            d_off: 0,
            d_reclen: 0,
            d_type: 0,
            d_name: [0; 256],
        }
    }

    pub fn new(mut name: String, inode: u64, mode: ModeFlags, offset: i64) -> Self {
        name += "\0";
        let mut len = name.len() + 19;
        let align = 8 - len % 8;
        len += align;
        for _ in 0..align {
            name.push('\0');
        }
        Self {
            d_ino: inode,
            d_off: offset,
            d_reclen: len as u16,
            d_type: match mode.get_type() {
                ModeFlags::SOCKET => DT_SOCK,
                ModeFlags::LINK => DT_LNK,
                ModeFlags::REGULAR => DT_REG,
                ModeFlags::BLK => DT_BLK,
                ModeFlags::DIRECTORY => DT_DIR,
                ModeFlags::CHAR => DT_CHR,
                ModeFlags::FIFO => DT_FIFO,
                _ => DT_UNKNOWN,
            },
            d_name: {
                let mut tmp: [u8; 256] = [0; 256];
                tmp[..name.len()].copy_from_slice(name.as_bytes());
                tmp
            },
        }
    }

    pub fn as_bytes(&self) -> &[u8] {
        let size = self.d_reclen as usize;
        // println!("in dirent: size = {}", size);
        unsafe { core::slice::from_raw_parts(self as *const _ as usize as *const u8, size) }
    }

    pub fn as_mut_bytes(&mut self) -> &mut [u8] {
        let size = self.d_reclen as usize;
        unsafe { core::slice::from_raw_parts_mut(self as *mut _ as usize as *mut u8, size) }
    }
}

#[repr(C)]
#[derive(Debug)]
pub struct Kstat {
    pub st_dev: u64,
    pub st_ino: u64,
    pub st_mode: u32,
    pub st_nlink: u32,
    pub st_uid: u32,
    pub st_gid: u32,
    pub st_rdev: u64,
    pub __pad: u64,
    pub st_size: i64,
    pub st_blksize: i32,
    pub __pad2: i32,
    pub st_blocks: i64,
    pub st_atime_sec: i64,
    pub st_atime_nsec: i64,
    pub st_mtime_sec: i64,
    pub st_mtime_nsec: i64,
    pub st_ctime_sec: i64,
    pub st_ctime_nsec: i64,
    pub __unused: [u32; 2],
}
impl Kstat {
    pub fn get_mode(&self) -> ModeFlags {
        ModeFlags::from_bits(self.st_mode).unwrap()
    }
}

#[repr(C)]
pub struct IOVec {
    pub iov_base: u64,
    pub iov_len: u64,
}

#[repr(C)]
pub struct Winsize {
    ws_row: u16,
    ws_col: u16,
    ws_xpixel: u16,
    ws_ypixel: u16,
}

impl Winsize {
    pub fn new() -> Self {
        Self {
            ws_row: 24,
            ws_col: 80,
            ws_xpixel: 0,
            ws_ypixel: 0,
        }
    }
}

#[repr(C)]
pub struct Statfs {
    pub f_type: u64,
    pub f_bsize: u64,
    pub f_blocks: u64,
    pub f_bfree: u64,
    pub f_bavail: u64,
    pub f_files: u64,
    pub f_ffree: u64,
    pub f_fsid: [i32; 2],
    pub f_namelen: u64,
    pub f_frsize: u64,
    pub f_flags: u64,
    pub f_spare: [u64; 4],
}

const FD_SETSIZE: usize = 1024;
const NFDBITS: usize = core::mem::size_of::<usize>() * 8;

#[repr(C)]
#[derive(Debug, Clone, Copy)]
pub struct FdSet {
    fds: [usize; FD_SETSIZE / NFDBITS],
}

impl FdSet {
    pub fn new() -> Self {
        Self {
            fds: [0; FD_SETSIZE / NFDBITS],
        }
    }

    pub fn set(&mut self, fd: usize) {
        if fd < FD_SETSIZE {
            let idx = fd / NFDBITS;
            let bit = fd % NFDBITS;
            self.fds[idx] |= 1 << bit;
        }
    }

    pub fn clear(&mut self, fd: usize) {
        if fd < FD_SETSIZE {
            let idx = fd / NFDBITS;
            let bit = fd % NFDBITS;
            self.fds[idx] &= !(1 << bit);
        }
    }

    pub fn is_set(&self, fd: usize) -> bool {
        if fd < FD_SETSIZE {
            let idx = fd / NFDBITS;
            let bit = fd % NFDBITS;
            (self.fds[idx] & (1 << bit)) != 0
        } else {
            false
        }
    }

    pub fn zero(&mut self) {
        for fd in self.fds.iter_mut() {
            *fd = 0;
        }
    }
}

#[repr(C)]
#[derive(Debug, Clone, Copy)]
pub struct Pollfd {
    pub fd: i32,
    pub events: i16,
    pub revents: i16,
}
