#![no_std]
#![feature(linkage)]
#![feature(panic_info_message)]
#![feature(alloc_error_handler)]

#[macro_use]
pub mod console;
mod lang_items;
mod syscall;

extern crate alloc;
#[macro_use]
extern crate bitflags;

use alloc::vec::Vec;
use buddy_system_allocator::LockedHeap;
use core::arch::global_asm;
use syscall::*;

global_asm!(include_str!("restore.S"));

const USER_HEAP_SIZE: usize = 32768;

static mut HEAP_SPACE: [u8; USER_HEAP_SIZE] = [0; USER_HEAP_SIZE];

#[global_allocator]
static HEAP: LockedHeap = LockedHeap::empty();

#[alloc_error_handler]
pub fn handle_alloc_error(layout: core::alloc::Layout) -> ! {
    panic!("Heap allocation error, layout = {:?}", layout);
}

#[no_mangle]
#[link_section = ".text.entry"]
pub extern "C" fn _start(argc: usize, argv: usize) -> ! {
    unsafe {
        HEAP.lock()
            .init(HEAP_SPACE.as_ptr() as usize, USER_HEAP_SIZE);
    }
    let mut v: Vec<&'static str> = Vec::new();
    for i in 0..argc {
        let str_start =
            unsafe { ((argv + i * core::mem::size_of::<usize>()) as *const usize).read_volatile() };
        let len = (0usize..)
            .find(|i| unsafe { ((str_start + *i) as *const u8).read_volatile() == 0 })
            .unwrap();
        v.push(
            core::str::from_utf8(unsafe {
                core::slice::from_raw_parts(str_start as *const u8, len)
            })
            .unwrap(),
        );
    }
    exit(main(argc, v.as_slice()));
}

// 弱符号
#[linkage = "weak"]
#[no_mangle]
fn main(_argc: usize, _argv: &[&str]) -> i32 {
    panic!("Cannot find main!");
}

bitflags! {
    pub struct OpenFlags: u32 {
        const RDONLY = 0;
        const WRONLY = 1 << 0;
        const RDWR = 1 << 1;
        const CREATE = 0x40;
        const O_EXCL = 0x80;
        const O_TRUNC = 0x200;
        const O_APPEND = 0x400;
        const O_NONBLOCK = 0x800;
        const LARGEFILE = 0x8000;
        const O_DIRECTORY = 0x10000;
        const CLOEXEC = 0x80000;
    }

    pub struct ModeFlags: u32 {
        // 权限
        const READ = 0o444;
        const WRITE = 0o222;
        const EXEC = 0o111;
        const RW = 0o600;
        const RWX = Self::READ.bits | Self::WRITE.bits | Self::EXEC.bits;
        // 文件类型
        const DIRECTORY = 0o40000;
        const REGULAR = 0o100000;
    }
}

pub fn dup(fd: u32) -> isize {
    sys_dup(fd)
}
pub fn open(path: &str, flags: OpenFlags) -> isize {
    sys_open(path, flags.bits)
}
pub fn mkdir(path: &str, mode: ModeFlags) -> isize {
    sys_mkdirat(path, mode.bits())
}
pub fn close(fd: u32) -> isize {
    sys_close(fd)
}
pub fn pipe(pipe_fd: &mut [u32]) -> isize {
    sys_pipe(pipe_fd)
}
pub fn read(fd: usize, buf: &mut [u8]) -> isize {
    sys_read(fd, buf)
}
pub fn write(fd: usize, buf: &[u8]) -> isize {
    sys_write(fd, buf)
}
pub fn exit(exit_code: i32) -> ! {
    sys_exit(exit_code);
}
pub fn yield_() -> isize {
    sys_yield()
}

pub struct Timespec {
    pub tv_sec: u64,
    pub tv_nsec: u64,
}

pub fn sleep(time: usize) -> isize {
    let mut req = Timespec {
        tv_sec: time as u64,
        tv_nsec: 0,
    };
    sys_nanosleep(&mut req)
}
pub fn getpid() -> isize {
    sys_getpid()
}
pub fn fork() -> isize {
    sys_fork()
}
pub fn exec(path: &str, args: &[*const u8], envp: &[*const u8]) -> isize {
    sys_exec(path, args, envp)
}
pub fn wait(exit_code: &mut i32) -> isize {
    sys_waitpid(-1, exit_code as *mut _)
}
pub fn waitpid(pid: usize, exit_code: &mut i32) -> isize {
    sys_waitpid(pid as isize, exit_code as *mut _)
}

pub const SIGDEF: i32 = 0; // Default signal handling
pub const SIGHUP: i32 = 1;
pub const SIGINT: i32 = 2;
pub const SIGQUIT: i32 = 3;
pub const SIGILL: i32 = 4;
pub const SIGTRAP: i32 = 5;
pub const SIGABRT: i32 = 6;
pub const SIGBUS: i32 = 7;
pub const SIGFPE: i32 = 8;
pub const SIGKILL: i32 = 9;
pub const SIGUSR1: i32 = 10;
pub const SIGSEGV: i32 = 11;
pub const SIGUSR2: i32 = 12;
pub const SIGPIPE: i32 = 13;
pub const SIGALRM: i32 = 14;
pub const SIGTERM: i32 = 15;
pub const SIGSTKFLT: i32 = 16;
pub const SIGCHLD: i32 = 17;
pub const SIGCONT: i32 = 18;
pub const SIGSTOP: i32 = 19;
pub const SIGTSTP: i32 = 20;
pub const SIGTTIN: i32 = 21;
pub const SIGTTOU: i32 = 22;
pub const SIGURG: i32 = 23;
pub const SIGXCPU: i32 = 24;
pub const SIGXFSZ: i32 = 25;
pub const SIGVTALRM: i32 = 26;
pub const SIGPROF: i32 = 27;
pub const SIGWINCH: i32 = 28;
pub const SIGIO: i32 = 29;
pub const SIGPWR: i32 = 30;
pub const SIGSYS: i32 = 31;

bitflags! {
    pub struct SignalFlags: u64 {
        const SIGHUP = 1 << (1 - 1);
        const SIGINT = 1 << (2 - 1);
        const SIGQUIT = 1 << (3 - 1);
        const SIGILL = 1 << (4 - 1);
        const SIGTRAP = 1 << (5 - 1);
        const SIGABRT = 1 << (6 - 1);
        const SIGBUS = 1 << (7 - 1);
        const SIGFPE = 1 << (8 - 1);
        const SIGKILL = 1 << (9 - 1);
        const SIGUSR1 = 1 << (10 - 1);
        const SIGSEGV = 1 << (11 - 1);
        const SIGUSR2 = 1 << (12 - 1);
        const SIGPIPE = 1 << (13 - 1);
        const SIGALRM = 1 << (14 - 1);
        const SIGTERM = 1 << (15 - 1);
        const SIGSTKFLT = 1 << (16 - 1);
        const SIGCHLD = 1 << (17 - 1);
        const SIGCONT = 1 << (18 - 1);
        const SIGSTOP = 1 << (19 - 1);
        const SIGTSTP = 1 << (20 - 1);
        const SIGTTIN = 1 << (21 - 1);
        const SIGTTOU = 1 << (22 - 1);
        const SIGURG = 1 << (23 - 1);
        const SIGXCPU = 1 << (24 - 1);
        const SIGXFSZ = 1 << (25 - 1);
        const SIGVTALRM = 1 << (26 - 1);
        const SIGPROF = 1 << (27 - 1);
        const SIGWINCH = 1 << (28 - 1);
        const SIGIO = 1 << (29 - 1);
        const SIGPWR = 1 << (30 - 1);
        const SIGSYS = 1 << (31 - 1);
    }
}

/// Action for a signal
#[repr(C)]
#[derive(Debug, Clone, Copy)]
pub struct SignalAction {
    pub handler: usize,
    pub flags: usize,
    restorer: usize,
    pub mask: SignalFlags,
}

impl Default for SignalAction {
    fn default() -> Self {
        Self {
            handler: 0,
            flags: 0,
            restorer: {
                extern "C" {
                    fn __restore();
                }
                __restore as usize
            },
            mask: SignalFlags::from_bits(0).unwrap(),
        }
    }
}

pub fn sigaction(
    signum: i32,
    action: Option<&SignalAction>,
    old_action: Option<&mut SignalAction>,
) -> isize {
    sys_sigaction(
        signum,
        action.map_or(core::ptr::null(), |a| a),
        old_action.map_or(core::ptr::null_mut(), |a| a),
    )
}

pub const SIG_BLOCK: u32 = 0;
pub const SIG_UNBLOCK: u32 = 1;
pub const SIG_SETMASK: u32 = 2;

pub fn sigprocmask(
    how: u32,
    set: Option<&SignalFlags>,
    old_set: Option<&mut SignalFlags>,
) -> isize {
    sys_sigprocmask(
        how,
        set.map_or(core::ptr::null(), |a| a),
        old_set.map_or(core::ptr::null_mut(), |a| a),
    )
}

pub fn kill(pid: usize, signum: i32) -> isize {
    sys_kill(pid, signum)
}

pub fn mount(source: &str, target: &str, filesystemtype: &str) {
    sys_mount(source, target, filesystemtype, 0, core::ptr::null());
}

pub fn umount(target: &str) {
    sys_umount2(target, 0);
}

pub fn shutdown() -> isize {
    sys_shutdown()
}
