use crate::{SignalAction, SignalFlags, Timespec};
use core::arch::asm;

const SYSCALL_DUP: usize = 23;
const SYSCALL_MKDIRAT: usize = 34;
const SYSCALL_UMOUNT2: usize = 39;
const SYSCALL_MOUNT: usize = 40;
const SYSCALL_OPENAT: usize = 56;
const SYSCALL_CLOSE: usize = 57;
const SYSCALL_PIPE: usize = 59;
const SYSCALL_READ: usize = 63;
const SYSCALL_WRITE: usize = 64;
const SYSCALL_EXIT_GROUP: usize = 94;
const SYSCALL_NANOSLEEP: usize = 101;
const SYSCALL_SCHED_YIELD: usize = 124;
const SYSCALL_KILL: usize = 129;
const SYSCALL_SIGACTION: usize = 134;
const SYSCALL_SIGPROCMASK: usize = 135;
const SYSCALL_GETPID: usize = 172;
const SYSCALL_CLONE: usize = 220;
const SYSCALL_EXEC: usize = 221;
const SYSCALL_WAIT4: usize = 260;
const SYSCALL_SHUTDOWN: usize = 0;

const AT_FDCWD: i32 = -100;

fn syscall(id: usize, args: [usize; 6]) -> isize {
    let mut ret: isize;
    unsafe {
        asm!(
            "ecall",
            inlateout("x10") args[0] => ret,
            in("x11") args[1],
            in("x12") args[2],
            in("x13") args[3],
            in("x14") args[4],
            in("x15") args[5],
            in("x17") id
        );
    }
    ret
}

pub fn sys_dup(fd: u32) -> isize {
    syscall(SYSCALL_DUP, [fd as usize, 0, 0, 0, 0, 0])
}

pub fn sys_open(path: &str, flags: u32) -> isize {
    syscall(
        SYSCALL_OPENAT,
        [
            AT_FDCWD as usize,
            path.as_ptr() as usize,
            flags as usize,
            0o777,
            0,
            0,
        ],
    )
}

pub fn sys_mkdirat(path: &str, mode: u32) -> isize {
    syscall(
        SYSCALL_MKDIRAT,
        [
            AT_FDCWD as usize,
            path.as_ptr() as usize,
            mode as usize,
            0,
            0,
            0,
        ],
    )
}

pub fn sys_close(fd: u32) -> isize {
    syscall(SYSCALL_CLOSE, [fd as usize, 0, 0, 0, 0, 0])
}

pub fn sys_pipe(pipe: &mut [u32]) -> isize {
    syscall(SYSCALL_PIPE, [pipe.as_mut_ptr() as usize, 0, 0, 0, 0, 0])
}

pub fn sys_read(fd: usize, buffer: &mut [u8]) -> isize {
    syscall(
        SYSCALL_READ,
        [fd, buffer.as_mut_ptr() as usize, buffer.len(), 0, 0, 0],
    )
}

pub fn sys_write(fd: usize, buffer: &[u8]) -> isize {
    syscall(
        SYSCALL_WRITE,
        [fd, buffer.as_ptr() as usize, buffer.len(), 0, 0, 0],
    )
}

pub fn sys_exit(exit_code: i32) -> ! {
    syscall(SYSCALL_EXIT_GROUP, [exit_code as usize, 0, 0, 0, 0, 0]);
    panic!("sys_exit never returns!");
}

pub fn sys_yield() -> isize {
    syscall(SYSCALL_SCHED_YIELD, [0, 0, 0, 0, 0, 0])
}

pub fn sys_nanosleep(req: *mut Timespec) -> isize {
    syscall(SYSCALL_NANOSLEEP, [req as usize, 0, 0, 0, 0, 0])
}

pub fn sys_kill(pid: usize, signal: i32) -> isize {
    syscall(SYSCALL_KILL, [pid, signal as usize, 0, 0, 0, 0])
}

pub fn sys_sigaction(
    signum: i32,
    action: *const SignalAction,
    old_action: *mut SignalAction,
) -> isize {
    syscall(
        SYSCALL_SIGACTION,
        [
            signum as usize,
            action as usize,
            old_action as usize,
            0,
            0,
            0,
        ],
    )
}

pub fn sys_sigprocmask(how: u32, set: *const SignalFlags, old_set: *mut SignalFlags) -> isize {
    syscall(
        SYSCALL_SIGPROCMASK,
        [how as usize, set as usize, old_set as usize, 0, 0, 0],
    )
}

pub fn sys_getpid() -> isize {
    syscall(SYSCALL_GETPID, [0, 0, 0, 0, 0, 0])
}

pub fn sys_fork() -> isize {
    syscall(SYSCALL_CLONE, [0, 0, 0, 0, 0, 0])
}

pub fn sys_exec(path: &str, args: &[*const u8], envp: &[*const u8]) -> isize {
    syscall(
        SYSCALL_EXEC,
        [
            path.as_ptr() as usize,
            args.as_ptr() as usize,
            envp.as_ptr() as usize,
            0,
            0,
            0,
        ],
    )
}

pub fn sys_waitpid(pid: isize, exit_code: *mut i32) -> isize {
    syscall(
        SYSCALL_WAIT4,
        [pid as usize, exit_code as usize, 0, 0, 0, 0],
    )
}

pub fn sys_mount(
    source: &str,
    target: &str,
    filesystemtype: &str,
    mountflags: usize,
    data: *const usize,
) -> isize {
    syscall(
        SYSCALL_MOUNT,
        [
            source.as_ptr() as usize,
            target.as_ptr() as usize,
            filesystemtype.as_ptr() as usize,
            mountflags,
            data as usize,
            0,
        ],
    )
}

pub fn sys_umount2(target: &str, flags: i32) -> isize {
    syscall(
        SYSCALL_UMOUNT2,
        [target.as_ptr() as usize, flags as usize, 0, 0, 0, 0],
    )
}

pub fn sys_shutdown() -> isize {
    syscall(SYSCALL_SHUTDOWN, [0, 0, 0, 0, 0, 0])
}
